/**********************************************************************
Field Info Display

Author: Cameron Hummels <chummels@gmail.com>
Affiliation: Columbia
Author: Jeffrey S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: MSU
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
***********************************************************************/

Ext.define("Reason.view.widgets.FieldPanel", {
    extend: 'Ext.panel.Panel',
    alias: 'widget.fieldpanel',
    iconCls: 'graph',
    autoScroll: true,
    layout: {
        type: 'vbox',
        pack: 'start',
        align: 'stretch',
    },
    width: '100%',
    height: '100%',

    items: [ 
        {
          xtype: 'combo',
          text: 'Field',
          itemId: 'fieldSelector',
          height: 30,
          width: 700,
          queryMode: 'local',
          editable: false,
          triggerAction: 'all',
          validateOnBlur: false,
        }, {
          xtype: 'panel',
          title: 'Field Source',
          itemId: 'fieldSourcePanel',
          flex: 3.0,
          width: 700,
          autoScroll: true,
          bodyCls: 'pfdisplay',
        }, {
          xtype: 'panel',
          title: 'Field Parameters',
          height: 200,
          width: 700,
          autoScroll: true,
          bodyCls: 'pfdisplay',
        }
    ],
});

