/* Generated by Cython 0.17.1 on Fri Mar  1 10:41:36 2013 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02040000
    #error Cython requires Python 2.4+.
#else
#include <stddef.h> /* For offsetof */
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if PY_VERSION_HEX < 0x02050000
  typedef int Py_ssize_t;
  #define PY_SSIZE_T_MAX INT_MAX
  #define PY_SSIZE_T_MIN INT_MIN
  #define PY_FORMAT_SIZE_T ""
  #define CYTHON_FORMAT_SSIZE_T ""
  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
  #define PyInt_AsSsize_t(o)   __Pyx_PyInt_AsInt(o)
  #define PyNumber_Index(o)    ((PyNumber_Check(o) && !PyFloat_Check(o)) ? PyNumber_Int(o) : \
                                (PyErr_Format(PyExc_TypeError, \
                                              "expected index value, got %.200s", Py_TYPE(o)->tp_name), \
                                 (PyObject*)0))
  #define PyIndex_Check(o)     (PyNumber_Check(o) && !PyFloat_Check(o) && !PyComplex_Check(o))
  #define PyErr_WarnEx(category, message, stacklevel) PyErr_Warn(category, message)
  #define __PYX_BUILD_PY_SSIZE_T "i"
#else
  #define __PYX_BUILD_PY_SSIZE_T "n"
  #define CYTHON_FORMAT_SSIZE_T "z"
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_REFCNT(ob) (((PyObject*)(ob))->ob_refcnt)
  #define Py_TYPE(ob)   (((PyObject*)(ob))->ob_type)
  #define Py_SIZE(ob)   (((PyVarObject*)(ob))->ob_size)
  #define PyVarObject_HEAD_INIT(type, size) \
          PyObject_HEAD_INIT(type) size,
  #define PyType_Modified(t)
  typedef struct {
     void *buf;
     PyObject *obj;
     Py_ssize_t len;
     Py_ssize_t itemsize;
     int readonly;
     int ndim;
     char *format;
     Py_ssize_t *shape;
     Py_ssize_t *strides;
     Py_ssize_t *suboffsets;
     void *internal;
  } Py_buffer;
  #define PyBUF_SIMPLE 0
  #define PyBUF_WRITABLE 0x0001
  #define PyBUF_FORMAT 0x0004
  #define PyBUF_ND 0x0008
  #define PyBUF_STRIDES (0x0010 | PyBUF_ND)
  #define PyBUF_C_CONTIGUOUS (0x0020 | PyBUF_STRIDES)
  #define PyBUF_F_CONTIGUOUS (0x0040 | PyBUF_STRIDES)
  #define PyBUF_ANY_CONTIGUOUS (0x0080 | PyBUF_STRIDES)
  #define PyBUF_INDIRECT (0x0100 | PyBUF_STRIDES)
  #define PyBUF_RECORDS (PyBUF_STRIDES | PyBUF_FORMAT | PyBUF_WRITABLE)
  #define PyBUF_FULL (PyBUF_INDIRECT | PyBUF_FORMAT | PyBUF_WRITABLE)
  typedef int (*getbufferproc)(PyObject *, Py_buffer *, int);
  typedef void (*releasebufferproc)(PyObject *, Py_buffer *);
#endif
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#if PY_MAJOR_VERSION < 3 && PY_MINOR_VERSION < 6
  #define PyUnicode_FromString(s) PyUnicode_Decode(s, strlen(s), "UTF-8", "strict")
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#if (PY_VERSION_HEX < 0x02060000) || (PY_MAJOR_VERSION >= 3)
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((k=k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PyBytesObject                PyStringObject
  #define PyBytes_Type                 PyString_Type
  #define PyBytes_Check                PyString_Check
  #define PyBytes_CheckExact           PyString_CheckExact
  #define PyBytes_FromString           PyString_FromString
  #define PyBytes_FromStringAndSize    PyString_FromStringAndSize
  #define PyBytes_FromFormat           PyString_FromFormat
  #define PyBytes_DecodeEscape         PyString_DecodeEscape
  #define PyBytes_AsString             PyString_AsString
  #define PyBytes_AsStringAndSize      PyString_AsStringAndSize
  #define PyBytes_Size                 PyString_Size
  #define PyBytes_AS_STRING            PyString_AS_STRING
  #define PyBytes_GET_SIZE             PyString_GET_SIZE
  #define PyBytes_Repr                 PyString_Repr
  #define PyBytes_Concat               PyString_Concat
  #define PyBytes_ConcatAndDel         PyString_ConcatAndDel
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PySet_Check(obj)             PyObject_TypeCheck(obj, &PySet_Type)
  #define PyFrozenSet_Check(obj)       PyObject_TypeCheck(obj, &PyFrozenSet_Type)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_VERSION_HEX < 0x03020000
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if (PY_MAJOR_VERSION < 3) || (PY_VERSION_HEX >= 0x03010300)
  #define __Pyx_PySequence_GetSlice(obj, a, b) PySequence_GetSlice(obj, a, b)
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) PySequence_SetSlice(obj, a, b, value)
  #define __Pyx_PySequence_DelSlice(obj, a, b) PySequence_DelSlice(obj, a, b)
#else
  #define __Pyx_PySequence_GetSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), (PyObject*)0) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_GetSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object is unsliceable", (obj)->ob_type->tp_name), (PyObject*)0)))
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_SetSlice(obj, a, b, value)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice assignment", (obj)->ob_type->tp_name), -1)))
  #define __Pyx_PySequence_DelSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_DelSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice deletion", (obj)->ob_type->tp_name), -1)))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),((char *)(n)))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),((char *)(n)),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),((char *)(n)))
#else
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),(n))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),(n),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),(n))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_NAMESTR(n) ((char *)(n))
  #define __Pyx_DOCSTR(n)  ((char *)(n))
#else
  #define __Pyx_NAMESTR(n) (n)
  #define __Pyx_DOCSTR(n)  (n)
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE__yt__utilities__lib__misc_utilities
#define __PYX_HAVE_API__yt__utilities__lib__misc_utilities
#include "stdio.h"
#include "stdlib.h"
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif


/* inline attribute */
#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* unused attribute */
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif

typedef struct {PyObject **p; char *s; const long n; const char* encoding; const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry; /*proto*/


/* Type Conversion Predeclarations */

#define __Pyx_PyBytes_FromUString(s) PyBytes_FromString((char*)s)
#define __Pyx_PyBytes_AsUString(s)   ((unsigned char*) PyBytes_AsString(s))

#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);

static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject*);

#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))

#ifdef __GNUC__
  /* Test for GCC > 2.95 */
  #if __GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95))
    #define likely(x)   __builtin_expect(!!(x), 1)
    #define unlikely(x) __builtin_expect(!!(x), 0)
  #else /* __GNUC__ > 2 ... */
    #define likely(x)   (x)
    #define unlikely(x) (x)
  #endif /* __GNUC__ > 2 ... */
#else /* __GNUC__ */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
    
static PyObject *__pyx_m;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "misc_utilities.pyx",
  "numpy.pxd",
  "type.pxd",
};
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name; /* for error messages only */
  struct __Pyx_StructField_* fields;
  size_t size;     /* sizeof(type) */
  size_t arraysize[8]; /* length of array in each dimension */
  int ndim;
  char typegroup; /* _R_eal, _C_omplex, Signed _I_nt, _U_nsigned int, _S_truct, _P_ointer, _O_bject, c_H_ar */
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "numpy.pxd":723
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "numpy.pxd":724
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "numpy.pxd":725
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "numpy.pxd":726
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "numpy.pxd":730
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "numpy.pxd":731
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "numpy.pxd":732
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "numpy.pxd":733
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "numpy.pxd":737
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "numpy.pxd":738
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "numpy.pxd":747
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "numpy.pxd":748
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "numpy.pxd":749
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "numpy.pxd":751
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "numpy.pxd":752
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "numpy.pxd":753
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "numpy.pxd":755
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "numpy.pxd":756
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "numpy.pxd":758
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "numpy.pxd":759
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "numpy.pxd":760
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif


/*--- Type declarations ---*/

/* "numpy.pxd":762
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "numpy.pxd":763
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "numpy.pxd":764
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "numpy.pxd":766
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname); /*proto*/
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif /* CYTHON_REFNANNY */
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name); /*proto*/

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found); /*proto*/

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name); /*proto*/

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name); /*proto*/

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact); /*proto*/

static CYTHON_INLINE int  __Pyx_GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb); /*proto*/
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb); /*proto*/

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
#define __Pyx_BufPtrStrided3d(type, buf, i0, s0, i1, s1, i2, s2) (type)((char*)buf + i0 * s0 + i1 * s1 + i2 * s2)
static void __Pyx_RaiseBufferIndexError(int axis); /*proto*/

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type); /*proto*/

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
#define __Pyx_GetItemInt_List(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_List_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyList_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyList_GET_ITEM(o, PyList_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt_Tuple(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Tuple_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyTuple_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyTuple_GET_ITEM(o, PyTuple_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (PyList_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyList_GET_SIZE(o)))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {  /* inlined PySequence_GetItem() */
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (unlikely(l < 0)) return NULL;
                i += l;
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

static void __Pyx_RaiseBufferFallbackError(void); /*proto*/

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause); /*proto*/

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

static CYTHON_INLINE int __Pyx_IterFinish(void); /*proto*/

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected); /*proto*/

typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


static Py_ssize_t __Pyx_zeros[] = {0, 0, 0, 0, 0, 0, 0, 0};
static Py_ssize_t __Pyx_minusones[] = {-1, -1, -1, -1, -1, -1, -1, -1};

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, long level); /*proto*/

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t);

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_npy_int64(npy_int64);

#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(_WIN32) && defined(__cplusplus) && CYTHON_CCOMPLEX
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eqf(a, b)   ((a)==(b))
    #define __Pyx_c_sumf(a, b)  ((a)+(b))
    #define __Pyx_c_difff(a, b) ((a)-(b))
    #define __Pyx_c_prodf(a, b) ((a)*(b))
    #define __Pyx_c_quotf(a, b) ((a)/(b))
    #define __Pyx_c_negf(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zerof(z) ((z)==(float)0)
    #define __Pyx_c_conjf(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_absf(z)     (::std::abs(z))
        #define __Pyx_c_powf(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zerof(z) ((z)==0)
    #define __Pyx_c_conjf(z)    (conjf(z))
    #if 1
        #define __Pyx_c_absf(z)     (cabsf(z))
        #define __Pyx_c_powf(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq(a, b)   ((a)==(b))
    #define __Pyx_c_sum(a, b)  ((a)+(b))
    #define __Pyx_c_diff(a, b) ((a)-(b))
    #define __Pyx_c_prod(a, b) ((a)*(b))
    #define __Pyx_c_quot(a, b) ((a)/(b))
    #define __Pyx_c_neg(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero(z) ((z)==(double)0)
    #define __Pyx_c_conj(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (::std::abs(z))
        #define __Pyx_c_pow(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero(z) ((z)==0)
    #define __Pyx_c_conj(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (cabs(z))
        #define __Pyx_c_pow(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject *);

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject *);

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject *);

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject *);

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject *);

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject *);

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject *);

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject *);

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject *);

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject *);

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject *);

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject *);

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name); /*proto*/

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);  /*proto*/

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename); /*proto*/

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t); /*proto*/


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'yt.utilities.lib.misc_utilities' */
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t = { "int64_t", NULL, sizeof(__pyx_t_5numpy_int64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int64_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t = { "uint8_t", NULL, sizeof(__pyx_t_5numpy_uint8_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint8_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint8_t), 0 };
#define __Pyx_MODULE_NAME "yt.utilities.lib.misc_utilities"
int __pyx_module_is_main_yt__utilities__lib__misc_utilities = 0;

/* Implementation of 'yt.utilities.lib.misc_utilities' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_bin_profile1d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_bins_x, PyArrayObject *__pyx_v_wsource, PyArrayObject *__pyx_v_bsource, PyArrayObject *__pyx_v_wresult, PyArrayObject *__pyx_v_bresult, PyArrayObject *__pyx_v_mresult, PyArrayObject *__pyx_v_qresult, PyArrayObject *__pyx_v_used); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_2bin_profile2d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_bins_x, PyArrayObject *__pyx_v_bins_y, PyArrayObject *__pyx_v_wsource, PyArrayObject *__pyx_v_bsource, PyArrayObject *__pyx_v_wresult, PyArrayObject *__pyx_v_bresult, PyArrayObject *__pyx_v_mresult, PyArrayObject *__pyx_v_qresult, PyArrayObject *__pyx_v_used); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_4bin_profile3d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_bins_x, PyArrayObject *__pyx_v_bins_y, PyArrayObject *__pyx_v_bins_z, PyArrayObject *__pyx_v_wsource, PyArrayObject *__pyx_v_bsource, PyArrayObject *__pyx_v_wresult, PyArrayObject *__pyx_v_bresult, PyArrayObject *__pyx_v_mresult, PyArrayObject *__pyx_v_qresult, PyArrayObject *__pyx_v_used); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_6lines(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_image, PyArrayObject *__pyx_v_xs, PyArrayObject *__pyx_v_ys, PyArrayObject *__pyx_v_colors, int __pyx_v_points_per_color); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_8rotate_vectors(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_vecs, PyArrayObject *__pyx_v_R); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_10get_color_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_px, PyArrayObject *__pyx_v_py, PyArrayObject *__pyx_v_pdx, PyArrayObject *__pyx_v_pdy, PyArrayObject *__pyx_v_value, __pyx_t_5numpy_float64_t __pyx_v_leftx, __pyx_t_5numpy_float64_t __pyx_v_rightx, __pyx_t_5numpy_float64_t __pyx_v_lefty, __pyx_t_5numpy_float64_t __pyx_v_righty, __pyx_t_5numpy_float64_t __pyx_v_mindx, __pyx_t_5numpy_float64_t __pyx_v_maxdx); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_12get_box_grids_level(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_left_edge, PyArrayObject *__pyx_v_right_edge, int __pyx_v_level, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels, PyArrayObject *__pyx_v_mask, int __pyx_v_min_index); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_14get_box_grids_below_level(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_left_edge, PyArrayObject *__pyx_v_right_edge, int __pyx_v_level, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels, PyArrayObject *__pyx_v_mask, int __pyx_v_min_level); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_16find_values_at_point(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_point, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_dimensions, PyObject *__pyx_v_field_names, PyObject *__pyx_v_grid_objects); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_18kdtree_get_choices(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_data, PyArrayObject *__pyx_v_l_corner, PyArrayObject *__pyx_v_r_corner); /* proto */
static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_20grow_flagging_field(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_oofield); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static char __pyx_k_5[] = "ndarray is not C contiguous";
static char __pyx_k_7[] = "ndarray is not Fortran contiguous";
static char __pyx_k_9[] = "Non-native byte order not supported";
static char __pyx_k_11[] = "unknown dtype code in numpy.pxd (%d)";
static char __pyx_k_12[] = "Format string allocated too short, see comment in numpy.pxd";
static char __pyx_k_15[] = "Format string allocated too short.";
static char __pyx_k_17[] = "\nSimple utilities that don't fit anywhere else\n\nAuthor: Matthew Turk <matthewturk@gmail.com>\nAffiliation: Columbia University\nHomepage: http://yt-project.org/\nLicense:\n  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.\n\n  This file is part of yt.\n\n  yt is free software; you can redistribute it and/or modify\n  it under the terms of the GNU General Public License as published by\n  the Free Software Foundation; either version 3 of the License, or\n  (at your option) any later version.\n\n  This program is distributed in the hope that it will be useful,\n  but WITHOUT ANY WARRANTY; without even the implied warranty of\n  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n  GNU General Public License for more details.\n\n  You should have received a copy of the GNU General Public License\n  along with this program.  If not, see <http://www.gnu.org/licenses/>.\n";
static char __pyx_k_20[] = "/home/mturk/yt/yt-clean/yt/utilities/lib/misc_utilities.pyx";
static char __pyx_k_21[] = "yt.utilities.lib.misc_utilities";
static char __pyx_k_36[] = "get_box_grids_below_level";
static char __pyx_k_39[] = "find_values_at_point";
static char __pyx_k__B[] = "B";
static char __pyx_k__H[] = "H";
static char __pyx_k__I[] = "I";
static char __pyx_k__L[] = "L";
static char __pyx_k__O[] = "O";
static char __pyx_k__Q[] = "Q";
static char __pyx_k__R[] = "R";
static char __pyx_k__b[] = "b";
static char __pyx_k__d[] = "d";
static char __pyx_k__f[] = "f";
static char __pyx_k__g[] = "g";
static char __pyx_k__h[] = "h";
static char __pyx_k__i[] = "i";
static char __pyx_k__j[] = "j";
static char __pyx_k__k[] = "k";
static char __pyx_k__l[] = "l";
static char __pyx_k__n[] = "n";
static char __pyx_k__q[] = "q";
static char __pyx_k__v[] = "v";
static char __pyx_k__Zd[] = "Zd";
static char __pyx_k__Zf[] = "Zf";
static char __pyx_k__Zg[] = "Zg";
static char __pyx_k__dx[] = "dx";
static char __pyx_k__dy[] = "dy";
static char __pyx_k__e2[] = "e2";
static char __pyx_k__fi[] = "fi";
static char __pyx_k__gi[] = "gi";
static char __pyx_k__ma[] = "ma";
static char __pyx_k__mi[] = "mi";
static char __pyx_k__nf[] = "nf";
static char __pyx_k__ni[] = "ni";
static char __pyx_k__nj[] = "nj";
static char __pyx_k__nk[] = "nk";
static char __pyx_k__nl[] = "nl";
static char __pyx_k__np[] = "np";
static char __pyx_k__nx[] = "nx";
static char __pyx_k__ny[] = "ny";
static char __pyx_k__oi[] = "oi";
static char __pyx_k__oj[] = "oj";
static char __pyx_k__ok[] = "ok";
static char __pyx_k__px[] = "px";
static char __pyx_k__py[] = "py";
static char __pyx_k__rv[] = "rv";
static char __pyx_k__sx[] = "sx";
static char __pyx_k__sy[] = "sy";
static char __pyx_k__x0[] = "x0";
static char __pyx_k__x1[] = "x1";
static char __pyx_k__xs[] = "xs";
static char __pyx_k__y0[] = "y0";
static char __pyx_k__y1[] = "y1";
static char __pyx_k__ys[] = "ys";
static char __pyx_k__bin[] = "bin";
static char __pyx_k__dds[] = "dds";
static char __pyx_k__dim[] = "dim";
static char __pyx_k__eps[] = "eps";
static char __pyx_k__err[] = "err";
static char __pyx_k__ind[] = "ind";
static char __pyx_k__pdx[] = "pdx";
static char __pyx_k__pdy[] = "pdy";
static char __pyx_k__bini[] = "bini";
static char __pyx_k__binj[] = "binj";
static char __pyx_k__bink[] = "bink";
static char __pyx_k__bool[] = "bool";
static char __pyx_k__bval[] = "bval";
static char __pyx_k__data[] = "data";
static char __pyx_k__grid[] = "grid";
static char __pyx_k__mask[] = "mask";
static char __pyx_k__sort[] = "sort";
static char __pyx_k__tarr[] = "tarr";
static char __pyx_k__used[] = "used";
static char __pyx_k__vecs[] = "vecs";
static char __pyx_k__view[] = "view";
static char __pyx_k__wval[] = "wval";
static char __pyx_k__addit[] = "addit";
static char __pyx_k__alpha[] = "alpha";
static char __pyx_k__dtype[] = "dtype";
static char __pyx_k__empty[] = "empty";
static char __pyx_k__field[] = "field";
static char __pyx_k__finfo[] = "finfo";
static char __pyx_k__image[] = "image";
static char __pyx_k__leftx[] = "leftx";
static char __pyx_k__lefty[] = "lefty";
static char __pyx_k__level[] = "level";
static char __pyx_k__lines[] = "lines";
static char __pyx_k__maxdx[] = "maxdx";
static char __pyx_k__mindx[] = "mindx";
static char __pyx_k__numpy[] = "numpy";
static char __pyx_k__point[] = "point";
static char __pyx_k__range[] = "range";
static char __pyx_k__split[] = "split";
static char __pyx_k__uint8[] = "uint8";
static char __pyx_k__value[] = "value";
static char __pyx_k__zeros[] = "zeros";
static char __pyx_k__astype[] = "astype";
static char __pyx_k__bins_x[] = "bins_x";
static char __pyx_k__bins_y[] = "bins_y";
static char __pyx_k__bins_z[] = "bins_z";
static char __pyx_k__colors[] = "colors";
static char __pyx_k__inside[] = "inside";
static char __pyx_k__levels[] = "levels";
static char __pyx_k__my_max[] = "my_max";
static char __pyx_k__n_best[] = "n_best";
static char __pyx_k__nfield[] = "nfield";
static char __pyx_k__ofield[] = "ofield";
static char __pyx_k__rightx[] = "rightx";
static char __pyx_k__righty[] = "righty";
static char __pyx_k__skipit[] = "skipit";
static char __pyx_k__bresult[] = "bresult";
static char __pyx_k__bsource[] = "bsource";
static char __pyx_k__float64[] = "float64";
static char __pyx_k__mresult[] = "mresult";
static char __pyx_k__n_grids[] = "n_grids";
static char __pyx_k__oofield[] = "oofield";
static char __pyx_k__qresult[] = "qresult";
static char __pyx_k__rotated[] = "rotated";
static char __pyx_k__uniques[] = "uniques";
static char __pyx_k__wresult[] = "wresult";
static char __pyx_k__wsource[] = "wsource";
static char __pyx_k__KeyError[] = "KeyError";
static char __pyx_k____main__[] = "__main__";
static char __pyx_k____test__[] = "__test__";
static char __pyx_k__best_dim[] = "best_dim";
static char __pyx_k__l_corner[] = "l_corner";
static char __pyx_k__less_ids[] = "less_ids";
static char __pyx_k__my_split[] = "my_split";
static char __pyx_k__n_unique[] = "n_unique";
static char __pyx_k__r_corner[] = "r_corner";
static char __pyx_k__has_alpha[] = "has_alpha";
static char __pyx_k__left_edge[] = "left_edge";
static char __pyx_k__min_index[] = "min_index";
static char __pyx_k__min_level[] = "min_level";
static char __pyx_k__ValueError[] = "ValueError";
static char __pyx_k__dimensions[] = "dimensions";
static char __pyx_k__left_edges[] = "left_edges";
static char __pyx_k__right_edge[] = "right_edge";
static char __pyx_k__uniquedims[] = "uniquedims";
static char __pyx_k__zeros_like[] = "zeros_like";
static char __pyx_k__field_names[] = "field_names";
static char __pyx_k__greater_ids[] = "greater_ids";
static char __pyx_k__right_edges[] = "right_edges";
static char __pyx_k__RuntimeError[] = "RuntimeError";
static char __pyx_k__grid_objects[] = "grid_objects";
static char __pyx_k__bin_profile1d[] = "bin_profile1d";
static char __pyx_k__bin_profile2d[] = "bin_profile2d";
static char __pyx_k__bin_profile3d[] = "bin_profile3d";
static char __pyx_k__rotate_vectors[] = "rotate_vectors";
static char __pyx_k__get_color_bounds[] = "get_color_bounds";
static char __pyx_k__points_per_color[] = "points_per_color";
static char __pyx_k__kdtree_get_choices[] = "kdtree_get_choices";
static char __pyx_k__get_box_grids_level[] = "get_box_grids_level";
static char __pyx_k__grow_flagging_field[] = "grow_flagging_field";
static PyObject *__pyx_kp_u_11;
static PyObject *__pyx_kp_u_12;
static PyObject *__pyx_kp_u_15;
static PyObject *__pyx_kp_s_20;
static PyObject *__pyx_n_s_21;
static PyObject *__pyx_n_s_36;
static PyObject *__pyx_n_s_39;
static PyObject *__pyx_kp_u_5;
static PyObject *__pyx_kp_u_7;
static PyObject *__pyx_kp_u_9;
static PyObject *__pyx_n_s__KeyError;
static PyObject *__pyx_n_s__R;
static PyObject *__pyx_n_s__RuntimeError;
static PyObject *__pyx_n_s__ValueError;
static PyObject *__pyx_n_s____main__;
static PyObject *__pyx_n_s____test__;
static PyObject *__pyx_n_s__addit;
static PyObject *__pyx_n_s__alpha;
static PyObject *__pyx_n_s__astype;
static PyObject *__pyx_n_s__best_dim;
static PyObject *__pyx_n_s__bin;
static PyObject *__pyx_n_s__bin_profile1d;
static PyObject *__pyx_n_s__bin_profile2d;
static PyObject *__pyx_n_s__bin_profile3d;
static PyObject *__pyx_n_s__bini;
static PyObject *__pyx_n_s__binj;
static PyObject *__pyx_n_s__bink;
static PyObject *__pyx_n_s__bins_x;
static PyObject *__pyx_n_s__bins_y;
static PyObject *__pyx_n_s__bins_z;
static PyObject *__pyx_n_s__bool;
static PyObject *__pyx_n_s__bresult;
static PyObject *__pyx_n_s__bsource;
static PyObject *__pyx_n_s__bval;
static PyObject *__pyx_n_s__colors;
static PyObject *__pyx_n_s__data;
static PyObject *__pyx_n_s__dds;
static PyObject *__pyx_n_s__dim;
static PyObject *__pyx_n_s__dimensions;
static PyObject *__pyx_n_s__dtype;
static PyObject *__pyx_n_s__dx;
static PyObject *__pyx_n_s__dy;
static PyObject *__pyx_n_s__e2;
static PyObject *__pyx_n_s__empty;
static PyObject *__pyx_n_s__eps;
static PyObject *__pyx_n_s__err;
static PyObject *__pyx_n_s__fi;
static PyObject *__pyx_n_s__field;
static PyObject *__pyx_n_s__field_names;
static PyObject *__pyx_n_s__finfo;
static PyObject *__pyx_n_s__float64;
static PyObject *__pyx_n_s__get_box_grids_level;
static PyObject *__pyx_n_s__get_color_bounds;
static PyObject *__pyx_n_s__gi;
static PyObject *__pyx_n_s__greater_ids;
static PyObject *__pyx_n_s__grid;
static PyObject *__pyx_n_s__grid_objects;
static PyObject *__pyx_n_s__grow_flagging_field;
static PyObject *__pyx_n_s__has_alpha;
static PyObject *__pyx_n_s__i;
static PyObject *__pyx_n_s__image;
static PyObject *__pyx_n_s__ind;
static PyObject *__pyx_n_s__inside;
static PyObject *__pyx_n_s__j;
static PyObject *__pyx_n_s__k;
static PyObject *__pyx_n_s__kdtree_get_choices;
static PyObject *__pyx_n_s__l_corner;
static PyObject *__pyx_n_s__left_edge;
static PyObject *__pyx_n_s__left_edges;
static PyObject *__pyx_n_s__leftx;
static PyObject *__pyx_n_s__lefty;
static PyObject *__pyx_n_s__less_ids;
static PyObject *__pyx_n_s__level;
static PyObject *__pyx_n_s__levels;
static PyObject *__pyx_n_s__lines;
static PyObject *__pyx_n_s__ma;
static PyObject *__pyx_n_s__mask;
static PyObject *__pyx_n_s__maxdx;
static PyObject *__pyx_n_s__mi;
static PyObject *__pyx_n_s__min_index;
static PyObject *__pyx_n_s__min_level;
static PyObject *__pyx_n_s__mindx;
static PyObject *__pyx_n_s__mresult;
static PyObject *__pyx_n_s__my_max;
static PyObject *__pyx_n_s__my_split;
static PyObject *__pyx_n_s__n;
static PyObject *__pyx_n_s__n_best;
static PyObject *__pyx_n_s__n_grids;
static PyObject *__pyx_n_s__n_unique;
static PyObject *__pyx_n_s__nf;
static PyObject *__pyx_n_s__nfield;
static PyObject *__pyx_n_s__ni;
static PyObject *__pyx_n_s__nj;
static PyObject *__pyx_n_s__nk;
static PyObject *__pyx_n_s__nl;
static PyObject *__pyx_n_s__np;
static PyObject *__pyx_n_s__numpy;
static PyObject *__pyx_n_s__nx;
static PyObject *__pyx_n_s__ny;
static PyObject *__pyx_n_s__ofield;
static PyObject *__pyx_n_s__oi;
static PyObject *__pyx_n_s__oj;
static PyObject *__pyx_n_s__ok;
static PyObject *__pyx_n_s__oofield;
static PyObject *__pyx_n_s__pdx;
static PyObject *__pyx_n_s__pdy;
static PyObject *__pyx_n_s__point;
static PyObject *__pyx_n_s__points_per_color;
static PyObject *__pyx_n_s__px;
static PyObject *__pyx_n_s__py;
static PyObject *__pyx_n_s__qresult;
static PyObject *__pyx_n_s__r_corner;
static PyObject *__pyx_n_s__range;
static PyObject *__pyx_n_s__right_edge;
static PyObject *__pyx_n_s__right_edges;
static PyObject *__pyx_n_s__rightx;
static PyObject *__pyx_n_s__righty;
static PyObject *__pyx_n_s__rotate_vectors;
static PyObject *__pyx_n_s__rotated;
static PyObject *__pyx_n_s__rv;
static PyObject *__pyx_n_s__skipit;
static PyObject *__pyx_n_s__sort;
static PyObject *__pyx_n_s__split;
static PyObject *__pyx_n_s__sx;
static PyObject *__pyx_n_s__sy;
static PyObject *__pyx_n_s__tarr;
static PyObject *__pyx_n_s__uint8;
static PyObject *__pyx_n_s__uniquedims;
static PyObject *__pyx_n_s__uniques;
static PyObject *__pyx_n_s__used;
static PyObject *__pyx_n_s__v;
static PyObject *__pyx_n_s__value;
static PyObject *__pyx_n_s__vecs;
static PyObject *__pyx_n_s__view;
static PyObject *__pyx_n_s__wresult;
static PyObject *__pyx_n_s__wsource;
static PyObject *__pyx_n_s__wval;
static PyObject *__pyx_n_s__x0;
static PyObject *__pyx_n_s__x1;
static PyObject *__pyx_n_s__xs;
static PyObject *__pyx_n_s__y0;
static PyObject *__pyx_n_s__y1;
static PyObject *__pyx_n_s__ys;
static PyObject *__pyx_n_s__zeros;
static PyObject *__pyx_n_s__zeros_like;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_15;
static PyObject *__pyx_k_tuple_1;
static PyObject *__pyx_k_tuple_2;
static PyObject *__pyx_k_tuple_3;
static PyObject *__pyx_k_tuple_4;
static PyObject *__pyx_k_tuple_6;
static PyObject *__pyx_k_tuple_8;
static PyObject *__pyx_k_tuple_10;
static PyObject *__pyx_k_tuple_13;
static PyObject *__pyx_k_tuple_14;
static PyObject *__pyx_k_tuple_16;
static PyObject *__pyx_k_tuple_18;
static PyObject *__pyx_k_tuple_22;
static PyObject *__pyx_k_tuple_24;
static PyObject *__pyx_k_tuple_26;
static PyObject *__pyx_k_tuple_28;
static PyObject *__pyx_k_tuple_30;
static PyObject *__pyx_k_tuple_32;
static PyObject *__pyx_k_tuple_34;
static PyObject *__pyx_k_tuple_37;
static PyObject *__pyx_k_tuple_40;
static PyObject *__pyx_k_tuple_42;
static PyObject *__pyx_k_codeobj_19;
static PyObject *__pyx_k_codeobj_23;
static PyObject *__pyx_k_codeobj_25;
static PyObject *__pyx_k_codeobj_27;
static PyObject *__pyx_k_codeobj_29;
static PyObject *__pyx_k_codeobj_31;
static PyObject *__pyx_k_codeobj_33;
static PyObject *__pyx_k_codeobj_35;
static PyObject *__pyx_k_codeobj_38;
static PyObject *__pyx_k_codeobj_41;
static PyObject *__pyx_k_codeobj_43;

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_1bin_profile1d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_1bin_profile1d = {__Pyx_NAMESTR("bin_profile1d"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_1bin_profile1d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_1bin_profile1d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_bins_x = 0;
  PyArrayObject *__pyx_v_wsource = 0;
  PyArrayObject *__pyx_v_bsource = 0;
  PyArrayObject *__pyx_v_wresult = 0;
  PyArrayObject *__pyx_v_bresult = 0;
  PyArrayObject *__pyx_v_mresult = 0;
  PyArrayObject *__pyx_v_qresult = 0;
  PyArrayObject *__pyx_v_used = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin_profile1d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__bins_x,&__pyx_n_s__wsource,&__pyx_n_s__bsource,&__pyx_n_s__wresult,&__pyx_n_s__bresult,&__pyx_n_s__mresult,&__pyx_n_s__qresult,&__pyx_n_s__used,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bins_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__wsource)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bsource)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__wresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__qresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__used)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bin_profile1d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_bins_x = ((PyArrayObject *)values[0]);
    __pyx_v_wsource = ((PyArrayObject *)values[1]);
    __pyx_v_bsource = ((PyArrayObject *)values[2]);
    __pyx_v_wresult = ((PyArrayObject *)values[3]);
    __pyx_v_bresult = ((PyArrayObject *)values[4]);
    __pyx_v_mresult = ((PyArrayObject *)values[5]);
    __pyx_v_qresult = ((PyArrayObject *)values[6]);
    __pyx_v_used = ((PyArrayObject *)values[7]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bin_profile1d", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.bin_profile1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bins_x), __pyx_ptype_5numpy_ndarray, 1, "bins_x", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wsource), __pyx_ptype_5numpy_ndarray, 1, "wsource", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bsource), __pyx_ptype_5numpy_ndarray, 1, "bsource", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 39; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wresult), __pyx_ptype_5numpy_ndarray, 1, "wresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 40; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bresult), __pyx_ptype_5numpy_ndarray, 1, "bresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 41; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mresult), __pyx_ptype_5numpy_ndarray, 1, "mresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 42; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qresult), __pyx_ptype_5numpy_ndarray, 1, "qresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 43; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_used), __pyx_ptype_5numpy_ndarray, 1, "used", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 44; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_bin_profile1d(__pyx_self, __pyx_v_bins_x, __pyx_v_wsource, __pyx_v_bsource, __pyx_v_wresult, __pyx_v_bresult, __pyx_v_mresult, __pyx_v_qresult, __pyx_v_used);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":37
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile1d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.float64_t, ndim=1] wsource,
 *                   np.ndarray[np.float64_t, ndim=1] bsource,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_bin_profile1d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_bins_x, PyArrayObject *__pyx_v_wsource, PyArrayObject *__pyx_v_bsource, PyArrayObject *__pyx_v_wresult, PyArrayObject *__pyx_v_bresult, PyArrayObject *__pyx_v_mresult, PyArrayObject *__pyx_v_qresult, PyArrayObject *__pyx_v_used) {
  int __pyx_v_n;
  int __pyx_v_bin;
  __pyx_t_5numpy_float64_t __pyx_v_wval;
  __pyx_t_5numpy_float64_t __pyx_v_bval;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bins_x;
  __Pyx_Buffer __pyx_pybuffer_bins_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bresult;
  __Pyx_Buffer __pyx_pybuffer_bresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bsource;
  __Pyx_Buffer __pyx_pybuffer_bsource;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mresult;
  __Pyx_Buffer __pyx_pybuffer_mresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_qresult;
  __Pyx_Buffer __pyx_pybuffer_qresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_used;
  __Pyx_Buffer __pyx_pybuffer_used;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wresult;
  __Pyx_Buffer __pyx_pybuffer_wresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wsource;
  __Pyx_Buffer __pyx_pybuffer_wsource;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  npy_intp __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin_profile1d", 0);
  __pyx_pybuffer_bins_x.pybuffer.buf = NULL;
  __pyx_pybuffer_bins_x.refcount = 0;
  __pyx_pybuffernd_bins_x.data = NULL;
  __pyx_pybuffernd_bins_x.rcbuffer = &__pyx_pybuffer_bins_x;
  __pyx_pybuffer_wsource.pybuffer.buf = NULL;
  __pyx_pybuffer_wsource.refcount = 0;
  __pyx_pybuffernd_wsource.data = NULL;
  __pyx_pybuffernd_wsource.rcbuffer = &__pyx_pybuffer_wsource;
  __pyx_pybuffer_bsource.pybuffer.buf = NULL;
  __pyx_pybuffer_bsource.refcount = 0;
  __pyx_pybuffernd_bsource.data = NULL;
  __pyx_pybuffernd_bsource.rcbuffer = &__pyx_pybuffer_bsource;
  __pyx_pybuffer_wresult.pybuffer.buf = NULL;
  __pyx_pybuffer_wresult.refcount = 0;
  __pyx_pybuffernd_wresult.data = NULL;
  __pyx_pybuffernd_wresult.rcbuffer = &__pyx_pybuffer_wresult;
  __pyx_pybuffer_bresult.pybuffer.buf = NULL;
  __pyx_pybuffer_bresult.refcount = 0;
  __pyx_pybuffernd_bresult.data = NULL;
  __pyx_pybuffernd_bresult.rcbuffer = &__pyx_pybuffer_bresult;
  __pyx_pybuffer_mresult.pybuffer.buf = NULL;
  __pyx_pybuffer_mresult.refcount = 0;
  __pyx_pybuffernd_mresult.data = NULL;
  __pyx_pybuffernd_mresult.rcbuffer = &__pyx_pybuffer_mresult;
  __pyx_pybuffer_qresult.pybuffer.buf = NULL;
  __pyx_pybuffer_qresult.refcount = 0;
  __pyx_pybuffernd_qresult.data = NULL;
  __pyx_pybuffernd_qresult.rcbuffer = &__pyx_pybuffer_qresult;
  __pyx_pybuffer_used.pybuffer.buf = NULL;
  __pyx_pybuffer_used.refcount = 0;
  __pyx_pybuffernd_used.data = NULL;
  __pyx_pybuffernd_used.rcbuffer = &__pyx_pybuffer_used;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_bins_x, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bins_x.diminfo[0].strides = __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bins_x.diminfo[0].shape = __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer, (PyObject*)__pyx_v_wsource, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_wsource.diminfo[0].strides = __pyx_pybuffernd_wsource.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wsource.diminfo[0].shape = __pyx_pybuffernd_wsource.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer, (PyObject*)__pyx_v_bsource, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bsource.diminfo[0].strides = __pyx_pybuffernd_bsource.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bsource.diminfo[0].shape = __pyx_pybuffernd_bsource.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_wresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_wresult.diminfo[0].strides = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wresult.diminfo[0].shape = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_bresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bresult.diminfo[0].strides = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bresult.diminfo[0].shape = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_mresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mresult.diminfo[0].strides = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mresult.diminfo[0].shape = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_qresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_qresult.diminfo[0].strides = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_qresult.diminfo[0].shape = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_used.rcbuffer->pybuffer, (PyObject*)__pyx_v_used, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_used.diminfo[0].strides = __pyx_pybuffernd_used.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_used.diminfo[0].shape = __pyx_pybuffernd_used.rcbuffer->pybuffer.shape[0];

  /* "yt/utilities/lib/misc_utilities.pyx":47
 *     cdef int n, bin
 *     cdef np.float64_t wval, bval
 *     for n in range(bins_x.shape[0]):             # <<<<<<<<<<<<<<
 *         bin = bins_x[n]
 *         bval = bsource[n]
 */
  __pyx_t_1 = (__pyx_v_bins_x->dimensions[0]);
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_n = __pyx_t_2;

    /* "yt/utilities/lib/misc_utilities.pyx":48
 *     cdef np.float64_t wval, bval
 *     for n in range(bins_x.shape[0]):
 *         bin = bins_x[n]             # <<<<<<<<<<<<<<
 *         bval = bsource[n]
 *         wval = wsource[n]
 */
    __pyx_t_3 = __pyx_v_n;
    __pyx_v_bin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_bins_x.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":49
 *     for n in range(bins_x.shape[0]):
 *         bin = bins_x[n]
 *         bval = bsource[n]             # <<<<<<<<<<<<<<
 *         wval = wsource[n]
 *         qresult[bin] += (wresult[bin] * wval * (bval - mresult[bin])**2) / \
 */
    __pyx_t_4 = __pyx_v_n;
    __pyx_v_bval = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bsource.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_bsource.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":50
 *         bin = bins_x[n]
 *         bval = bsource[n]
 *         wval = wsource[n]             # <<<<<<<<<<<<<<
 *         qresult[bin] += (wresult[bin] * wval * (bval - mresult[bin])**2) / \
 *             (wresult[bin] + wval)
 */
    __pyx_t_5 = __pyx_v_n;
    __pyx_v_wval = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wsource.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_wsource.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":51
 *         bval = bsource[n]
 *         wval = wsource[n]
 *         qresult[bin] += (wresult[bin] * wval * (bval - mresult[bin])**2) / \             # <<<<<<<<<<<<<<
 *             (wresult[bin] + wval)
 *         wresult[bin] += wval
 */
    __pyx_t_6 = __pyx_v_bin;
    __pyx_t_7 = __pyx_v_bin;

    /* "yt/utilities/lib/misc_utilities.pyx":52
 *         wval = wsource[n]
 *         qresult[bin] += (wresult[bin] * wval * (bval - mresult[bin])**2) / \
 *             (wresult[bin] + wval)             # <<<<<<<<<<<<<<
 *         wresult[bin] += wval
 *         bresult[bin] += wval*bval
 */
    __pyx_t_8 = __pyx_v_bin;

    /* "yt/utilities/lib/misc_utilities.pyx":51
 *         bval = bsource[n]
 *         wval = wsource[n]
 *         qresult[bin] += (wresult[bin] * wval * (bval - mresult[bin])**2) / \             # <<<<<<<<<<<<<<
 *             (wresult[bin] + wval)
 *         wresult[bin] += wval
 */
    __pyx_t_9 = __pyx_v_bin;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_qresult.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_qresult.diminfo[0].strides) += ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_wresult.diminfo[0].strides)) * __pyx_v_wval) * pow((__pyx_v_bval - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_mresult.diminfo[0].strides))), 2.0)) / ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_wresult.diminfo[0].strides)) + __pyx_v_wval));

    /* "yt/utilities/lib/misc_utilities.pyx":53
 *         qresult[bin] += (wresult[bin] * wval * (bval - mresult[bin])**2) / \
 *             (wresult[bin] + wval)
 *         wresult[bin] += wval             # <<<<<<<<<<<<<<
 *         bresult[bin] += wval*bval
 *         mresult[bin] += wval * (bval - mresult[bin]) / wresult[bin]
 */
    __pyx_t_10 = __pyx_v_bin;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_wresult.diminfo[0].strides) += __pyx_v_wval;

    /* "yt/utilities/lib/misc_utilities.pyx":54
 *             (wresult[bin] + wval)
 *         wresult[bin] += wval
 *         bresult[bin] += wval*bval             # <<<<<<<<<<<<<<
 *         mresult[bin] += wval * (bval - mresult[bin]) / wresult[bin]
 *         used[bin] = 1
 */
    __pyx_t_11 = __pyx_v_bin;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bresult.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_bresult.diminfo[0].strides) += (__pyx_v_wval * __pyx_v_bval);

    /* "yt/utilities/lib/misc_utilities.pyx":55
 *         wresult[bin] += wval
 *         bresult[bin] += wval*bval
 *         mresult[bin] += wval * (bval - mresult[bin]) / wresult[bin]             # <<<<<<<<<<<<<<
 *         used[bin] = 1
 *     return
 */
    __pyx_t_12 = __pyx_v_bin;
    __pyx_t_13 = __pyx_v_bin;
    __pyx_t_14 = __pyx_v_bin;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mresult.diminfo[0].strides) += ((__pyx_v_wval * (__pyx_v_bval - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_mresult.diminfo[0].strides)))) / (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_wresult.diminfo[0].strides)));

    /* "yt/utilities/lib/misc_utilities.pyx":56
 *         bresult[bin] += wval*bval
 *         mresult[bin] += wval * (bval - mresult[bin]) / wresult[bin]
 *         used[bin] = 1             # <<<<<<<<<<<<<<
 *     return
 * 
 */
    __pyx_t_15 = __pyx_v_bin;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_used.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_used.diminfo[0].strides) = 1.0;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":57
 *         mresult[bin] += wval * (bval - mresult[bin]) / wresult[bin]
 *         used[bin] = 1
 *     return             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_used.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.bin_profile1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_used.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_3bin_profile2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_3bin_profile2d = {__Pyx_NAMESTR("bin_profile2d"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_3bin_profile2d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_3bin_profile2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_bins_x = 0;
  PyArrayObject *__pyx_v_bins_y = 0;
  PyArrayObject *__pyx_v_wsource = 0;
  PyArrayObject *__pyx_v_bsource = 0;
  PyArrayObject *__pyx_v_wresult = 0;
  PyArrayObject *__pyx_v_bresult = 0;
  PyArrayObject *__pyx_v_mresult = 0;
  PyArrayObject *__pyx_v_qresult = 0;
  PyArrayObject *__pyx_v_used = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin_profile2d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__bins_x,&__pyx_n_s__bins_y,&__pyx_n_s__wsource,&__pyx_n_s__bsource,&__pyx_n_s__wresult,&__pyx_n_s__bresult,&__pyx_n_s__mresult,&__pyx_n_s__qresult,&__pyx_n_s__used,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bins_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bins_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__wsource)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bsource)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__wresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__qresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__used)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bin_profile2d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_bins_x = ((PyArrayObject *)values[0]);
    __pyx_v_bins_y = ((PyArrayObject *)values[1]);
    __pyx_v_wsource = ((PyArrayObject *)values[2]);
    __pyx_v_bsource = ((PyArrayObject *)values[3]);
    __pyx_v_wresult = ((PyArrayObject *)values[4]);
    __pyx_v_bresult = ((PyArrayObject *)values[5]);
    __pyx_v_mresult = ((PyArrayObject *)values[6]);
    __pyx_v_qresult = ((PyArrayObject *)values[7]);
    __pyx_v_used = ((PyArrayObject *)values[8]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bin_profile2d", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.bin_profile2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bins_x), __pyx_ptype_5numpy_ndarray, 1, "bins_x", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bins_y), __pyx_ptype_5numpy_ndarray, 1, "bins_y", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 63; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wsource), __pyx_ptype_5numpy_ndarray, 1, "wsource", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 64; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bsource), __pyx_ptype_5numpy_ndarray, 1, "bsource", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 65; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wresult), __pyx_ptype_5numpy_ndarray, 1, "wresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 66; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bresult), __pyx_ptype_5numpy_ndarray, 1, "bresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 67; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mresult), __pyx_ptype_5numpy_ndarray, 1, "mresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 68; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qresult), __pyx_ptype_5numpy_ndarray, 1, "qresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_used), __pyx_ptype_5numpy_ndarray, 1, "used", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 70; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_2bin_profile2d(__pyx_self, __pyx_v_bins_x, __pyx_v_bins_y, __pyx_v_wsource, __pyx_v_bsource, __pyx_v_wresult, __pyx_v_bresult, __pyx_v_mresult, __pyx_v_qresult, __pyx_v_used);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":62
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile2d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=1] bins_y,
 *                   np.ndarray[np.float64_t, ndim=1] wsource,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_2bin_profile2d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_bins_x, PyArrayObject *__pyx_v_bins_y, PyArrayObject *__pyx_v_wsource, PyArrayObject *__pyx_v_bsource, PyArrayObject *__pyx_v_wresult, PyArrayObject *__pyx_v_bresult, PyArrayObject *__pyx_v_mresult, PyArrayObject *__pyx_v_qresult, PyArrayObject *__pyx_v_used) {
  int __pyx_v_n;
  int __pyx_v_bini;
  int __pyx_v_binj;
  __pyx_t_5numpy_float64_t __pyx_v_wval;
  __pyx_t_5numpy_float64_t __pyx_v_bval;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bins_x;
  __Pyx_Buffer __pyx_pybuffer_bins_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bins_y;
  __Pyx_Buffer __pyx_pybuffer_bins_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bresult;
  __Pyx_Buffer __pyx_pybuffer_bresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bsource;
  __Pyx_Buffer __pyx_pybuffer_bsource;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mresult;
  __Pyx_Buffer __pyx_pybuffer_mresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_qresult;
  __Pyx_Buffer __pyx_pybuffer_qresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_used;
  __Pyx_Buffer __pyx_pybuffer_used;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wresult;
  __Pyx_Buffer __pyx_pybuffer_wresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wsource;
  __Pyx_Buffer __pyx_pybuffer_wsource;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  npy_intp __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  int __pyx_t_23;
  int __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin_profile2d", 0);
  __pyx_pybuffer_bins_x.pybuffer.buf = NULL;
  __pyx_pybuffer_bins_x.refcount = 0;
  __pyx_pybuffernd_bins_x.data = NULL;
  __pyx_pybuffernd_bins_x.rcbuffer = &__pyx_pybuffer_bins_x;
  __pyx_pybuffer_bins_y.pybuffer.buf = NULL;
  __pyx_pybuffer_bins_y.refcount = 0;
  __pyx_pybuffernd_bins_y.data = NULL;
  __pyx_pybuffernd_bins_y.rcbuffer = &__pyx_pybuffer_bins_y;
  __pyx_pybuffer_wsource.pybuffer.buf = NULL;
  __pyx_pybuffer_wsource.refcount = 0;
  __pyx_pybuffernd_wsource.data = NULL;
  __pyx_pybuffernd_wsource.rcbuffer = &__pyx_pybuffer_wsource;
  __pyx_pybuffer_bsource.pybuffer.buf = NULL;
  __pyx_pybuffer_bsource.refcount = 0;
  __pyx_pybuffernd_bsource.data = NULL;
  __pyx_pybuffernd_bsource.rcbuffer = &__pyx_pybuffer_bsource;
  __pyx_pybuffer_wresult.pybuffer.buf = NULL;
  __pyx_pybuffer_wresult.refcount = 0;
  __pyx_pybuffernd_wresult.data = NULL;
  __pyx_pybuffernd_wresult.rcbuffer = &__pyx_pybuffer_wresult;
  __pyx_pybuffer_bresult.pybuffer.buf = NULL;
  __pyx_pybuffer_bresult.refcount = 0;
  __pyx_pybuffernd_bresult.data = NULL;
  __pyx_pybuffernd_bresult.rcbuffer = &__pyx_pybuffer_bresult;
  __pyx_pybuffer_mresult.pybuffer.buf = NULL;
  __pyx_pybuffer_mresult.refcount = 0;
  __pyx_pybuffernd_mresult.data = NULL;
  __pyx_pybuffernd_mresult.rcbuffer = &__pyx_pybuffer_mresult;
  __pyx_pybuffer_qresult.pybuffer.buf = NULL;
  __pyx_pybuffer_qresult.refcount = 0;
  __pyx_pybuffernd_qresult.data = NULL;
  __pyx_pybuffernd_qresult.rcbuffer = &__pyx_pybuffer_qresult;
  __pyx_pybuffer_used.pybuffer.buf = NULL;
  __pyx_pybuffer_used.refcount = 0;
  __pyx_pybuffernd_used.data = NULL;
  __pyx_pybuffernd_used.rcbuffer = &__pyx_pybuffer_used;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_bins_x, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bins_x.diminfo[0].strides = __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bins_x.diminfo[0].shape = __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bins_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_bins_y, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bins_y.diminfo[0].strides = __pyx_pybuffernd_bins_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bins_y.diminfo[0].shape = __pyx_pybuffernd_bins_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer, (PyObject*)__pyx_v_wsource, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_wsource.diminfo[0].strides = __pyx_pybuffernd_wsource.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wsource.diminfo[0].shape = __pyx_pybuffernd_wsource.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer, (PyObject*)__pyx_v_bsource, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bsource.diminfo[0].strides = __pyx_pybuffernd_bsource.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bsource.diminfo[0].shape = __pyx_pybuffernd_bsource.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_wresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_wresult.diminfo[0].strides = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wresult.diminfo[0].shape = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_wresult.diminfo[1].strides = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_wresult.diminfo[1].shape = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_bresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bresult.diminfo[0].strides = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bresult.diminfo[0].shape = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_bresult.diminfo[1].strides = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_bresult.diminfo[1].shape = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_mresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mresult.diminfo[0].strides = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mresult.diminfo[0].shape = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mresult.diminfo[1].strides = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mresult.diminfo[1].shape = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_qresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_qresult.diminfo[0].strides = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_qresult.diminfo[0].shape = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_qresult.diminfo[1].strides = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_qresult.diminfo[1].shape = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_used.rcbuffer->pybuffer, (PyObject*)__pyx_v_used, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_used.diminfo[0].strides = __pyx_pybuffernd_used.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_used.diminfo[0].shape = __pyx_pybuffernd_used.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_used.diminfo[1].strides = __pyx_pybuffernd_used.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_used.diminfo[1].shape = __pyx_pybuffernd_used.rcbuffer->pybuffer.shape[1];

  /* "yt/utilities/lib/misc_utilities.pyx":74
 *     cdef np.int64_t bin
 *     cdef np.float64_t wval, bval
 *     for n in range(bins_x.shape[0]):             # <<<<<<<<<<<<<<
 *         bini = bins_x[n]
 *         binj = bins_y[n]
 */
  __pyx_t_1 = (__pyx_v_bins_x->dimensions[0]);
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_n = __pyx_t_2;

    /* "yt/utilities/lib/misc_utilities.pyx":75
 *     cdef np.float64_t wval, bval
 *     for n in range(bins_x.shape[0]):
 *         bini = bins_x[n]             # <<<<<<<<<<<<<<
 *         binj = bins_y[n]
 *         bval = bsource[n]
 */
    __pyx_t_3 = __pyx_v_n;
    __pyx_v_bini = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_bins_x.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":76
 *     for n in range(bins_x.shape[0]):
 *         bini = bins_x[n]
 *         binj = bins_y[n]             # <<<<<<<<<<<<<<
 *         bval = bsource[n]
 *         wval = wsource[n]
 */
    __pyx_t_4 = __pyx_v_n;
    __pyx_v_binj = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_bins_y.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_bins_y.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":77
 *         bini = bins_x[n]
 *         binj = bins_y[n]
 *         bval = bsource[n]             # <<<<<<<<<<<<<<
 *         wval = wsource[n]
 *         qresult[bini, binj] += (wresult[bini, binj] * wval * (bval - mresult[bini, binj])**2) / \
 */
    __pyx_t_5 = __pyx_v_n;
    __pyx_v_bval = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bsource.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_bsource.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":78
 *         binj = bins_y[n]
 *         bval = bsource[n]
 *         wval = wsource[n]             # <<<<<<<<<<<<<<
 *         qresult[bini, binj] += (wresult[bini, binj] * wval * (bval - mresult[bini, binj])**2) / \
 *             (wresult[bini, binj] + wval)
 */
    __pyx_t_6 = __pyx_v_n;
    __pyx_v_wval = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wsource.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_wsource.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":79
 *         bval = bsource[n]
 *         wval = wsource[n]
 *         qresult[bini, binj] += (wresult[bini, binj] * wval * (bval - mresult[bini, binj])**2) / \             # <<<<<<<<<<<<<<
 *             (wresult[bini, binj] + wval)
 *         wresult[bini, binj] += wval
 */
    __pyx_t_7 = __pyx_v_bini;
    __pyx_t_8 = __pyx_v_binj;
    __pyx_t_9 = __pyx_v_bini;
    __pyx_t_10 = __pyx_v_binj;

    /* "yt/utilities/lib/misc_utilities.pyx":80
 *         wval = wsource[n]
 *         qresult[bini, binj] += (wresult[bini, binj] * wval * (bval - mresult[bini, binj])**2) / \
 *             (wresult[bini, binj] + wval)             # <<<<<<<<<<<<<<
 *         wresult[bini, binj] += wval
 *         bresult[bini, binj] += wval*bval
 */
    __pyx_t_11 = __pyx_v_bini;
    __pyx_t_12 = __pyx_v_binj;

    /* "yt/utilities/lib/misc_utilities.pyx":79
 *         bval = bsource[n]
 *         wval = wsource[n]
 *         qresult[bini, binj] += (wresult[bini, binj] * wval * (bval - mresult[bini, binj])**2) / \             # <<<<<<<<<<<<<<
 *             (wresult[bini, binj] + wval)
 *         wresult[bini, binj] += wval
 */
    __pyx_t_13 = __pyx_v_bini;
    __pyx_t_14 = __pyx_v_binj;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_qresult.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_qresult.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_qresult.diminfo[1].strides) += ((((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_8, __pyx_pybuffernd_wresult.diminfo[1].strides)) * __pyx_v_wval) * pow((__pyx_v_bval - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_mresult.diminfo[0].strides, __pyx_t_10, __pyx_pybuffernd_mresult.diminfo[1].strides))), 2.0)) / ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_12, __pyx_pybuffernd_wresult.diminfo[1].strides)) + __pyx_v_wval));

    /* "yt/utilities/lib/misc_utilities.pyx":81
 *         qresult[bini, binj] += (wresult[bini, binj] * wval * (bval - mresult[bini, binj])**2) / \
 *             (wresult[bini, binj] + wval)
 *         wresult[bini, binj] += wval             # <<<<<<<<<<<<<<
 *         bresult[bini, binj] += wval*bval
 *         mresult[bini, binj] += wval * (bval - mresult[bini, binj]) / wresult[bini, binj]
 */
    __pyx_t_15 = __pyx_v_bini;
    __pyx_t_16 = __pyx_v_binj;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_wresult.diminfo[1].strides) += __pyx_v_wval;

    /* "yt/utilities/lib/misc_utilities.pyx":82
 *             (wresult[bini, binj] + wval)
 *         wresult[bini, binj] += wval
 *         bresult[bini, binj] += wval*bval             # <<<<<<<<<<<<<<
 *         mresult[bini, binj] += wval * (bval - mresult[bini, binj]) / wresult[bini, binj]
 *         used[bini, binj] = 1
 */
    __pyx_t_17 = __pyx_v_bini;
    __pyx_t_18 = __pyx_v_binj;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bresult.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_bresult.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_bresult.diminfo[1].strides) += (__pyx_v_wval * __pyx_v_bval);

    /* "yt/utilities/lib/misc_utilities.pyx":83
 *         wresult[bini, binj] += wval
 *         bresult[bini, binj] += wval*bval
 *         mresult[bini, binj] += wval * (bval - mresult[bini, binj]) / wresult[bini, binj]             # <<<<<<<<<<<<<<
 *         used[bini, binj] = 1
 *     return
 */
    __pyx_t_19 = __pyx_v_bini;
    __pyx_t_20 = __pyx_v_binj;
    __pyx_t_21 = __pyx_v_bini;
    __pyx_t_22 = __pyx_v_binj;
    __pyx_t_23 = __pyx_v_bini;
    __pyx_t_24 = __pyx_v_binj;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_mresult.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_mresult.diminfo[1].strides) += ((__pyx_v_wval * (__pyx_v_bval - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_mresult.diminfo[0].strides, __pyx_t_20, __pyx_pybuffernd_mresult.diminfo[1].strides)))) / (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_wresult.diminfo[1].strides)));

    /* "yt/utilities/lib/misc_utilities.pyx":84
 *         bresult[bini, binj] += wval*bval
 *         mresult[bini, binj] += wval * (bval - mresult[bini, binj]) / wresult[bini, binj]
 *         used[bini, binj] = 1             # <<<<<<<<<<<<<<
 *     return
 * 
 */
    __pyx_t_25 = __pyx_v_bini;
    __pyx_t_26 = __pyx_v_binj;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_used.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_used.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_used.diminfo[1].strides) = 1.0;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":85
 *         mresult[bini, binj] += wval * (bval - mresult[bini, binj]) / wresult[bini, binj]
 *         used[bini, binj] = 1
 *     return             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_used.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.bin_profile2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_used.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_5bin_profile3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_5bin_profile3d = {__Pyx_NAMESTR("bin_profile3d"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_5bin_profile3d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_5bin_profile3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_bins_x = 0;
  PyArrayObject *__pyx_v_bins_y = 0;
  PyArrayObject *__pyx_v_bins_z = 0;
  PyArrayObject *__pyx_v_wsource = 0;
  PyArrayObject *__pyx_v_bsource = 0;
  PyArrayObject *__pyx_v_wresult = 0;
  PyArrayObject *__pyx_v_bresult = 0;
  PyArrayObject *__pyx_v_mresult = 0;
  PyArrayObject *__pyx_v_qresult = 0;
  PyArrayObject *__pyx_v_used = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin_profile3d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__bins_x,&__pyx_n_s__bins_y,&__pyx_n_s__bins_z,&__pyx_n_s__wsource,&__pyx_n_s__bsource,&__pyx_n_s__wresult,&__pyx_n_s__bresult,&__pyx_n_s__mresult,&__pyx_n_s__qresult,&__pyx_n_s__used,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bins_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bins_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bins_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__wsource)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bsource)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__wresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__qresult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__used)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, 9); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bin_profile3d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_bins_x = ((PyArrayObject *)values[0]);
    __pyx_v_bins_y = ((PyArrayObject *)values[1]);
    __pyx_v_bins_z = ((PyArrayObject *)values[2]);
    __pyx_v_wsource = ((PyArrayObject *)values[3]);
    __pyx_v_bsource = ((PyArrayObject *)values[4]);
    __pyx_v_wresult = ((PyArrayObject *)values[5]);
    __pyx_v_bresult = ((PyArrayObject *)values[6]);
    __pyx_v_mresult = ((PyArrayObject *)values[7]);
    __pyx_v_qresult = ((PyArrayObject *)values[8]);
    __pyx_v_used = ((PyArrayObject *)values[9]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bin_profile3d", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.bin_profile3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bins_x), __pyx_ptype_5numpy_ndarray, 1, "bins_x", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bins_y), __pyx_ptype_5numpy_ndarray, 1, "bins_y", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bins_z), __pyx_ptype_5numpy_ndarray, 1, "bins_z", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wsource), __pyx_ptype_5numpy_ndarray, 1, "wsource", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 93; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bsource), __pyx_ptype_5numpy_ndarray, 1, "bsource", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wresult), __pyx_ptype_5numpy_ndarray, 1, "wresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bresult), __pyx_ptype_5numpy_ndarray, 1, "bresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mresult), __pyx_ptype_5numpy_ndarray, 1, "mresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qresult), __pyx_ptype_5numpy_ndarray, 1, "qresult", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 98; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_used), __pyx_ptype_5numpy_ndarray, 1, "used", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 99; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_4bin_profile3d(__pyx_self, __pyx_v_bins_x, __pyx_v_bins_y, __pyx_v_bins_z, __pyx_v_wsource, __pyx_v_bsource, __pyx_v_wresult, __pyx_v_bresult, __pyx_v_mresult, __pyx_v_qresult, __pyx_v_used);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":90
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile3d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=1] bins_y,
 *                   np.ndarray[np.int64_t, ndim=1] bins_z,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_4bin_profile3d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_bins_x, PyArrayObject *__pyx_v_bins_y, PyArrayObject *__pyx_v_bins_z, PyArrayObject *__pyx_v_wsource, PyArrayObject *__pyx_v_bsource, PyArrayObject *__pyx_v_wresult, PyArrayObject *__pyx_v_bresult, PyArrayObject *__pyx_v_mresult, PyArrayObject *__pyx_v_qresult, PyArrayObject *__pyx_v_used) {
  int __pyx_v_n;
  int __pyx_v_bini;
  int __pyx_v_binj;
  int __pyx_v_bink;
  __pyx_t_5numpy_float64_t __pyx_v_wval;
  __pyx_t_5numpy_float64_t __pyx_v_bval;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bins_x;
  __Pyx_Buffer __pyx_pybuffer_bins_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bins_y;
  __Pyx_Buffer __pyx_pybuffer_bins_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bins_z;
  __Pyx_Buffer __pyx_pybuffer_bins_z;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bresult;
  __Pyx_Buffer __pyx_pybuffer_bresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bsource;
  __Pyx_Buffer __pyx_pybuffer_bsource;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mresult;
  __Pyx_Buffer __pyx_pybuffer_mresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_qresult;
  __Pyx_Buffer __pyx_pybuffer_qresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_used;
  __Pyx_Buffer __pyx_pybuffer_used;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wresult;
  __Pyx_Buffer __pyx_pybuffer_wresult;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wsource;
  __Pyx_Buffer __pyx_pybuffer_wsource;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  npy_intp __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  int __pyx_t_23;
  int __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  int __pyx_t_32;
  int __pyx_t_33;
  int __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin_profile3d", 0);
  __pyx_pybuffer_bins_x.pybuffer.buf = NULL;
  __pyx_pybuffer_bins_x.refcount = 0;
  __pyx_pybuffernd_bins_x.data = NULL;
  __pyx_pybuffernd_bins_x.rcbuffer = &__pyx_pybuffer_bins_x;
  __pyx_pybuffer_bins_y.pybuffer.buf = NULL;
  __pyx_pybuffer_bins_y.refcount = 0;
  __pyx_pybuffernd_bins_y.data = NULL;
  __pyx_pybuffernd_bins_y.rcbuffer = &__pyx_pybuffer_bins_y;
  __pyx_pybuffer_bins_z.pybuffer.buf = NULL;
  __pyx_pybuffer_bins_z.refcount = 0;
  __pyx_pybuffernd_bins_z.data = NULL;
  __pyx_pybuffernd_bins_z.rcbuffer = &__pyx_pybuffer_bins_z;
  __pyx_pybuffer_wsource.pybuffer.buf = NULL;
  __pyx_pybuffer_wsource.refcount = 0;
  __pyx_pybuffernd_wsource.data = NULL;
  __pyx_pybuffernd_wsource.rcbuffer = &__pyx_pybuffer_wsource;
  __pyx_pybuffer_bsource.pybuffer.buf = NULL;
  __pyx_pybuffer_bsource.refcount = 0;
  __pyx_pybuffernd_bsource.data = NULL;
  __pyx_pybuffernd_bsource.rcbuffer = &__pyx_pybuffer_bsource;
  __pyx_pybuffer_wresult.pybuffer.buf = NULL;
  __pyx_pybuffer_wresult.refcount = 0;
  __pyx_pybuffernd_wresult.data = NULL;
  __pyx_pybuffernd_wresult.rcbuffer = &__pyx_pybuffer_wresult;
  __pyx_pybuffer_bresult.pybuffer.buf = NULL;
  __pyx_pybuffer_bresult.refcount = 0;
  __pyx_pybuffernd_bresult.data = NULL;
  __pyx_pybuffernd_bresult.rcbuffer = &__pyx_pybuffer_bresult;
  __pyx_pybuffer_mresult.pybuffer.buf = NULL;
  __pyx_pybuffer_mresult.refcount = 0;
  __pyx_pybuffernd_mresult.data = NULL;
  __pyx_pybuffernd_mresult.rcbuffer = &__pyx_pybuffer_mresult;
  __pyx_pybuffer_qresult.pybuffer.buf = NULL;
  __pyx_pybuffer_qresult.refcount = 0;
  __pyx_pybuffernd_qresult.data = NULL;
  __pyx_pybuffernd_qresult.rcbuffer = &__pyx_pybuffer_qresult;
  __pyx_pybuffer_used.pybuffer.buf = NULL;
  __pyx_pybuffer_used.refcount = 0;
  __pyx_pybuffernd_used.data = NULL;
  __pyx_pybuffernd_used.rcbuffer = &__pyx_pybuffer_used;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_bins_x, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bins_x.diminfo[0].strides = __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bins_x.diminfo[0].shape = __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bins_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_bins_y, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bins_y.diminfo[0].strides = __pyx_pybuffernd_bins_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bins_y.diminfo[0].shape = __pyx_pybuffernd_bins_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bins_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_bins_z, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bins_z.diminfo[0].strides = __pyx_pybuffernd_bins_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bins_z.diminfo[0].shape = __pyx_pybuffernd_bins_z.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer, (PyObject*)__pyx_v_wsource, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_wsource.diminfo[0].strides = __pyx_pybuffernd_wsource.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wsource.diminfo[0].shape = __pyx_pybuffernd_wsource.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer, (PyObject*)__pyx_v_bsource, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bsource.diminfo[0].strides = __pyx_pybuffernd_bsource.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bsource.diminfo[0].shape = __pyx_pybuffernd_bsource.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_wresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_wresult.diminfo[0].strides = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wresult.diminfo[0].shape = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_wresult.diminfo[1].strides = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_wresult.diminfo[1].shape = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_wresult.diminfo[2].strides = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_wresult.diminfo[2].shape = __pyx_pybuffernd_wresult.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_bresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_bresult.diminfo[0].strides = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bresult.diminfo[0].shape = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_bresult.diminfo[1].strides = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_bresult.diminfo[1].shape = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_bresult.diminfo[2].strides = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_bresult.diminfo[2].shape = __pyx_pybuffernd_bresult.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_mresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mresult.diminfo[0].strides = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mresult.diminfo[0].shape = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mresult.diminfo[1].strides = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mresult.diminfo[1].shape = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mresult.diminfo[2].strides = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mresult.diminfo[2].shape = __pyx_pybuffernd_mresult.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer, (PyObject*)__pyx_v_qresult, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_qresult.diminfo[0].strides = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_qresult.diminfo[0].shape = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_qresult.diminfo[1].strides = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_qresult.diminfo[1].shape = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_qresult.diminfo[2].strides = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_qresult.diminfo[2].shape = __pyx_pybuffernd_qresult.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_used.rcbuffer->pybuffer, (PyObject*)__pyx_v_used, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_used.diminfo[0].strides = __pyx_pybuffernd_used.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_used.diminfo[0].shape = __pyx_pybuffernd_used.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_used.diminfo[1].strides = __pyx_pybuffernd_used.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_used.diminfo[1].shape = __pyx_pybuffernd_used.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_used.diminfo[2].strides = __pyx_pybuffernd_used.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_used.diminfo[2].shape = __pyx_pybuffernd_used.rcbuffer->pybuffer.shape[2];

  /* "yt/utilities/lib/misc_utilities.pyx":103
 *     cdef np.int64_t bin
 *     cdef np.float64_t wval, bval
 *     for n in range(bins_x.shape[0]):             # <<<<<<<<<<<<<<
 *         bini = bins_x[n]
 *         binj = bins_y[n]
 */
  __pyx_t_1 = (__pyx_v_bins_x->dimensions[0]);
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_n = __pyx_t_2;

    /* "yt/utilities/lib/misc_utilities.pyx":104
 *     cdef np.float64_t wval, bval
 *     for n in range(bins_x.shape[0]):
 *         bini = bins_x[n]             # <<<<<<<<<<<<<<
 *         binj = bins_y[n]
 *         bink = bins_z[n]
 */
    __pyx_t_3 = __pyx_v_n;
    __pyx_v_bini = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_bins_x.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_bins_x.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":105
 *     for n in range(bins_x.shape[0]):
 *         bini = bins_x[n]
 *         binj = bins_y[n]             # <<<<<<<<<<<<<<
 *         bink = bins_z[n]
 *         bval = bsource[n]
 */
    __pyx_t_4 = __pyx_v_n;
    __pyx_v_binj = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_bins_y.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_bins_y.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":106
 *         bini = bins_x[n]
 *         binj = bins_y[n]
 *         bink = bins_z[n]             # <<<<<<<<<<<<<<
 *         bval = bsource[n]
 *         wval = wsource[n]
 */
    __pyx_t_5 = __pyx_v_n;
    __pyx_v_bink = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_bins_z.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_bins_z.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":107
 *         binj = bins_y[n]
 *         bink = bins_z[n]
 *         bval = bsource[n]             # <<<<<<<<<<<<<<
 *         wval = wsource[n]
 *         qresult[bini, binj, bink] += (wresult[bini, binj, bink] * wval * (bval - mresult[bini, binj, bink])**2) / \
 */
    __pyx_t_6 = __pyx_v_n;
    __pyx_v_bval = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bsource.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_bsource.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":108
 *         bink = bins_z[n]
 *         bval = bsource[n]
 *         wval = wsource[n]             # <<<<<<<<<<<<<<
 *         qresult[bini, binj, bink] += (wresult[bini, binj, bink] * wval * (bval - mresult[bini, binj, bink])**2) / \
 *             (wresult[bini, binj, bink] + wval)
 */
    __pyx_t_7 = __pyx_v_n;
    __pyx_v_wval = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wsource.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_wsource.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":109
 *         bval = bsource[n]
 *         wval = wsource[n]
 *         qresult[bini, binj, bink] += (wresult[bini, binj, bink] * wval * (bval - mresult[bini, binj, bink])**2) / \             # <<<<<<<<<<<<<<
 *             (wresult[bini, binj, bink] + wval)
 *         wresult[bini, binj, bink] += wval
 */
    __pyx_t_8 = __pyx_v_bini;
    __pyx_t_9 = __pyx_v_binj;
    __pyx_t_10 = __pyx_v_bink;
    __pyx_t_11 = __pyx_v_bini;
    __pyx_t_12 = __pyx_v_binj;
    __pyx_t_13 = __pyx_v_bink;

    /* "yt/utilities/lib/misc_utilities.pyx":110
 *         wval = wsource[n]
 *         qresult[bini, binj, bink] += (wresult[bini, binj, bink] * wval * (bval - mresult[bini, binj, bink])**2) / \
 *             (wresult[bini, binj, bink] + wval)             # <<<<<<<<<<<<<<
 *         wresult[bini, binj, bink] += wval
 *         bresult[bini, binj, bink] += wval*bval
 */
    __pyx_t_14 = __pyx_v_bini;
    __pyx_t_15 = __pyx_v_binj;
    __pyx_t_16 = __pyx_v_bink;

    /* "yt/utilities/lib/misc_utilities.pyx":109
 *         bval = bsource[n]
 *         wval = wsource[n]
 *         qresult[bini, binj, bink] += (wresult[bini, binj, bink] * wval * (bval - mresult[bini, binj, bink])**2) / \             # <<<<<<<<<<<<<<
 *             (wresult[bini, binj, bink] + wval)
 *         wresult[bini, binj, bink] += wval
 */
    __pyx_t_17 = __pyx_v_bini;
    __pyx_t_18 = __pyx_v_binj;
    __pyx_t_19 = __pyx_v_bink;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_qresult.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_qresult.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_qresult.diminfo[1].strides, __pyx_t_19, __pyx_pybuffernd_qresult.diminfo[2].strides) += ((((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_9, __pyx_pybuffernd_wresult.diminfo[1].strides, __pyx_t_10, __pyx_pybuffernd_wresult.diminfo[2].strides)) * __pyx_v_wval) * pow((__pyx_v_bval - (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_mresult.diminfo[0].strides, __pyx_t_12, __pyx_pybuffernd_mresult.diminfo[1].strides, __pyx_t_13, __pyx_pybuffernd_mresult.diminfo[2].strides))), 2.0)) / ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_wresult.diminfo[1].strides, __pyx_t_16, __pyx_pybuffernd_wresult.diminfo[2].strides)) + __pyx_v_wval));

    /* "yt/utilities/lib/misc_utilities.pyx":111
 *         qresult[bini, binj, bink] += (wresult[bini, binj, bink] * wval * (bval - mresult[bini, binj, bink])**2) / \
 *             (wresult[bini, binj, bink] + wval)
 *         wresult[bini, binj, bink] += wval             # <<<<<<<<<<<<<<
 *         bresult[bini, binj, bink] += wval*bval
 *         mresult[bini, binj, bink] += wval * (bval - mresult[bini, binj, bink]) / wresult[bini, binj, bink]
 */
    __pyx_t_20 = __pyx_v_bini;
    __pyx_t_21 = __pyx_v_binj;
    __pyx_t_22 = __pyx_v_bink;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_wresult.diminfo[1].strides, __pyx_t_22, __pyx_pybuffernd_wresult.diminfo[2].strides) += __pyx_v_wval;

    /* "yt/utilities/lib/misc_utilities.pyx":112
 *             (wresult[bini, binj, bink] + wval)
 *         wresult[bini, binj, bink] += wval
 *         bresult[bini, binj, bink] += wval*bval             # <<<<<<<<<<<<<<
 *         mresult[bini, binj, bink] += wval * (bval - mresult[bini, binj, bink]) / wresult[bini, binj, bink]
 *         used[bini, binj, bink] = 1
 */
    __pyx_t_23 = __pyx_v_bini;
    __pyx_t_24 = __pyx_v_binj;
    __pyx_t_25 = __pyx_v_bink;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bresult.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_bresult.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_bresult.diminfo[1].strides, __pyx_t_25, __pyx_pybuffernd_bresult.diminfo[2].strides) += (__pyx_v_wval * __pyx_v_bval);

    /* "yt/utilities/lib/misc_utilities.pyx":113
 *         wresult[bini, binj, bink] += wval
 *         bresult[bini, binj, bink] += wval*bval
 *         mresult[bini, binj, bink] += wval * (bval - mresult[bini, binj, bink]) / wresult[bini, binj, bink]             # <<<<<<<<<<<<<<
 *         used[bini, binj, bink] = 1
 *     return
 */
    __pyx_t_26 = __pyx_v_bini;
    __pyx_t_27 = __pyx_v_binj;
    __pyx_t_28 = __pyx_v_bink;
    __pyx_t_29 = __pyx_v_bini;
    __pyx_t_30 = __pyx_v_binj;
    __pyx_t_31 = __pyx_v_bink;
    __pyx_t_32 = __pyx_v_bini;
    __pyx_t_33 = __pyx_v_binj;
    __pyx_t_34 = __pyx_v_bink;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_mresult.diminfo[0].strides, __pyx_t_33, __pyx_pybuffernd_mresult.diminfo[1].strides, __pyx_t_34, __pyx_pybuffernd_mresult.diminfo[2].strides) += ((__pyx_v_wval * (__pyx_v_bval - (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mresult.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_mresult.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_mresult.diminfo[1].strides, __pyx_t_28, __pyx_pybuffernd_mresult.diminfo[2].strides)))) / (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_wresult.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_wresult.diminfo[0].strides, __pyx_t_30, __pyx_pybuffernd_wresult.diminfo[1].strides, __pyx_t_31, __pyx_pybuffernd_wresult.diminfo[2].strides)));

    /* "yt/utilities/lib/misc_utilities.pyx":114
 *         bresult[bini, binj, bink] += wval*bval
 *         mresult[bini, binj, bink] += wval * (bval - mresult[bini, binj, bink]) / wresult[bini, binj, bink]
 *         used[bini, binj, bink] = 1             # <<<<<<<<<<<<<<
 *     return
 * 
 */
    __pyx_t_35 = __pyx_v_bini;
    __pyx_t_36 = __pyx_v_binj;
    __pyx_t_37 = __pyx_v_bink;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_used.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_used.diminfo[0].strides, __pyx_t_36, __pyx_pybuffernd_used.diminfo[1].strides, __pyx_t_37, __pyx_pybuffernd_used.diminfo[2].strides) = 1.0;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":115
 *         mresult[bini, binj, bink] += wval * (bval - mresult[bini, binj, bink]) / wresult[bini, binj, bink]
 *         used[bini, binj, bink] = 1
 *     return             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_z.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_used.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.bin_profile3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bins_z.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bsource.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_qresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_used.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wresult.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wsource.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_7lines(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_7lines = {__Pyx_NAMESTR("lines"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_7lines, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_7lines(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_image = 0;
  PyArrayObject *__pyx_v_xs = 0;
  PyArrayObject *__pyx_v_ys = 0;
  PyArrayObject *__pyx_v_colors = 0;
  int __pyx_v_points_per_color;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lines (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__image,&__pyx_n_s__xs,&__pyx_n_s__ys,&__pyx_n_s__colors,&__pyx_n_s__points_per_color,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__xs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lines", 0, 4, 5, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__ys)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lines", 0, 4, 5, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__colors)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lines", 0, 4, 5, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__points_per_color);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lines") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = ((PyArrayObject *)values[0]);
    __pyx_v_xs = ((PyArrayObject *)values[1]);
    __pyx_v_ys = ((PyArrayObject *)values[2]);
    __pyx_v_colors = ((PyArrayObject *)values[3]);
    if (values[4]) {
      __pyx_v_points_per_color = __Pyx_PyInt_AsInt(values[4]); if (unlikely((__pyx_v_points_per_color == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 124; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {
      __pyx_v_points_per_color = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lines", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_image), __pyx_ptype_5numpy_ndarray, 1, "image", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_xs), __pyx_ptype_5numpy_ndarray, 1, "xs", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 121; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ys), __pyx_ptype_5numpy_ndarray, 1, "ys", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 122; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_colors), __pyx_ptype_5numpy_ndarray, 1, "colors", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 123; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_6lines(__pyx_self, __pyx_v_image, __pyx_v_xs, __pyx_v_ys, __pyx_v_colors, __pyx_v_points_per_color);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":120
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def lines(np.ndarray[np.float64_t, ndim=3] image,             # <<<<<<<<<<<<<<
 *           np.ndarray[np.int64_t, ndim=1] xs,
 *           np.ndarray[np.int64_t, ndim=1] ys,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_6lines(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_image, PyArrayObject *__pyx_v_xs, PyArrayObject *__pyx_v_ys, PyArrayObject *__pyx_v_colors, int __pyx_v_points_per_color) {
  int __pyx_v_nx;
  int __pyx_v_ny;
  int __pyx_v_nl;
  __pyx_t_5numpy_float64_t __pyx_v_alpha[4];
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_dx;
  int __pyx_v_dy;
  int __pyx_v_sx;
  int __pyx_v_sy;
  int __pyx_v_e2;
  int __pyx_v_err;
  __pyx_t_5numpy_int64_t __pyx_v_x0;
  __pyx_t_5numpy_int64_t __pyx_v_x1;
  __pyx_t_5numpy_int64_t __pyx_v_y0;
  __pyx_t_5numpy_int64_t __pyx_v_y1;
  int __pyx_v_has_alpha;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_colors;
  __Pyx_Buffer __pyx_pybuffer_colors;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_image;
  __Pyx_Buffer __pyx_pybuffer_image;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_xs;
  __Pyx_Buffer __pyx_pybuffer_xs;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_ys;
  __Pyx_Buffer __pyx_pybuffer_ys;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  long __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  long __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  __pyx_t_5numpy_int64_t __pyx_t_21;
  __pyx_t_5numpy_int64_t __pyx_t_22;
  int __pyx_t_23;
  __pyx_t_5numpy_int64_t __pyx_t_24;
  __pyx_t_5numpy_int64_t __pyx_t_25;
  int __pyx_t_26;
  __pyx_t_5numpy_int64_t __pyx_t_27;
  __pyx_t_5numpy_int64_t __pyx_t_28;
  int __pyx_t_29;
  __pyx_t_5numpy_int64_t __pyx_t_30;
  __pyx_t_5numpy_int64_t __pyx_t_31;
  int __pyx_t_32;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lines", 0);
  __pyx_pybuffer_image.pybuffer.buf = NULL;
  __pyx_pybuffer_image.refcount = 0;
  __pyx_pybuffernd_image.data = NULL;
  __pyx_pybuffernd_image.rcbuffer = &__pyx_pybuffer_image;
  __pyx_pybuffer_xs.pybuffer.buf = NULL;
  __pyx_pybuffer_xs.refcount = 0;
  __pyx_pybuffernd_xs.data = NULL;
  __pyx_pybuffernd_xs.rcbuffer = &__pyx_pybuffer_xs;
  __pyx_pybuffer_ys.pybuffer.buf = NULL;
  __pyx_pybuffer_ys.refcount = 0;
  __pyx_pybuffernd_ys.data = NULL;
  __pyx_pybuffernd_ys.rcbuffer = &__pyx_pybuffer_ys;
  __pyx_pybuffer_colors.pybuffer.buf = NULL;
  __pyx_pybuffer_colors.refcount = 0;
  __pyx_pybuffernd_colors.data = NULL;
  __pyx_pybuffernd_colors.rcbuffer = &__pyx_pybuffer_colors;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_image.rcbuffer->pybuffer, (PyObject*)__pyx_v_image, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_image.diminfo[0].strides = __pyx_pybuffernd_image.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_image.diminfo[0].shape = __pyx_pybuffernd_image.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_image.diminfo[1].strides = __pyx_pybuffernd_image.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_image.diminfo[1].shape = __pyx_pybuffernd_image.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_image.diminfo[2].strides = __pyx_pybuffernd_image.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_image.diminfo[2].shape = __pyx_pybuffernd_image.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_xs.rcbuffer->pybuffer, (PyObject*)__pyx_v_xs, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_xs.diminfo[0].strides = __pyx_pybuffernd_xs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_xs.diminfo[0].shape = __pyx_pybuffernd_xs.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_ys.rcbuffer->pybuffer, (PyObject*)__pyx_v_ys, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_ys.diminfo[0].strides = __pyx_pybuffernd_ys.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_ys.diminfo[0].shape = __pyx_pybuffernd_ys.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_colors.rcbuffer->pybuffer, (PyObject*)__pyx_v_colors, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_colors.diminfo[0].strides = __pyx_pybuffernd_colors.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_colors.diminfo[0].shape = __pyx_pybuffernd_colors.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_colors.diminfo[1].strides = __pyx_pybuffernd_colors.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_colors.diminfo[1].shape = __pyx_pybuffernd_colors.rcbuffer->pybuffer.shape[1];

  /* "yt/utilities/lib/misc_utilities.pyx":126
 *           int points_per_color=1):
 * 
 *     cdef int nx = image.shape[0]             # <<<<<<<<<<<<<<
 *     cdef int ny = image.shape[1]
 *     cdef int nl = xs.shape[0]
 */
  __pyx_v_nx = (__pyx_v_image->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":127
 * 
 *     cdef int nx = image.shape[0]
 *     cdef int ny = image.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int nl = xs.shape[0]
 *     cdef np.float64_t alpha[4]
 */
  __pyx_v_ny = (__pyx_v_image->dimensions[1]);

  /* "yt/utilities/lib/misc_utilities.pyx":128
 *     cdef int nx = image.shape[0]
 *     cdef int ny = image.shape[1]
 *     cdef int nl = xs.shape[0]             # <<<<<<<<<<<<<<
 *     cdef np.float64_t alpha[4]
 *     cdef int i, j
 */
  __pyx_v_nl = (__pyx_v_xs->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":133
 *     cdef int dx, dy, sx, sy, e2, err
 *     cdef np.int64_t x0, x1, y0, y1
 *     cdef int has_alpha = (image.shape[-1] == 4)             # <<<<<<<<<<<<<<
 *     for j in range(0, nl, 2):
 *         # From wikipedia http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
 */
  __pyx_v_has_alpha = ((__pyx_v_image->dimensions[-1]) == 4);

  /* "yt/utilities/lib/misc_utilities.pyx":134
 *     cdef np.int64_t x0, x1, y0, y1
 *     cdef int has_alpha = (image.shape[-1] == 4)
 *     for j in range(0, nl, 2):             # <<<<<<<<<<<<<<
 *         # From wikipedia http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
 *         x0 = xs[j]; y0 = ys[j]; x1 = xs[j+1]; y1 = ys[j+1]
 */
  __pyx_t_1 = __pyx_v_nl;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=2) {
    __pyx_v_j = __pyx_t_2;

    /* "yt/utilities/lib/misc_utilities.pyx":136
 *     for j in range(0, nl, 2):
 *         # From wikipedia http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
 *         x0 = xs[j]; y0 = ys[j]; x1 = xs[j+1]; y1 = ys[j+1]             # <<<<<<<<<<<<<<
 *         dx = abs(x1-x0)
 *         dy = abs(y1-y0)
 */
    __pyx_t_3 = __pyx_v_j;
    __pyx_v_x0 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_xs.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_xs.diminfo[0].strides));
    __pyx_t_4 = __pyx_v_j;
    __pyx_v_y0 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_ys.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_ys.diminfo[0].strides));
    __pyx_t_5 = (__pyx_v_j + 1);
    __pyx_v_x1 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_xs.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_xs.diminfo[0].strides));
    __pyx_t_6 = (__pyx_v_j + 1);
    __pyx_v_y1 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_ys.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_ys.diminfo[0].strides));

    /* "yt/utilities/lib/misc_utilities.pyx":137
 *         # From wikipedia http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
 *         x0 = xs[j]; y0 = ys[j]; x1 = xs[j+1]; y1 = ys[j+1]
 *         dx = abs(x1-x0)             # <<<<<<<<<<<<<<
 *         dy = abs(y1-y0)
 *         err = dx - dy
 */
    __pyx_t_7 = __Pyx_PyInt_to_py_npy_int64((__pyx_v_x1 - __pyx_v_x0)); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 137; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyNumber_Absolute(__pyx_t_7); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 137; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyInt_AsInt(__pyx_t_8); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 137; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_dx = __pyx_t_9;

    /* "yt/utilities/lib/misc_utilities.pyx":138
 *         x0 = xs[j]; y0 = ys[j]; x1 = xs[j+1]; y1 = ys[j+1]
 *         dx = abs(x1-x0)
 *         dy = abs(y1-y0)             # <<<<<<<<<<<<<<
 *         err = dx - dy
 *         if has_alpha:
 */
    __pyx_t_8 = __Pyx_PyInt_to_py_npy_int64((__pyx_v_y1 - __pyx_v_y0)); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyNumber_Absolute(__pyx_t_8); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = __Pyx_PyInt_AsInt(__pyx_t_7); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_dy = __pyx_t_9;

    /* "yt/utilities/lib/misc_utilities.pyx":139
 *         dx = abs(x1-x0)
 *         dy = abs(y1-y0)
 *         err = dx - dy             # <<<<<<<<<<<<<<
 *         if has_alpha:
 *             for i in range(4):
 */
    __pyx_v_err = (__pyx_v_dx - __pyx_v_dy);

    /* "yt/utilities/lib/misc_utilities.pyx":140
 *         dy = abs(y1-y0)
 *         err = dx - dy
 *         if has_alpha:             # <<<<<<<<<<<<<<
 *             for i in range(4):
 *                 alpha[i] = colors[j/points_per_color,i]
 */
    if (__pyx_v_has_alpha) {

      /* "yt/utilities/lib/misc_utilities.pyx":141
 *         err = dx - dy
 *         if has_alpha:
 *             for i in range(4):             # <<<<<<<<<<<<<<
 *                 alpha[i] = colors[j/points_per_color,i]
 *         else:
 */
      for (__pyx_t_9 = 0; __pyx_t_9 < 4; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "yt/utilities/lib/misc_utilities.pyx":142
 *         if has_alpha:
 *             for i in range(4):
 *                 alpha[i] = colors[j/points_per_color,i]             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(3):
 */
        __pyx_t_10 = (__pyx_v_j / __pyx_v_points_per_color);
        __pyx_t_11 = __pyx_v_i;
        (__pyx_v_alpha[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_colors.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_colors.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_colors.diminfo[1].strides));
      }
      goto __pyx_L5;
    }
    /*else*/ {

      /* "yt/utilities/lib/misc_utilities.pyx":144
 *                 alpha[i] = colors[j/points_per_color,i]
 *         else:
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 alpha[i] = colors[j/points_per_color,3]*\
 *                         colors[j/points_per_color,i]
 */
      for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "yt/utilities/lib/misc_utilities.pyx":145
 *         else:
 *             for i in range(3):
 *                 alpha[i] = colors[j/points_per_color,3]*\             # <<<<<<<<<<<<<<
 *                         colors[j/points_per_color,i]
 *         if x0 < x1:
 */
        __pyx_t_12 = (__pyx_v_j / __pyx_v_points_per_color);
        __pyx_t_13 = 3;

        /* "yt/utilities/lib/misc_utilities.pyx":146
 *             for i in range(3):
 *                 alpha[i] = colors[j/points_per_color,3]*\
 *                         colors[j/points_per_color,i]             # <<<<<<<<<<<<<<
 *         if x0 < x1:
 *             sx = 1
 */
        __pyx_t_14 = (__pyx_v_j / __pyx_v_points_per_color);
        __pyx_t_15 = __pyx_v_i;

        /* "yt/utilities/lib/misc_utilities.pyx":145
 *         else:
 *             for i in range(3):
 *                 alpha[i] = colors[j/points_per_color,3]*\             # <<<<<<<<<<<<<<
 *                         colors[j/points_per_color,i]
 *         if x0 < x1:
 */
        (__pyx_v_alpha[__pyx_v_i]) = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_colors.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_colors.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_colors.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_colors.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_colors.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_colors.diminfo[1].strides)));
      }
    }
    __pyx_L5:;

    /* "yt/utilities/lib/misc_utilities.pyx":147
 *                 alpha[i] = colors[j/points_per_color,3]*\
 *                         colors[j/points_per_color,i]
 *         if x0 < x1:             # <<<<<<<<<<<<<<
 *             sx = 1
 *         else:
 */
    __pyx_t_16 = (__pyx_v_x0 < __pyx_v_x1);
    if (__pyx_t_16) {

      /* "yt/utilities/lib/misc_utilities.pyx":148
 *                         colors[j/points_per_color,i]
 *         if x0 < x1:
 *             sx = 1             # <<<<<<<<<<<<<<
 *         else:
 *             sx = -1
 */
      __pyx_v_sx = 1;
      goto __pyx_L10;
    }
    /*else*/ {

      /* "yt/utilities/lib/misc_utilities.pyx":150
 *             sx = 1
 *         else:
 *             sx = -1             # <<<<<<<<<<<<<<
 *         if y0 < y1:
 *             sy = 1
 */
      __pyx_v_sx = -1;
    }
    __pyx_L10:;

    /* "yt/utilities/lib/misc_utilities.pyx":151
 *         else:
 *             sx = -1
 *         if y0 < y1:             # <<<<<<<<<<<<<<
 *             sy = 1
 *         else:
 */
    __pyx_t_16 = (__pyx_v_y0 < __pyx_v_y1);
    if (__pyx_t_16) {

      /* "yt/utilities/lib/misc_utilities.pyx":152
 *             sx = -1
 *         if y0 < y1:
 *             sy = 1             # <<<<<<<<<<<<<<
 *         else:
 *             sy = -1
 */
      __pyx_v_sy = 1;
      goto __pyx_L11;
    }
    /*else*/ {

      /* "yt/utilities/lib/misc_utilities.pyx":154
 *             sy = 1
 *         else:
 *             sy = -1             # <<<<<<<<<<<<<<
 *         while(1):
 *             if (x0 < 0 and sx == -1): break
 */
      __pyx_v_sy = -1;
    }
    __pyx_L11:;

    /* "yt/utilities/lib/misc_utilities.pyx":155
 *         else:
 *             sy = -1
 *         while(1):             # <<<<<<<<<<<<<<
 *             if (x0 < 0 and sx == -1): break
 *             elif (x0 >= nx and sx == 1): break
 */
    while (1) {
      if (!1) break;

      /* "yt/utilities/lib/misc_utilities.pyx":156
 *             sy = -1
 *         while(1):
 *             if (x0 < 0 and sx == -1): break             # <<<<<<<<<<<<<<
 *             elif (x0 >= nx and sx == 1): break
 *             elif (y0 < 0 and sy == -1): break
 */
      __pyx_t_16 = (__pyx_v_x0 < 0);
      if (__pyx_t_16) {
        __pyx_t_17 = (__pyx_v_sx == -1);
        __pyx_t_18 = __pyx_t_17;
      } else {
        __pyx_t_18 = __pyx_t_16;
      }
      if (__pyx_t_18) {
        goto __pyx_L13_break;
        goto __pyx_L14;
      }

      /* "yt/utilities/lib/misc_utilities.pyx":157
 *         while(1):
 *             if (x0 < 0 and sx == -1): break
 *             elif (x0 >= nx and sx == 1): break             # <<<<<<<<<<<<<<
 *             elif (y0 < 0 and sy == -1): break
 *             elif (y0 >= nx and sy == 1): break
 */
      __pyx_t_18 = (__pyx_v_x0 >= __pyx_v_nx);
      if (__pyx_t_18) {
        __pyx_t_16 = (__pyx_v_sx == 1);
        __pyx_t_17 = __pyx_t_16;
      } else {
        __pyx_t_17 = __pyx_t_18;
      }
      if (__pyx_t_17) {
        goto __pyx_L13_break;
        goto __pyx_L14;
      }

      /* "yt/utilities/lib/misc_utilities.pyx":158
 *             if (x0 < 0 and sx == -1): break
 *             elif (x0 >= nx and sx == 1): break
 *             elif (y0 < 0 and sy == -1): break             # <<<<<<<<<<<<<<
 *             elif (y0 >= nx and sy == 1): break
 *             if (x0 >=0 and x0 < nx and y0 >= 0 and y0 < ny):
 */
      __pyx_t_17 = (__pyx_v_y0 < 0);
      if (__pyx_t_17) {
        __pyx_t_18 = (__pyx_v_sy == -1);
        __pyx_t_16 = __pyx_t_18;
      } else {
        __pyx_t_16 = __pyx_t_17;
      }
      if (__pyx_t_16) {
        goto __pyx_L13_break;
        goto __pyx_L14;
      }

      /* "yt/utilities/lib/misc_utilities.pyx":159
 *             elif (x0 >= nx and sx == 1): break
 *             elif (y0 < 0 and sy == -1): break
 *             elif (y0 >= nx and sy == 1): break             # <<<<<<<<<<<<<<
 *             if (x0 >=0 and x0 < nx and y0 >= 0 and y0 < ny):
 *                 if has_alpha:
 */
      __pyx_t_16 = (__pyx_v_y0 >= __pyx_v_nx);
      if (__pyx_t_16) {
        __pyx_t_17 = (__pyx_v_sy == 1);
        __pyx_t_18 = __pyx_t_17;
      } else {
        __pyx_t_18 = __pyx_t_16;
      }
      if (__pyx_t_18) {
        goto __pyx_L13_break;
        goto __pyx_L14;
      }
      __pyx_L14:;

      /* "yt/utilities/lib/misc_utilities.pyx":160
 *             elif (y0 < 0 and sy == -1): break
 *             elif (y0 >= nx and sy == 1): break
 *             if (x0 >=0 and x0 < nx and y0 >= 0 and y0 < ny):             # <<<<<<<<<<<<<<
 *                 if has_alpha:
 *                     for i in range(4):
 */
      __pyx_t_18 = (__pyx_v_x0 >= 0);
      if (__pyx_t_18) {
        __pyx_t_16 = (__pyx_v_x0 < __pyx_v_nx);
        if (__pyx_t_16) {
          __pyx_t_17 = (__pyx_v_y0 >= 0);
          if (__pyx_t_17) {
            __pyx_t_19 = (__pyx_v_y0 < __pyx_v_ny);
            __pyx_t_20 = __pyx_t_19;
          } else {
            __pyx_t_20 = __pyx_t_17;
          }
          __pyx_t_17 = __pyx_t_20;
        } else {
          __pyx_t_17 = __pyx_t_16;
        }
        __pyx_t_16 = __pyx_t_17;
      } else {
        __pyx_t_16 = __pyx_t_18;
      }
      if (__pyx_t_16) {

        /* "yt/utilities/lib/misc_utilities.pyx":161
 *             elif (y0 >= nx and sy == 1): break
 *             if (x0 >=0 and x0 < nx and y0 >= 0 and y0 < ny):
 *                 if has_alpha:             # <<<<<<<<<<<<<<
 *                     for i in range(4):
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]
 */
        if (__pyx_v_has_alpha) {

          /* "yt/utilities/lib/misc_utilities.pyx":162
 *             if (x0 >=0 and x0 < nx and y0 >= 0 and y0 < ny):
 *                 if has_alpha:
 *                     for i in range(4):             # <<<<<<<<<<<<<<
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]
 *                 else:
 */
          for (__pyx_t_9 = 0; __pyx_t_9 < 4; __pyx_t_9+=1) {
            __pyx_v_i = __pyx_t_9;

            /* "yt/utilities/lib/misc_utilities.pyx":163
 *                 if has_alpha:
 *                     for i in range(4):
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]             # <<<<<<<<<<<<<<
 *                 else:
 *                     for i in range(3):
 */
            __pyx_t_21 = __pyx_v_x0;
            __pyx_t_22 = __pyx_v_y0;
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_x0;
            __pyx_t_25 = __pyx_v_y0;
            __pyx_t_26 = __pyx_v_i;
            *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_image.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_image.diminfo[0].strides, __pyx_t_25, __pyx_pybuffernd_image.diminfo[1].strides, __pyx_t_26, __pyx_pybuffernd_image.diminfo[2].strides) = (((1. - (__pyx_v_alpha[__pyx_v_i])) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_image.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_image.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_image.diminfo[1].strides, __pyx_t_23, __pyx_pybuffernd_image.diminfo[2].strides))) + (__pyx_v_alpha[__pyx_v_i]));
          }
          goto __pyx_L16;
        }
        /*else*/ {

          /* "yt/utilities/lib/misc_utilities.pyx":165
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]
 *                 else:
 *                     for i in range(3):             # <<<<<<<<<<<<<<
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]
 * 
 */
          for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
            __pyx_v_i = __pyx_t_9;

            /* "yt/utilities/lib/misc_utilities.pyx":166
 *                 else:
 *                     for i in range(3):
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]             # <<<<<<<<<<<<<<
 * 
 *             if (x0 == x1 and y0 == y1):
 */
            __pyx_t_27 = __pyx_v_x0;
            __pyx_t_28 = __pyx_v_y0;
            __pyx_t_29 = __pyx_v_i;
            __pyx_t_30 = __pyx_v_x0;
            __pyx_t_31 = __pyx_v_y0;
            __pyx_t_32 = __pyx_v_i;
            *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_image.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_image.diminfo[0].strides, __pyx_t_31, __pyx_pybuffernd_image.diminfo[1].strides, __pyx_t_32, __pyx_pybuffernd_image.diminfo[2].strides) = (((1. - (__pyx_v_alpha[__pyx_v_i])) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_image.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_image.diminfo[0].strides, __pyx_t_28, __pyx_pybuffernd_image.diminfo[1].strides, __pyx_t_29, __pyx_pybuffernd_image.diminfo[2].strides))) + (__pyx_v_alpha[__pyx_v_i]));
          }
        }
        __pyx_L16:;
        goto __pyx_L15;
      }
      __pyx_L15:;

      /* "yt/utilities/lib/misc_utilities.pyx":168
 *                         image[x0,y0,i] = (1.-alpha[i])*image[x0,y0,i] + alpha[i]
 * 
 *             if (x0 == x1 and y0 == y1):             # <<<<<<<<<<<<<<
 *                 break
 *             e2 = 2*err
 */
      __pyx_t_16 = (__pyx_v_x0 == __pyx_v_x1);
      if (__pyx_t_16) {
        __pyx_t_18 = (__pyx_v_y0 == __pyx_v_y1);
        __pyx_t_17 = __pyx_t_18;
      } else {
        __pyx_t_17 = __pyx_t_16;
      }
      if (__pyx_t_17) {

        /* "yt/utilities/lib/misc_utilities.pyx":169
 * 
 *             if (x0 == x1 and y0 == y1):
 *                 break             # <<<<<<<<<<<<<<
 *             e2 = 2*err
 *             if e2 > -dy:
 */
        goto __pyx_L13_break;
        goto __pyx_L21;
      }
      __pyx_L21:;

      /* "yt/utilities/lib/misc_utilities.pyx":170
 *             if (x0 == x1 and y0 == y1):
 *                 break
 *             e2 = 2*err             # <<<<<<<<<<<<<<
 *             if e2 > -dy:
 *                 err = err - dy
 */
      __pyx_v_e2 = (2 * __pyx_v_err);

      /* "yt/utilities/lib/misc_utilities.pyx":171
 *                 break
 *             e2 = 2*err
 *             if e2 > -dy:             # <<<<<<<<<<<<<<
 *                 err = err - dy
 *                 x0 += sx
 */
      __pyx_t_17 = (__pyx_v_e2 > (-__pyx_v_dy));
      if (__pyx_t_17) {

        /* "yt/utilities/lib/misc_utilities.pyx":172
 *             e2 = 2*err
 *             if e2 > -dy:
 *                 err = err - dy             # <<<<<<<<<<<<<<
 *                 x0 += sx
 *             if e2 < dx :
 */
        __pyx_v_err = (__pyx_v_err - __pyx_v_dy);

        /* "yt/utilities/lib/misc_utilities.pyx":173
 *             if e2 > -dy:
 *                 err = err - dy
 *                 x0 += sx             # <<<<<<<<<<<<<<
 *             if e2 < dx :
 *                 err = err + dx
 */
        __pyx_v_x0 = (__pyx_v_x0 + __pyx_v_sx);
        goto __pyx_L22;
      }
      __pyx_L22:;

      /* "yt/utilities/lib/misc_utilities.pyx":174
 *                 err = err - dy
 *                 x0 += sx
 *             if e2 < dx :             # <<<<<<<<<<<<<<
 *                 err = err + dx
 *                 y0 += sy
 */
      __pyx_t_17 = (__pyx_v_e2 < __pyx_v_dx);
      if (__pyx_t_17) {

        /* "yt/utilities/lib/misc_utilities.pyx":175
 *                 x0 += sx
 *             if e2 < dx :
 *                 err = err + dx             # <<<<<<<<<<<<<<
 *                 y0 += sy
 *     return
 */
        __pyx_v_err = (__pyx_v_err + __pyx_v_dx);

        /* "yt/utilities/lib/misc_utilities.pyx":176
 *             if e2 < dx :
 *                 err = err + dx
 *                 y0 += sy             # <<<<<<<<<<<<<<
 *     return
 * 
 */
        __pyx_v_y0 = (__pyx_v_y0 + __pyx_v_sy);
        goto __pyx_L23;
      }
      __pyx_L23:;
    }
    __pyx_L13_break:;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":177
 *                 err = err + dx
 *                 y0 += sy
 *     return             # <<<<<<<<<<<<<<
 * 
 * def rotate_vectors(np.ndarray[np.float64_t, ndim=3] vecs,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_colors.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_image.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ys.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_colors.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_image.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ys.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_9rotate_vectors(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_9rotate_vectors = {__Pyx_NAMESTR("rotate_vectors"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_9rotate_vectors, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_9rotate_vectors(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_vecs = 0;
  PyArrayObject *__pyx_v_R = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rotate_vectors (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__vecs,&__pyx_n_s__R,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__vecs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__R)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate_vectors", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "rotate_vectors") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_vecs = ((PyArrayObject *)values[0]);
    __pyx_v_R = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("rotate_vectors", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.rotate_vectors", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_vecs), __pyx_ptype_5numpy_ndarray, 1, "vecs", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_R), __pyx_ptype_5numpy_ndarray, 1, "R", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_8rotate_vectors(__pyx_self, __pyx_v_vecs, __pyx_v_R);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":179
 *     return
 * 
 * def rotate_vectors(np.ndarray[np.float64_t, ndim=3] vecs,             # <<<<<<<<<<<<<<
 *         np.ndarray[np.float64_t, ndim=2] R):
 *     cdef int nx = vecs.shape[0]
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_8rotate_vectors(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_vecs, PyArrayObject *__pyx_v_R) {
  int __pyx_v_nx;
  int __pyx_v_ny;
  PyObject *__pyx_v_rotated = NULL;
  int __pyx_v_i;
  int __pyx_v_j;
  long __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_R;
  __Pyx_Buffer __pyx_pybuffer_R;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_vecs;
  __Pyx_Buffer __pyx_pybuffer_vecs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  long __pyx_t_9;
  long __pyx_t_10;
  long __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  long __pyx_t_14;
  int __pyx_t_15;
  long __pyx_t_16;
  long __pyx_t_17;
  int __pyx_t_18;
  long __pyx_t_19;
  int __pyx_t_20;
  long __pyx_t_21;
  long __pyx_t_22;
  int __pyx_t_23;
  long __pyx_t_24;
  int __pyx_t_25;
  PyObject *__pyx_t_26 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rotate_vectors", 0);
  __pyx_pybuffer_vecs.pybuffer.buf = NULL;
  __pyx_pybuffer_vecs.refcount = 0;
  __pyx_pybuffernd_vecs.data = NULL;
  __pyx_pybuffernd_vecs.rcbuffer = &__pyx_pybuffer_vecs;
  __pyx_pybuffer_R.pybuffer.buf = NULL;
  __pyx_pybuffer_R.refcount = 0;
  __pyx_pybuffernd_R.data = NULL;
  __pyx_pybuffernd_R.rcbuffer = &__pyx_pybuffer_R;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_vecs.rcbuffer->pybuffer, (PyObject*)__pyx_v_vecs, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_vecs.diminfo[0].strides = __pyx_pybuffernd_vecs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_vecs.diminfo[0].shape = __pyx_pybuffernd_vecs.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_vecs.diminfo[1].strides = __pyx_pybuffernd_vecs.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_vecs.diminfo[1].shape = __pyx_pybuffernd_vecs.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_vecs.diminfo[2].strides = __pyx_pybuffernd_vecs.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_vecs.diminfo[2].shape = __pyx_pybuffernd_vecs.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_R.rcbuffer->pybuffer, (PyObject*)__pyx_v_R, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_R.diminfo[0].strides = __pyx_pybuffernd_R.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_R.diminfo[0].shape = __pyx_pybuffernd_R.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_R.diminfo[1].strides = __pyx_pybuffernd_R.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_R.diminfo[1].shape = __pyx_pybuffernd_R.rcbuffer->pybuffer.shape[1];

  /* "yt/utilities/lib/misc_utilities.pyx":181
 * def rotate_vectors(np.ndarray[np.float64_t, ndim=3] vecs,
 *         np.ndarray[np.float64_t, ndim=2] R):
 *     cdef int nx = vecs.shape[0]             # <<<<<<<<<<<<<<
 *     cdef int ny = vecs.shape[1]
 *     rotated = np.empty((nx,ny,3),dtype='float64')
 */
  __pyx_v_nx = (__pyx_v_vecs->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":182
 *         np.ndarray[np.float64_t, ndim=2] R):
 *     cdef int nx = vecs.shape[0]
 *     cdef int ny = vecs.shape[1]             # <<<<<<<<<<<<<<
 *     rotated = np.empty((nx,ny,3),dtype='float64')
 *     for i in range(nx):
 */
  __pyx_v_ny = (__pyx_v_vecs->dimensions[1]);

  /* "yt/utilities/lib/misc_utilities.pyx":183
 *     cdef int nx = vecs.shape[0]
 *     cdef int ny = vecs.shape[1]
 *     rotated = np.empty((nx,ny,3),dtype='float64')             # <<<<<<<<<<<<<<
 *     for i in range(nx):
 *         for j in range(ny):
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__empty); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromLong(__pyx_v_nx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromLong(__pyx_v_ny); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_4));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_4));
  __pyx_t_4 = 0;
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_4));
  if (PyDict_SetItem(__pyx_t_4, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__float64)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_4)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __pyx_v_rotated = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":184
 *     cdef int ny = vecs.shape[1]
 *     rotated = np.empty((nx,ny,3),dtype='float64')
 *     for i in range(nx):             # <<<<<<<<<<<<<<
 *         for j in range(ny):
 *             for k in range(3):
 */
  __pyx_t_5 = __pyx_v_nx;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/utilities/lib/misc_utilities.pyx":185
 *     rotated = np.empty((nx,ny,3),dtype='float64')
 *     for i in range(nx):
 *         for j in range(ny):             # <<<<<<<<<<<<<<
 *             for k in range(3):
 *                 rotated[i,j,k] =\
 */
    __pyx_t_7 = __pyx_v_ny;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_j = __pyx_t_8;

      /* "yt/utilities/lib/misc_utilities.pyx":186
 *     for i in range(nx):
 *         for j in range(ny):
 *             for k in range(3):             # <<<<<<<<<<<<<<
 *                 rotated[i,j,k] =\
 *                     R[k,0]*vecs[i,j,0]+R[k,1]*vecs[i,j,1]+R[k,2]*vecs[i,j,2]
 */
      for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/utilities/lib/misc_utilities.pyx":188
 *             for k in range(3):
 *                 rotated[i,j,k] =\
 *                     R[k,0]*vecs[i,j,0]+R[k,1]*vecs[i,j,1]+R[k,2]*vecs[i,j,2]             # <<<<<<<<<<<<<<
 *     return rotated
 * 
 */
        __pyx_t_10 = __pyx_v_k;
        __pyx_t_11 = 0;
        __pyx_t_12 = -1;
        if (__pyx_t_10 < 0) {
          __pyx_t_10 += __pyx_pybuffernd_R.diminfo[0].shape;
          if (unlikely(__pyx_t_10 < 0)) __pyx_t_12 = 0;
        } else if (unlikely(__pyx_t_10 >= __pyx_pybuffernd_R.diminfo[0].shape)) __pyx_t_12 = 0;
        if (__pyx_t_11 < 0) {
          __pyx_t_11 += __pyx_pybuffernd_R.diminfo[1].shape;
          if (unlikely(__pyx_t_11 < 0)) __pyx_t_12 = 1;
        } else if (unlikely(__pyx_t_11 >= __pyx_pybuffernd_R.diminfo[1].shape)) __pyx_t_12 = 1;
        if (unlikely(__pyx_t_12 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_12);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_12 = __pyx_v_i;
        __pyx_t_13 = __pyx_v_j;
        __pyx_t_14 = 0;
        __pyx_t_15 = -1;
        if (__pyx_t_12 < 0) {
          __pyx_t_12 += __pyx_pybuffernd_vecs.diminfo[0].shape;
          if (unlikely(__pyx_t_12 < 0)) __pyx_t_15 = 0;
        } else if (unlikely(__pyx_t_12 >= __pyx_pybuffernd_vecs.diminfo[0].shape)) __pyx_t_15 = 0;
        if (__pyx_t_13 < 0) {
          __pyx_t_13 += __pyx_pybuffernd_vecs.diminfo[1].shape;
          if (unlikely(__pyx_t_13 < 0)) __pyx_t_15 = 1;
        } else if (unlikely(__pyx_t_13 >= __pyx_pybuffernd_vecs.diminfo[1].shape)) __pyx_t_15 = 1;
        if (__pyx_t_14 < 0) {
          __pyx_t_14 += __pyx_pybuffernd_vecs.diminfo[2].shape;
          if (unlikely(__pyx_t_14 < 0)) __pyx_t_15 = 2;
        } else if (unlikely(__pyx_t_14 >= __pyx_pybuffernd_vecs.diminfo[2].shape)) __pyx_t_15 = 2;
        if (unlikely(__pyx_t_15 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_15);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_16 = __pyx_v_k;
        __pyx_t_17 = 1;
        __pyx_t_15 = -1;
        if (__pyx_t_16 < 0) {
          __pyx_t_16 += __pyx_pybuffernd_R.diminfo[0].shape;
          if (unlikely(__pyx_t_16 < 0)) __pyx_t_15 = 0;
        } else if (unlikely(__pyx_t_16 >= __pyx_pybuffernd_R.diminfo[0].shape)) __pyx_t_15 = 0;
        if (__pyx_t_17 < 0) {
          __pyx_t_17 += __pyx_pybuffernd_R.diminfo[1].shape;
          if (unlikely(__pyx_t_17 < 0)) __pyx_t_15 = 1;
        } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_R.diminfo[1].shape)) __pyx_t_15 = 1;
        if (unlikely(__pyx_t_15 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_15);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_15 = __pyx_v_i;
        __pyx_t_18 = __pyx_v_j;
        __pyx_t_19 = 1;
        __pyx_t_20 = -1;
        if (__pyx_t_15 < 0) {
          __pyx_t_15 += __pyx_pybuffernd_vecs.diminfo[0].shape;
          if (unlikely(__pyx_t_15 < 0)) __pyx_t_20 = 0;
        } else if (unlikely(__pyx_t_15 >= __pyx_pybuffernd_vecs.diminfo[0].shape)) __pyx_t_20 = 0;
        if (__pyx_t_18 < 0) {
          __pyx_t_18 += __pyx_pybuffernd_vecs.diminfo[1].shape;
          if (unlikely(__pyx_t_18 < 0)) __pyx_t_20 = 1;
        } else if (unlikely(__pyx_t_18 >= __pyx_pybuffernd_vecs.diminfo[1].shape)) __pyx_t_20 = 1;
        if (__pyx_t_19 < 0) {
          __pyx_t_19 += __pyx_pybuffernd_vecs.diminfo[2].shape;
          if (unlikely(__pyx_t_19 < 0)) __pyx_t_20 = 2;
        } else if (unlikely(__pyx_t_19 >= __pyx_pybuffernd_vecs.diminfo[2].shape)) __pyx_t_20 = 2;
        if (unlikely(__pyx_t_20 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_20);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_21 = __pyx_v_k;
        __pyx_t_22 = 2;
        __pyx_t_20 = -1;
        if (__pyx_t_21 < 0) {
          __pyx_t_21 += __pyx_pybuffernd_R.diminfo[0].shape;
          if (unlikely(__pyx_t_21 < 0)) __pyx_t_20 = 0;
        } else if (unlikely(__pyx_t_21 >= __pyx_pybuffernd_R.diminfo[0].shape)) __pyx_t_20 = 0;
        if (__pyx_t_22 < 0) {
          __pyx_t_22 += __pyx_pybuffernd_R.diminfo[1].shape;
          if (unlikely(__pyx_t_22 < 0)) __pyx_t_20 = 1;
        } else if (unlikely(__pyx_t_22 >= __pyx_pybuffernd_R.diminfo[1].shape)) __pyx_t_20 = 1;
        if (unlikely(__pyx_t_20 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_20);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_23 = __pyx_v_j;
        __pyx_t_24 = 2;
        __pyx_t_25 = -1;
        if (__pyx_t_20 < 0) {
          __pyx_t_20 += __pyx_pybuffernd_vecs.diminfo[0].shape;
          if (unlikely(__pyx_t_20 < 0)) __pyx_t_25 = 0;
        } else if (unlikely(__pyx_t_20 >= __pyx_pybuffernd_vecs.diminfo[0].shape)) __pyx_t_25 = 0;
        if (__pyx_t_23 < 0) {
          __pyx_t_23 += __pyx_pybuffernd_vecs.diminfo[1].shape;
          if (unlikely(__pyx_t_23 < 0)) __pyx_t_25 = 1;
        } else if (unlikely(__pyx_t_23 >= __pyx_pybuffernd_vecs.diminfo[1].shape)) __pyx_t_25 = 1;
        if (__pyx_t_24 < 0) {
          __pyx_t_24 += __pyx_pybuffernd_vecs.diminfo[2].shape;
          if (unlikely(__pyx_t_24 < 0)) __pyx_t_25 = 2;
        } else if (unlikely(__pyx_t_24 >= __pyx_pybuffernd_vecs.diminfo[2].shape)) __pyx_t_25 = 2;
        if (unlikely(__pyx_t_25 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_25);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_1 = PyFloat_FromDouble(((((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_R.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_R.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_R.diminfo[1].strides)) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_vecs.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_vecs.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_vecs.diminfo[1].strides, __pyx_t_14, __pyx_pybuffernd_vecs.diminfo[2].strides))) + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_R.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_R.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_R.diminfo[1].strides)) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_vecs.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_vecs.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_vecs.diminfo[1].strides, __pyx_t_19, __pyx_pybuffernd_vecs.diminfo[2].strides)))) + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_R.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_R.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_R.diminfo[1].strides)) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_vecs.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_vecs.diminfo[0].strides, __pyx_t_23, __pyx_pybuffernd_vecs.diminfo[1].strides, __pyx_t_24, __pyx_pybuffernd_vecs.diminfo[2].strides))))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);

        /* "yt/utilities/lib/misc_utilities.pyx":187
 *         for j in range(ny):
 *             for k in range(3):
 *                 rotated[i,j,k] =\             # <<<<<<<<<<<<<<
 *                     R[k,0]*vecs[i,j,0]+R[k,1]*vecs[i,j,1]+R[k,2]*vecs[i,j,2]
 *     return rotated
 */
        __pyx_t_4 = PyInt_FromLong(__pyx_v_i); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = PyInt_FromLong(__pyx_v_j); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyInt_FromLong(__pyx_v_k); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_26 = PyTuple_New(3); if (unlikely(!__pyx_t_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_26);
        PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_26, 2, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __pyx_t_4 = 0;
        __pyx_t_3 = 0;
        __pyx_t_2 = 0;
        if (PyObject_SetItem(__pyx_v_rotated, ((PyObject *)__pyx_t_26), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(((PyObject *)__pyx_t_26)); __pyx_t_26 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
  }

  /* "yt/utilities/lib/misc_utilities.pyx":189
 *                 rotated[i,j,k] =\
 *                     R[k,0]*vecs[i,j,0]+R[k,1]*vecs[i,j,1]+R[k,2]*vecs[i,j,2]
 *     return rotated             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rotated);
  __pyx_r = __pyx_v_rotated;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_26);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_R.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_vecs.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.rotate_vectors", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_R.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_vecs.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_rotated);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_11get_color_bounds(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_11get_color_bounds = {__Pyx_NAMESTR("get_color_bounds"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_11get_color_bounds, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_11get_color_bounds(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_px = 0;
  PyArrayObject *__pyx_v_py = 0;
  PyArrayObject *__pyx_v_pdx = 0;
  PyArrayObject *__pyx_v_pdy = 0;
  PyArrayObject *__pyx_v_value = 0;
  __pyx_t_5numpy_float64_t __pyx_v_leftx;
  __pyx_t_5numpy_float64_t __pyx_v_rightx;
  __pyx_t_5numpy_float64_t __pyx_v_lefty;
  __pyx_t_5numpy_float64_t __pyx_v_righty;
  __pyx_t_5numpy_float64_t __pyx_v_mindx;
  __pyx_t_5numpy_float64_t __pyx_v_maxdx;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_color_bounds (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__px,&__pyx_n_s__py,&__pyx_n_s__pdx,&__pyx_n_s__pdy,&__pyx_n_s__value,&__pyx_n_s__leftx,&__pyx_n_s__rightx,&__pyx_n_s__lefty,&__pyx_n_s__righty,&__pyx_n_s__mindx,&__pyx_n_s__maxdx,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__px)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__py)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__pdx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__pdy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__leftx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__rightx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__lefty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__righty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  9:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mindx);
          if (value) { values[9] = value; kw_args--; }
        }
        case 10:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__maxdx);
          if (value) { values[10] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_color_bounds") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_px = ((PyArrayObject *)values[0]);
    __pyx_v_py = ((PyArrayObject *)values[1]);
    __pyx_v_pdx = ((PyArrayObject *)values[2]);
    __pyx_v_pdy = ((PyArrayObject *)values[3]);
    __pyx_v_value = ((PyArrayObject *)values[4]);
    __pyx_v_leftx = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_leftx == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 199; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_rightx = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_rightx == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 199; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_lefty = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_lefty == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_righty = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_righty == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    if (values[9]) {
      __pyx_v_mindx = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_mindx == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 201; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {

      /* "yt/utilities/lib/misc_utilities.pyx":201
 *                      np.float64_t leftx, np.float64_t rightx,
 *                      np.float64_t lefty, np.float64_t righty,
 *                      np.float64_t mindx = -1, np.float64_t maxdx = -1):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef np.float64_t mi = 1e100, ma = -1e100, v
 */
      __pyx_v_mindx = ((__pyx_t_5numpy_float64_t)-1.0);
    }
    if (values[10]) {
      __pyx_v_maxdx = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_maxdx == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 201; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {
      __pyx_v_maxdx = ((__pyx_t_5numpy_float64_t)-1.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_color_bounds", 0, 9, 11, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.get_color_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_px), __pyx_ptype_5numpy_ndarray, 1, "px", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_py), __pyx_ptype_5numpy_ndarray, 1, "py", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pdx), __pyx_ptype_5numpy_ndarray, 1, "pdx", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pdy), __pyx_ptype_5numpy_ndarray, 1, "pdy", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 197; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_5numpy_ndarray, 1, "value", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 198; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_10get_color_bounds(__pyx_self, __pyx_v_px, __pyx_v_py, __pyx_v_pdx, __pyx_v_pdy, __pyx_v_value, __pyx_v_leftx, __pyx_v_rightx, __pyx_v_lefty, __pyx_v_righty, __pyx_v_mindx, __pyx_v_maxdx);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":194
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_color_bounds(np.ndarray[np.float64_t, ndim=1] px,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=1] py,
 *                      np.ndarray[np.float64_t, ndim=1] pdx,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_10get_color_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_px, PyArrayObject *__pyx_v_py, PyArrayObject *__pyx_v_pdx, PyArrayObject *__pyx_v_pdy, PyArrayObject *__pyx_v_value, __pyx_t_5numpy_float64_t __pyx_v_leftx, __pyx_t_5numpy_float64_t __pyx_v_rightx, __pyx_t_5numpy_float64_t __pyx_v_lefty, __pyx_t_5numpy_float64_t __pyx_v_righty, __pyx_t_5numpy_float64_t __pyx_v_mindx, __pyx_t_5numpy_float64_t __pyx_v_maxdx) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_mi;
  __pyx_t_5numpy_float64_t __pyx_v_ma;
  __pyx_t_5numpy_float64_t __pyx_v_v;
  int __pyx_v_np;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pdx;
  __Pyx_Buffer __pyx_pybuffer_pdx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pdy;
  __Pyx_Buffer __pyx_pybuffer_pdy;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_px;
  __Pyx_Buffer __pyx_pybuffer_px;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_py;
  __Pyx_Buffer __pyx_pybuffer_py;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_value;
  __Pyx_Buffer __pyx_pybuffer_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_color_bounds", 0);
  __pyx_pybuffer_px.pybuffer.buf = NULL;
  __pyx_pybuffer_px.refcount = 0;
  __pyx_pybuffernd_px.data = NULL;
  __pyx_pybuffernd_px.rcbuffer = &__pyx_pybuffer_px;
  __pyx_pybuffer_py.pybuffer.buf = NULL;
  __pyx_pybuffer_py.refcount = 0;
  __pyx_pybuffernd_py.data = NULL;
  __pyx_pybuffernd_py.rcbuffer = &__pyx_pybuffer_py;
  __pyx_pybuffer_pdx.pybuffer.buf = NULL;
  __pyx_pybuffer_pdx.refcount = 0;
  __pyx_pybuffernd_pdx.data = NULL;
  __pyx_pybuffernd_pdx.rcbuffer = &__pyx_pybuffer_pdx;
  __pyx_pybuffer_pdy.pybuffer.buf = NULL;
  __pyx_pybuffer_pdy.refcount = 0;
  __pyx_pybuffernd_pdy.data = NULL;
  __pyx_pybuffernd_pdy.rcbuffer = &__pyx_pybuffer_pdy;
  __pyx_pybuffer_value.pybuffer.buf = NULL;
  __pyx_pybuffer_value.refcount = 0;
  __pyx_pybuffernd_value.data = NULL;
  __pyx_pybuffernd_value.rcbuffer = &__pyx_pybuffer_value;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_px.rcbuffer->pybuffer, (PyObject*)__pyx_v_px, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_px.diminfo[0].strides = __pyx_pybuffernd_px.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_px.diminfo[0].shape = __pyx_pybuffernd_px.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_py.rcbuffer->pybuffer, (PyObject*)__pyx_v_py, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_py.diminfo[0].strides = __pyx_pybuffernd_py.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_py.diminfo[0].shape = __pyx_pybuffernd_py.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pdx.rcbuffer->pybuffer, (PyObject*)__pyx_v_pdx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_pdx.diminfo[0].strides = __pyx_pybuffernd_pdx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pdx.diminfo[0].shape = __pyx_pybuffernd_pdx.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pdy.rcbuffer->pybuffer, (PyObject*)__pyx_v_pdy, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_pdy.diminfo[0].strides = __pyx_pybuffernd_pdy.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pdy.diminfo[0].shape = __pyx_pybuffernd_pdy.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_value.rcbuffer->pybuffer, (PyObject*)__pyx_v_value, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_value.diminfo[0].strides = __pyx_pybuffernd_value.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_value.diminfo[0].shape = __pyx_pybuffernd_value.rcbuffer->pybuffer.shape[0];

  /* "yt/utilities/lib/misc_utilities.pyx":203
 *                      np.float64_t mindx = -1, np.float64_t maxdx = -1):
 *     cdef int i
 *     cdef np.float64_t mi = 1e100, ma = -1e100, v             # <<<<<<<<<<<<<<
 *     cdef int np = px.shape[0]
 *     with nogil:
 */
  __pyx_v_mi = 1e100;
  __pyx_v_ma = -1e100;

  /* "yt/utilities/lib/misc_utilities.pyx":204
 *     cdef int i
 *     cdef np.float64_t mi = 1e100, ma = -1e100, v
 *     cdef int np = px.shape[0]             # <<<<<<<<<<<<<<
 *     with nogil:
 *         for i in range(np):
 */
  __pyx_v_np = (__pyx_v_px->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":205
 *     cdef np.float64_t mi = 1e100, ma = -1e100, v
 *     cdef int np = px.shape[0]
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(np):
 *             v = value[i]
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save = NULL;
      #endif
      Py_UNBLOCK_THREADS
      /*try:*/ {

        /* "yt/utilities/lib/misc_utilities.pyx":206
 *     cdef int np = px.shape[0]
 *     with nogil:
 *         for i in range(np):             # <<<<<<<<<<<<<<
 *             v = value[i]
 *             if v < mi or v > ma:
 */
        __pyx_t_1 = __pyx_v_np;
        for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "yt/utilities/lib/misc_utilities.pyx":207
 *     with nogil:
 *         for i in range(np):
 *             v = value[i]             # <<<<<<<<<<<<<<
 *             if v < mi or v > ma:
 *                 if px[i] + pdx[i] < leftx: continue
 */
          __pyx_t_3 = __pyx_v_i;
          __pyx_v_v = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_value.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_value.diminfo[0].strides));

          /* "yt/utilities/lib/misc_utilities.pyx":208
 *         for i in range(np):
 *             v = value[i]
 *             if v < mi or v > ma:             # <<<<<<<<<<<<<<
 *                 if px[i] + pdx[i] < leftx: continue
 *                 if px[i] - pdx[i] > rightx: continue
 */
          __pyx_t_4 = (__pyx_v_v < __pyx_v_mi);
          if (!__pyx_t_4) {
            __pyx_t_5 = (__pyx_v_v > __pyx_v_ma);
            __pyx_t_6 = __pyx_t_5;
          } else {
            __pyx_t_6 = __pyx_t_4;
          }
          if (__pyx_t_6) {

            /* "yt/utilities/lib/misc_utilities.pyx":209
 *             v = value[i]
 *             if v < mi or v > ma:
 *                 if px[i] + pdx[i] < leftx: continue             # <<<<<<<<<<<<<<
 *                 if px[i] - pdx[i] > rightx: continue
 *                 if py[i] + pdy[i] < lefty: continue
 */
            __pyx_t_7 = __pyx_v_i;
            __pyx_t_8 = __pyx_v_i;
            __pyx_t_6 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_px.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_px.diminfo[0].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdx.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_pdx.diminfo[0].strides))) < __pyx_v_leftx);
            if (__pyx_t_6) {
              goto __pyx_L6_continue;
              goto __pyx_L9;
            }
            __pyx_L9:;

            /* "yt/utilities/lib/misc_utilities.pyx":210
 *             if v < mi or v > ma:
 *                 if px[i] + pdx[i] < leftx: continue
 *                 if px[i] - pdx[i] > rightx: continue             # <<<<<<<<<<<<<<
 *                 if py[i] + pdy[i] < lefty: continue
 *                 if py[i] - pdy[i] > righty: continue
 */
            __pyx_t_9 = __pyx_v_i;
            __pyx_t_10 = __pyx_v_i;
            __pyx_t_6 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_px.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_px.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdx.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_pdx.diminfo[0].strides))) > __pyx_v_rightx);
            if (__pyx_t_6) {
              goto __pyx_L6_continue;
              goto __pyx_L10;
            }
            __pyx_L10:;

            /* "yt/utilities/lib/misc_utilities.pyx":211
 *                 if px[i] + pdx[i] < leftx: continue
 *                 if px[i] - pdx[i] > rightx: continue
 *                 if py[i] + pdy[i] < lefty: continue             # <<<<<<<<<<<<<<
 *                 if py[i] - pdy[i] > righty: continue
 *                 if pdx[i] < mindx or pdy[i] < mindx: continue
 */
            __pyx_t_11 = __pyx_v_i;
            __pyx_t_12 = __pyx_v_i;
            __pyx_t_6 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_py.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_py.diminfo[0].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdy.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_pdy.diminfo[0].strides))) < __pyx_v_lefty);
            if (__pyx_t_6) {
              goto __pyx_L6_continue;
              goto __pyx_L11;
            }
            __pyx_L11:;

            /* "yt/utilities/lib/misc_utilities.pyx":212
 *                 if px[i] - pdx[i] > rightx: continue
 *                 if py[i] + pdy[i] < lefty: continue
 *                 if py[i] - pdy[i] > righty: continue             # <<<<<<<<<<<<<<
 *                 if pdx[i] < mindx or pdy[i] < mindx: continue
 *                 if maxdx > 0 and (pdx[i] > maxdx or pdy[i] > maxdx): continue
 */
            __pyx_t_13 = __pyx_v_i;
            __pyx_t_14 = __pyx_v_i;
            __pyx_t_6 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_py.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_py.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdy.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_pdy.diminfo[0].strides))) > __pyx_v_righty);
            if (__pyx_t_6) {
              goto __pyx_L6_continue;
              goto __pyx_L12;
            }
            __pyx_L12:;

            /* "yt/utilities/lib/misc_utilities.pyx":213
 *                 if py[i] + pdy[i] < lefty: continue
 *                 if py[i] - pdy[i] > righty: continue
 *                 if pdx[i] < mindx or pdy[i] < mindx: continue             # <<<<<<<<<<<<<<
 *                 if maxdx > 0 and (pdx[i] > maxdx or pdy[i] > maxdx): continue
 *                 if v < mi: mi = v
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_t_6 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdx.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_pdx.diminfo[0].strides)) < __pyx_v_mindx);
            if (!__pyx_t_6) {
              __pyx_t_16 = __pyx_v_i;
              __pyx_t_4 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdy.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pdy.diminfo[0].strides)) < __pyx_v_mindx);
              __pyx_t_5 = __pyx_t_4;
            } else {
              __pyx_t_5 = __pyx_t_6;
            }
            if (__pyx_t_5) {
              goto __pyx_L6_continue;
              goto __pyx_L13;
            }
            __pyx_L13:;

            /* "yt/utilities/lib/misc_utilities.pyx":214
 *                 if py[i] - pdy[i] > righty: continue
 *                 if pdx[i] < mindx or pdy[i] < mindx: continue
 *                 if maxdx > 0 and (pdx[i] > maxdx or pdy[i] > maxdx): continue             # <<<<<<<<<<<<<<
 *                 if v < mi: mi = v
 *                 if v > ma: ma = v
 */
            __pyx_t_5 = (__pyx_v_maxdx > 0.0);
            if (__pyx_t_5) {
              __pyx_t_17 = __pyx_v_i;
              __pyx_t_6 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdx.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_pdx.diminfo[0].strides)) > __pyx_v_maxdx);
              if (!__pyx_t_6) {
                __pyx_t_18 = __pyx_v_i;
                __pyx_t_4 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pdy.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_pdy.diminfo[0].strides)) > __pyx_v_maxdx);
                __pyx_t_19 = __pyx_t_4;
              } else {
                __pyx_t_19 = __pyx_t_6;
              }
              __pyx_t_6 = __pyx_t_19;
            } else {
              __pyx_t_6 = __pyx_t_5;
            }
            if (__pyx_t_6) {
              goto __pyx_L6_continue;
              goto __pyx_L14;
            }
            __pyx_L14:;

            /* "yt/utilities/lib/misc_utilities.pyx":215
 *                 if pdx[i] < mindx or pdy[i] < mindx: continue
 *                 if maxdx > 0 and (pdx[i] > maxdx or pdy[i] > maxdx): continue
 *                 if v < mi: mi = v             # <<<<<<<<<<<<<<
 *                 if v > ma: ma = v
 *     return (mi, ma)
 */
            __pyx_t_6 = (__pyx_v_v < __pyx_v_mi);
            if (__pyx_t_6) {
              __pyx_v_mi = __pyx_v_v;
              goto __pyx_L15;
            }
            __pyx_L15:;

            /* "yt/utilities/lib/misc_utilities.pyx":216
 *                 if maxdx > 0 and (pdx[i] > maxdx or pdy[i] > maxdx): continue
 *                 if v < mi: mi = v
 *                 if v > ma: ma = v             # <<<<<<<<<<<<<<
 *     return (mi, ma)
 * 
 */
            __pyx_t_6 = (__pyx_v_v > __pyx_v_ma);
            if (__pyx_t_6) {
              __pyx_v_ma = __pyx_v_v;
              goto __pyx_L16;
            }
            __pyx_L16:;
            goto __pyx_L8;
          }
          __pyx_L8:;
          __pyx_L6_continue:;
        }
      }

      /* "yt/utilities/lib/misc_utilities.pyx":205
 *     cdef np.float64_t mi = 1e100, ma = -1e100, v
 *     cdef int np = px.shape[0]
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(np):
 *             v = value[i]
 */
      /*finally:*/ {
        Py_BLOCK_THREADS
      }
  }

  /* "yt/utilities/lib/misc_utilities.pyx":217
 *                 if v < mi: mi = v
 *                 if v > ma: ma = v
 *     return (mi, ma)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_20 = PyFloat_FromDouble(__pyx_v_mi); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = PyFloat_FromDouble(__pyx_v_ma); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_r = ((PyObject *)__pyx_t_22);
  __pyx_t_22 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pdx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pdy.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_px.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_py.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_value.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.get_color_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pdx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pdy.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_px.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_py.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_value.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_13get_box_grids_level(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_13get_box_grids_level = {__Pyx_NAMESTR("get_box_grids_level"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_13get_box_grids_level, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_13get_box_grids_level(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edge = 0;
  PyArrayObject *__pyx_v_right_edge = 0;
  int __pyx_v_level;
  PyArrayObject *__pyx_v_left_edges = 0;
  PyArrayObject *__pyx_v_right_edges = 0;
  PyArrayObject *__pyx_v_levels = 0;
  PyArrayObject *__pyx_v_mask = 0;
  int __pyx_v_min_index;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_box_grids_level (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__left_edge,&__pyx_n_s__right_edge,&__pyx_n_s__level,&__pyx_n_s__left_edges,&__pyx_n_s__right_edges,&__pyx_n_s__levels,&__pyx_n_s__mask,&__pyx_n_s__min_index,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__left_edge)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__right_edge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__left_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__min_index);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_box_grids_level") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_left_edge = ((PyArrayObject *)values[0]);
    __pyx_v_right_edge = ((PyArrayObject *)values[1]);
    __pyx_v_level = __Pyx_PyInt_AsInt(values[2]); if (unlikely((__pyx_v_level == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 224; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_left_edges = ((PyArrayObject *)values[3]);
    __pyx_v_right_edges = ((PyArrayObject *)values[4]);
    __pyx_v_levels = ((PyArrayObject *)values[5]);
    __pyx_v_mask = ((PyArrayObject *)values[6]);
    if (values[7]) {
      __pyx_v_min_index = __Pyx_PyInt_AsInt(values[7]); if (unlikely((__pyx_v_min_index == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {
      __pyx_v_min_index = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_box_grids_level", 0, 7, 8, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.get_box_grids_level", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edge), __pyx_ptype_5numpy_ndarray, 1, "left_edge", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edge), __pyx_ptype_5numpy_ndarray, 1, "right_edge", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 226; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 227; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_12get_box_grids_level(__pyx_self, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels, __pyx_v_mask, __pyx_v_min_index);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":222
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_box_grids_level(np.ndarray[np.float64_t, ndim=1] left_edge,             # <<<<<<<<<<<<<<
 *                         np.ndarray[np.float64_t, ndim=1] right_edge,
 *                         int level,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_12get_box_grids_level(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_left_edge, PyArrayObject *__pyx_v_right_edge, int __pyx_v_level, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels, PyArrayObject *__pyx_v_mask, int __pyx_v_min_index) {
  int __pyx_v_i;
  int __pyx_v_n;
  int __pyx_v_nx;
  int __pyx_v_inside;
  __pyx_t_5numpy_float64_t __pyx_v_eps;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edge;
  __Pyx_Buffer __pyx_pybuffer_left_edge;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edge;
  __Pyx_Buffer __pyx_pybuffer_right_edge;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  long __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_box_grids_level", 0);
  __pyx_pybuffer_left_edge.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edge.refcount = 0;
  __pyx_pybuffernd_left_edge.data = NULL;
  __pyx_pybuffernd_left_edge.rcbuffer = &__pyx_pybuffer_left_edge;
  __pyx_pybuffer_right_edge.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edge.refcount = 0;
  __pyx_pybuffernd_right_edge.data = NULL;
  __pyx_pybuffernd_right_edge.rcbuffer = &__pyx_pybuffer_right_edge;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edge.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edge, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_left_edge.diminfo[0].strides = __pyx_pybuffernd_left_edge.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edge.diminfo[0].shape = __pyx_pybuffernd_left_edge.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edge.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edge, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_right_edge.diminfo[0].strides = __pyx_pybuffernd_right_edge.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edge.diminfo[0].shape = __pyx_pybuffernd_right_edge.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "yt/utilities/lib/misc_utilities.pyx":231
 *                         int min_index = 0):
 *     cdef int i, n
 *     cdef int nx = left_edges.shape[0]             # <<<<<<<<<<<<<<
 *     cdef int inside
 *     cdef np.float64_t eps = np.finfo(np.float64).eps
 */
  __pyx_v_nx = (__pyx_v_left_edges->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":233
 *     cdef int nx = left_edges.shape[0]
 *     cdef int inside
 *     cdef np.float64_t eps = np.finfo(np.float64).eps             # <<<<<<<<<<<<<<
 *     for i in range(nx):
 *         if i < min_index or levels[i,0] != level:
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__finfo); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__float64); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__eps); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_eps = __pyx_t_4;

  /* "yt/utilities/lib/misc_utilities.pyx":234
 *     cdef int inside
 *     cdef np.float64_t eps = np.finfo(np.float64).eps
 *     for i in range(nx):             # <<<<<<<<<<<<<<
 *         if i < min_index or levels[i,0] != level:
 *             mask[i] = 0
 */
  __pyx_t_5 = __pyx_v_nx;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/utilities/lib/misc_utilities.pyx":235
 *     cdef np.float64_t eps = np.finfo(np.float64).eps
 *     for i in range(nx):
 *         if i < min_index or levels[i,0] != level:             # <<<<<<<<<<<<<<
 *             mask[i] = 0
 *             continue
 */
    __pyx_t_7 = (__pyx_v_i < __pyx_v_min_index);
    if (!__pyx_t_7) {
      __pyx_t_8 = __pyx_v_i;
      __pyx_t_9 = 0;
      __pyx_t_10 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_levels.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_levels.diminfo[0].strides, __pyx_t_9, __pyx_pybuffernd_levels.diminfo[1].strides)) != __pyx_v_level);
      __pyx_t_11 = __pyx_t_10;
    } else {
      __pyx_t_11 = __pyx_t_7;
    }
    if (__pyx_t_11) {

      /* "yt/utilities/lib/misc_utilities.pyx":236
 *     for i in range(nx):
 *         if i < min_index or levels[i,0] != level:
 *             mask[i] = 0             # <<<<<<<<<<<<<<
 *             continue
 *         inside = 1
 */
      __pyx_t_12 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_mask.diminfo[0].strides) = 0;

      /* "yt/utilities/lib/misc_utilities.pyx":237
 *         if i < min_index or levels[i,0] != level:
 *             mask[i] = 0
 *             continue             # <<<<<<<<<<<<<<
 *         inside = 1
 *         for n in range(3):
 */
      goto __pyx_L3_continue;
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "yt/utilities/lib/misc_utilities.pyx":238
 *             mask[i] = 0
 *             continue
 *         inside = 1             # <<<<<<<<<<<<<<
 *         for n in range(3):
 *             if (right_edges[i,n] - left_edge[n]) <= eps or \
 */
    __pyx_v_inside = 1;

    /* "yt/utilities/lib/misc_utilities.pyx":239
 *             continue
 *         inside = 1
 *         for n in range(3):             # <<<<<<<<<<<<<<
 *             if (right_edges[i,n] - left_edge[n]) <= eps or \
 *                (right_edge[n] - left_edges[i,n]) <= eps:
 */
    for (__pyx_t_13 = 0; __pyx_t_13 < 3; __pyx_t_13+=1) {
      __pyx_v_n = __pyx_t_13;

      /* "yt/utilities/lib/misc_utilities.pyx":240
 *         inside = 1
 *         for n in range(3):
 *             if (right_edges[i,n] - left_edge[n]) <= eps or \             # <<<<<<<<<<<<<<
 *                (right_edge[n] - left_edges[i,n]) <= eps:
 *                 inside = 0
 */
      __pyx_t_14 = __pyx_v_i;
      __pyx_t_15 = __pyx_v_n;
      __pyx_t_16 = __pyx_v_n;
      __pyx_t_11 = (((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_right_edges.diminfo[1].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edge.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_left_edge.diminfo[0].strides))) <= __pyx_v_eps);
      if (!__pyx_t_11) {

        /* "yt/utilities/lib/misc_utilities.pyx":241
 *         for n in range(3):
 *             if (right_edges[i,n] - left_edge[n]) <= eps or \
 *                (right_edge[n] - left_edges[i,n]) <= eps:             # <<<<<<<<<<<<<<
 *                 inside = 0
 *                 break
 */
        __pyx_t_17 = __pyx_v_n;
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_19 = __pyx_v_n;
        __pyx_t_7 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edge.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_right_edge.diminfo[0].strides)) - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_left_edges.diminfo[1].strides))) <= __pyx_v_eps);
        __pyx_t_10 = __pyx_t_7;
      } else {
        __pyx_t_10 = __pyx_t_11;
      }
      if (__pyx_t_10) {

        /* "yt/utilities/lib/misc_utilities.pyx":242
 *             if (right_edges[i,n] - left_edge[n]) <= eps or \
 *                (right_edge[n] - left_edges[i,n]) <= eps:
 *                 inside = 0             # <<<<<<<<<<<<<<
 *                 break
 *         if inside == 1: mask[i] = 1
 */
        __pyx_v_inside = 0;

        /* "yt/utilities/lib/misc_utilities.pyx":243
 *                (right_edge[n] - left_edges[i,n]) <= eps:
 *                 inside = 0
 *                 break             # <<<<<<<<<<<<<<
 *         if inside == 1: mask[i] = 1
 *         else: mask[i] = 0
 */
        goto __pyx_L7_break;
        goto __pyx_L8;
      }
      __pyx_L8:;
    }
    __pyx_L7_break:;

    /* "yt/utilities/lib/misc_utilities.pyx":244
 *                 inside = 0
 *                 break
 *         if inside == 1: mask[i] = 1             # <<<<<<<<<<<<<<
 *         else: mask[i] = 0
 * 
 */
    __pyx_t_10 = (__pyx_v_inside == 1);
    if (__pyx_t_10) {
      __pyx_t_13 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_mask.diminfo[0].strides) = 1;
      goto __pyx_L9;
    }
    /*else*/ {

      /* "yt/utilities/lib/misc_utilities.pyx":245
 *                 break
 *         if inside == 1: mask[i] = 1
 *         else: mask[i] = 0             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
      __pyx_t_20 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_mask.diminfo[0].strides) = 0;
    }
    __pyx_L9:;
    __pyx_L3_continue:;
  }

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edge.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edge.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.get_box_grids_level", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edge.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edge.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_15get_box_grids_below_level(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_15get_box_grids_below_level = {__Pyx_NAMESTR("get_box_grids_below_level"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_15get_box_grids_below_level, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_15get_box_grids_below_level(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edge = 0;
  PyArrayObject *__pyx_v_right_edge = 0;
  int __pyx_v_level;
  PyArrayObject *__pyx_v_left_edges = 0;
  PyArrayObject *__pyx_v_right_edges = 0;
  PyArrayObject *__pyx_v_levels = 0;
  PyArrayObject *__pyx_v_mask = 0;
  int __pyx_v_min_level;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_box_grids_below_level (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__left_edge,&__pyx_n_s__right_edge,&__pyx_n_s__level,&__pyx_n_s__left_edges,&__pyx_n_s__right_edges,&__pyx_n_s__levels,&__pyx_n_s__mask,&__pyx_n_s__min_level,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__left_edge)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__right_edge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__left_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__min_level);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_box_grids_below_level") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_left_edge = ((PyArrayObject *)values[0]);
    __pyx_v_right_edge = ((PyArrayObject *)values[1]);
    __pyx_v_level = __Pyx_PyInt_AsInt(values[2]); if (unlikely((__pyx_v_level == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_left_edges = ((PyArrayObject *)values[3]);
    __pyx_v_right_edges = ((PyArrayObject *)values[4]);
    __pyx_v_levels = ((PyArrayObject *)values[5]);
    __pyx_v_mask = ((PyArrayObject *)values[6]);
    if (values[7]) {
      __pyx_v_min_level = __Pyx_PyInt_AsInt(values[7]); if (unlikely((__pyx_v_min_level == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {
      __pyx_v_min_level = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_box_grids_below_level", 0, 7, 8, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.get_box_grids_below_level", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edge), __pyx_ptype_5numpy_ndarray, 1, "left_edge", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 251; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edge), __pyx_ptype_5numpy_ndarray, 1, "right_edge", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 252; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_14get_box_grids_below_level(__pyx_self, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels, __pyx_v_mask, __pyx_v_min_level);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":250
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_box_grids_below_level(             # <<<<<<<<<<<<<<
 *                         np.ndarray[np.float64_t, ndim=1] left_edge,
 *                         np.ndarray[np.float64_t, ndim=1] right_edge,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_14get_box_grids_below_level(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_left_edge, PyArrayObject *__pyx_v_right_edge, int __pyx_v_level, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels, PyArrayObject *__pyx_v_mask, int __pyx_v_min_level) {
  int __pyx_v_i;
  int __pyx_v_n;
  int __pyx_v_nx;
  int __pyx_v_inside;
  __pyx_t_5numpy_float64_t __pyx_v_eps;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edge;
  __Pyx_Buffer __pyx_pybuffer_left_edge;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edge;
  __Pyx_Buffer __pyx_pybuffer_right_edge;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  long __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  long __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_box_grids_below_level", 0);
  __pyx_pybuffer_left_edge.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edge.refcount = 0;
  __pyx_pybuffernd_left_edge.data = NULL;
  __pyx_pybuffernd_left_edge.rcbuffer = &__pyx_pybuffer_left_edge;
  __pyx_pybuffer_right_edge.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edge.refcount = 0;
  __pyx_pybuffernd_right_edge.data = NULL;
  __pyx_pybuffernd_right_edge.rcbuffer = &__pyx_pybuffer_right_edge;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edge.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edge, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_left_edge.diminfo[0].strides = __pyx_pybuffernd_left_edge.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edge.diminfo[0].shape = __pyx_pybuffernd_left_edge.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edge.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edge, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_right_edge.diminfo[0].strides = __pyx_pybuffernd_right_edge.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edge.diminfo[0].shape = __pyx_pybuffernd_right_edge.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "yt/utilities/lib/misc_utilities.pyx":260
 *                         int min_level = 0):
 *     cdef int i, n
 *     cdef int nx = left_edges.shape[0]             # <<<<<<<<<<<<<<
 *     cdef int inside
 *     cdef np.float64_t eps = np.finfo(np.float64).eps
 */
  __pyx_v_nx = (__pyx_v_left_edges->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":262
 *     cdef int nx = left_edges.shape[0]
 *     cdef int inside
 *     cdef np.float64_t eps = np.finfo(np.float64).eps             # <<<<<<<<<<<<<<
 *     for i in range(nx):
 *         mask[i] = 0
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__finfo); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__float64); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__eps); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_eps = __pyx_t_4;

  /* "yt/utilities/lib/misc_utilities.pyx":263
 *     cdef int inside
 *     cdef np.float64_t eps = np.finfo(np.float64).eps
 *     for i in range(nx):             # <<<<<<<<<<<<<<
 *         mask[i] = 0
 *         if levels[i,0] <= level and levels[i,0] >= min_level:
 */
  __pyx_t_5 = __pyx_v_nx;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/utilities/lib/misc_utilities.pyx":264
 *     cdef np.float64_t eps = np.finfo(np.float64).eps
 *     for i in range(nx):
 *         mask[i] = 0             # <<<<<<<<<<<<<<
 *         if levels[i,0] <= level and levels[i,0] >= min_level:
 *             inside = 1
 */
    __pyx_t_7 = __pyx_v_i;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_mask.diminfo[0].strides) = 0;

    /* "yt/utilities/lib/misc_utilities.pyx":265
 *     for i in range(nx):
 *         mask[i] = 0
 *         if levels[i,0] <= level and levels[i,0] >= min_level:             # <<<<<<<<<<<<<<
 *             inside = 1
 *             for n in range(3):
 */
    __pyx_t_8 = __pyx_v_i;
    __pyx_t_9 = 0;
    __pyx_t_10 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_levels.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_levels.diminfo[0].strides, __pyx_t_9, __pyx_pybuffernd_levels.diminfo[1].strides)) <= __pyx_v_level);
    if (__pyx_t_10) {
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_12 = 0;
      __pyx_t_13 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_levels.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_levels.diminfo[0].strides, __pyx_t_12, __pyx_pybuffernd_levels.diminfo[1].strides)) >= __pyx_v_min_level);
      __pyx_t_14 = __pyx_t_13;
    } else {
      __pyx_t_14 = __pyx_t_10;
    }
    if (__pyx_t_14) {

      /* "yt/utilities/lib/misc_utilities.pyx":266
 *         mask[i] = 0
 *         if levels[i,0] <= level and levels[i,0] >= min_level:
 *             inside = 1             # <<<<<<<<<<<<<<
 *             for n in range(3):
 *                 if (right_edges[i,n] - left_edge[n]) <= eps or \
 */
      __pyx_v_inside = 1;

      /* "yt/utilities/lib/misc_utilities.pyx":267
 *         if levels[i,0] <= level and levels[i,0] >= min_level:
 *             inside = 1
 *             for n in range(3):             # <<<<<<<<<<<<<<
 *                 if (right_edges[i,n] - left_edge[n]) <= eps or \
 *                    (right_edge[n] - left_edges[i,n]) <= eps:
 */
      for (__pyx_t_15 = 0; __pyx_t_15 < 3; __pyx_t_15+=1) {
        __pyx_v_n = __pyx_t_15;

        /* "yt/utilities/lib/misc_utilities.pyx":268
 *             inside = 1
 *             for n in range(3):
 *                 if (right_edges[i,n] - left_edge[n]) <= eps or \             # <<<<<<<<<<<<<<
 *                    (right_edge[n] - left_edges[i,n]) <= eps:
 *                     inside = 0
 */
        __pyx_t_16 = __pyx_v_i;
        __pyx_t_17 = __pyx_v_n;
        __pyx_t_18 = __pyx_v_n;
        __pyx_t_14 = (((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_right_edges.diminfo[1].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edge.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_left_edge.diminfo[0].strides))) <= __pyx_v_eps);
        if (!__pyx_t_14) {

          /* "yt/utilities/lib/misc_utilities.pyx":269
 *             for n in range(3):
 *                 if (right_edges[i,n] - left_edge[n]) <= eps or \
 *                    (right_edge[n] - left_edges[i,n]) <= eps:             # <<<<<<<<<<<<<<
 *                     inside = 0
 *                     break
 */
          __pyx_t_19 = __pyx_v_n;
          __pyx_t_20 = __pyx_v_i;
          __pyx_t_21 = __pyx_v_n;
          __pyx_t_10 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edge.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_right_edge.diminfo[0].strides)) - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_left_edges.diminfo[1].strides))) <= __pyx_v_eps);
          __pyx_t_13 = __pyx_t_10;
        } else {
          __pyx_t_13 = __pyx_t_14;
        }
        if (__pyx_t_13) {

          /* "yt/utilities/lib/misc_utilities.pyx":270
 *                 if (right_edges[i,n] - left_edge[n]) <= eps or \
 *                    (right_edge[n] - left_edges[i,n]) <= eps:
 *                     inside = 0             # <<<<<<<<<<<<<<
 *                     break
 *             if inside == 1: mask[i] = 1
 */
          __pyx_v_inside = 0;

          /* "yt/utilities/lib/misc_utilities.pyx":271
 *                    (right_edge[n] - left_edges[i,n]) <= eps:
 *                     inside = 0
 *                     break             # <<<<<<<<<<<<<<
 *             if inside == 1: mask[i] = 1
 * 
 */
          goto __pyx_L7_break;
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
      __pyx_L7_break:;

      /* "yt/utilities/lib/misc_utilities.pyx":272
 *                     inside = 0
 *                     break
 *             if inside == 1: mask[i] = 1             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
      __pyx_t_13 = (__pyx_v_inside == 1);
      if (__pyx_t_13) {
        __pyx_t_15 = __pyx_v_i;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_mask.diminfo[0].strides) = 1;
        goto __pyx_L9;
      }
      __pyx_L9:;
      goto __pyx_L5;
    }
    __pyx_L5:;
  }

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edge.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edge.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.get_box_grids_below_level", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edge.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edge.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_17find_values_at_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_17find_values_at_point = {__Pyx_NAMESTR("find_values_at_point"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_17find_values_at_point, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_17find_values_at_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_point = 0;
  PyArrayObject *__pyx_v_left_edges = 0;
  PyArrayObject *__pyx_v_right_edges = 0;
  PyArrayObject *__pyx_v_dimensions = 0;
  PyObject *__pyx_v_field_names = 0;
  PyObject *__pyx_v_grid_objects = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_values_at_point (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__point,&__pyx_n_s__left_edges,&__pyx_n_s__right_edges,&__pyx_n_s__dimensions,&__pyx_n_s__field_names,&__pyx_n_s__grid_objects,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__point)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__left_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_values_at_point", 1, 6, 6, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_values_at_point", 1, 6, 6, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__dimensions)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_values_at_point", 1, 6, 6, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__field_names)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_values_at_point", 1, 6, 6, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__grid_objects)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_values_at_point", 1, 6, 6, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_values_at_point") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_point = ((PyArrayObject *)values[0]);
    __pyx_v_left_edges = ((PyArrayObject *)values[1]);
    __pyx_v_right_edges = ((PyArrayObject *)values[2]);
    __pyx_v_dimensions = ((PyArrayObject *)values[3]);
    __pyx_v_field_names = values[4];
    __pyx_v_grid_objects = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_values_at_point", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.find_values_at_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_point), __pyx_ptype_5numpy_ndarray, 1, "point", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dimensions), __pyx_ptype_5numpy_ndarray, 1, "dimensions", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_16find_values_at_point(__pyx_self, __pyx_v_point, __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_dimensions, __pyx_v_field_names, __pyx_v_grid_objects);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":277
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def find_values_at_point(np.ndarray[np.float64_t, ndim=1] point,             # <<<<<<<<<<<<<<
 *                          np.ndarray[np.float64_t, ndim=2] left_edges,
 *                          np.ndarray[np.float64_t, ndim=2] right_edges,
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_16find_values_at_point(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_point, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_dimensions, PyObject *__pyx_v_field_names, PyObject *__pyx_v_grid_objects) {
  int __pyx_v_ind[3];
  int __pyx_v_gi;
  int __pyx_v_fi;
  int __pyx_v_nf;
  __pyx_t_5numpy_float64_t __pyx_v_dds;
  PyArrayObject *__pyx_v_field = 0;
  PyArrayObject *__pyx_v_rv = 0;
  PyObject *__pyx_v_grid = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dimensions;
  __Pyx_Buffer __pyx_pybuffer_dimensions;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_field;
  __Pyx_Buffer __pyx_pybuffer_field;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_point;
  __Pyx_Buffer __pyx_pybuffer_point;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_rv;
  __Pyx_Buffer __pyx_pybuffer_rv;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  npy_intp __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  long __pyx_t_10;
  long __pyx_t_11;
  __pyx_t_5numpy_float64_t __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  long __pyx_t_15;
  int __pyx_t_16;
  long __pyx_t_17;
  long __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  long __pyx_t_21;
  int __pyx_t_22;
  long __pyx_t_23;
  long __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  long __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  int __pyx_t_32;
  int __pyx_t_33;
  int __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  PyArrayObject *__pyx_t_40 = NULL;
  int __pyx_t_41;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  int __pyx_t_45;
  int __pyx_t_46;
  int __pyx_t_47;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_values_at_point", 0);
  __pyx_pybuffer_field.pybuffer.buf = NULL;
  __pyx_pybuffer_field.refcount = 0;
  __pyx_pybuffernd_field.data = NULL;
  __pyx_pybuffernd_field.rcbuffer = &__pyx_pybuffer_field;
  __pyx_pybuffer_rv.pybuffer.buf = NULL;
  __pyx_pybuffer_rv.refcount = 0;
  __pyx_pybuffernd_rv.data = NULL;
  __pyx_pybuffernd_rv.rcbuffer = &__pyx_pybuffer_rv;
  __pyx_pybuffer_point.pybuffer.buf = NULL;
  __pyx_pybuffer_point.refcount = 0;
  __pyx_pybuffernd_point.data = NULL;
  __pyx_pybuffernd_point.rcbuffer = &__pyx_pybuffer_point;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_dimensions.pybuffer.buf = NULL;
  __pyx_pybuffer_dimensions.refcount = 0;
  __pyx_pybuffernd_dimensions.data = NULL;
  __pyx_pybuffernd_dimensions.rcbuffer = &__pyx_pybuffer_dimensions;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_point.rcbuffer->pybuffer, (PyObject*)__pyx_v_point, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_point.diminfo[0].strides = __pyx_pybuffernd_point.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_point.diminfo[0].shape = __pyx_pybuffernd_point.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dimensions.rcbuffer->pybuffer, (PyObject*)__pyx_v_dimensions, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_dimensions.diminfo[0].strides = __pyx_pybuffernd_dimensions.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dimensions.diminfo[0].shape = __pyx_pybuffernd_dimensions.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_dimensions.diminfo[1].strides = __pyx_pybuffernd_dimensions.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_dimensions.diminfo[1].shape = __pyx_pybuffernd_dimensions.rcbuffer->pybuffer.shape[1];

  /* "yt/utilities/lib/misc_utilities.pyx":288
 *     # strings and whatnot, but they should not be terribly slow.
 *     cdef int ind[3], gi, fi
 *     cdef int nf = len(field_names)             # <<<<<<<<<<<<<<
 *     cdef np.float64_t dds
 *     cdef np.ndarray[np.float64_t, ndim=3] field
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_field_names); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 288; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_nf = __pyx_t_1;

  /* "yt/utilities/lib/misc_utilities.pyx":291
 *     cdef np.float64_t dds
 *     cdef np.ndarray[np.float64_t, ndim=3] field
 *     cdef np.ndarray[np.float64_t, ndim=1] rv = np.zeros(nf, dtype='float64')             # <<<<<<<<<<<<<<
 *     for gi in range(left_edges.shape[0]):
 *         if not ((left_edges[gi,0] < point[0] < right_edges[gi,0])
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromLong(__pyx_v_nf); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__float64)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_5 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_rv.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_rv = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_rv.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_rv.diminfo[0].strides = __pyx_pybuffernd_rv.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_rv.diminfo[0].shape = __pyx_pybuffernd_rv.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_rv = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":292
 *     cdef np.ndarray[np.float64_t, ndim=3] field
 *     cdef np.ndarray[np.float64_t, ndim=1] rv = np.zeros(nf, dtype='float64')
 *     for gi in range(left_edges.shape[0]):             # <<<<<<<<<<<<<<
 *         if not ((left_edges[gi,0] < point[0] < right_edges[gi,0])
 *             and (left_edges[gi,1] < point[1] < right_edges[gi,1])
 */
  __pyx_t_7 = (__pyx_v_left_edges->dimensions[0]);
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_gi = __pyx_t_8;

    /* "yt/utilities/lib/misc_utilities.pyx":293
 *     cdef np.ndarray[np.float64_t, ndim=1] rv = np.zeros(nf, dtype='float64')
 *     for gi in range(left_edges.shape[0]):
 *         if not ((left_edges[gi,0] < point[0] < right_edges[gi,0])             # <<<<<<<<<<<<<<
 *             and (left_edges[gi,1] < point[1] < right_edges[gi,1])
 *             and (left_edges[gi,2] < point[2] < right_edges[gi,2])):
 */
    __pyx_t_9 = __pyx_v_gi;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_point.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_point.diminfo[0].strides));
    __pyx_t_13 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_10, __pyx_pybuffernd_left_edges.diminfo[1].strides)) < __pyx_t_12);
    if (__pyx_t_13) {
      __pyx_t_14 = __pyx_v_gi;
      __pyx_t_15 = 0;
      __pyx_t_13 = (__pyx_t_12 < (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_right_edges.diminfo[1].strides)));
    }
    if (__pyx_t_13) {

      /* "yt/utilities/lib/misc_utilities.pyx":294
 *     for gi in range(left_edges.shape[0]):
 *         if not ((left_edges[gi,0] < point[0] < right_edges[gi,0])
 *             and (left_edges[gi,1] < point[1] < right_edges[gi,1])             # <<<<<<<<<<<<<<
 *             and (left_edges[gi,2] < point[2] < right_edges[gi,2])):
 *             continue
 */
      __pyx_t_16 = __pyx_v_gi;
      __pyx_t_17 = 1;
      __pyx_t_18 = 1;
      __pyx_t_12 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_point.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_point.diminfo[0].strides));
      __pyx_t_19 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_left_edges.diminfo[1].strides)) < __pyx_t_12);
      if (__pyx_t_19) {
        __pyx_t_20 = __pyx_v_gi;
        __pyx_t_21 = 1;
        __pyx_t_19 = (__pyx_t_12 < (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_right_edges.diminfo[1].strides)));
      }
      if (__pyx_t_19) {

        /* "yt/utilities/lib/misc_utilities.pyx":295
 *         if not ((left_edges[gi,0] < point[0] < right_edges[gi,0])
 *             and (left_edges[gi,1] < point[1] < right_edges[gi,1])
 *             and (left_edges[gi,2] < point[2] < right_edges[gi,2])):             # <<<<<<<<<<<<<<
 *             continue
 *         # We found our grid!
 */
        __pyx_t_22 = __pyx_v_gi;
        __pyx_t_23 = 2;
        __pyx_t_24 = 2;
        __pyx_t_12 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_point.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_point.diminfo[0].strides));
        __pyx_t_25 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_23, __pyx_pybuffernd_left_edges.diminfo[1].strides)) < __pyx_t_12);
        if (__pyx_t_25) {
          __pyx_t_26 = __pyx_v_gi;
          __pyx_t_27 = 2;
          __pyx_t_25 = (__pyx_t_12 < (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_right_edges.diminfo[1].strides)));
        }
        __pyx_t_28 = __pyx_t_25;
      } else {
        __pyx_t_28 = __pyx_t_19;
      }
      __pyx_t_19 = __pyx_t_28;
    } else {
      __pyx_t_19 = __pyx_t_13;
    }
    __pyx_t_13 = (!__pyx_t_19);
    if (__pyx_t_13) {

      /* "yt/utilities/lib/misc_utilities.pyx":296
 *             and (left_edges[gi,1] < point[1] < right_edges[gi,1])
 *             and (left_edges[gi,2] < point[2] < right_edges[gi,2])):
 *             continue             # <<<<<<<<<<<<<<
 *         # We found our grid!
 *         for fi in range(3):
 */
      goto __pyx_L3_continue;
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "yt/utilities/lib/misc_utilities.pyx":298
 *             continue
 *         # We found our grid!
 *         for fi in range(3):             # <<<<<<<<<<<<<<
 *             dds = ((right_edges[gi,fi] - left_edges[gi,fi])/
 *                    (<np.float64_t> dimensions[gi,fi]))
 */
    for (__pyx_t_29 = 0; __pyx_t_29 < 3; __pyx_t_29+=1) {
      __pyx_v_fi = __pyx_t_29;

      /* "yt/utilities/lib/misc_utilities.pyx":299
 *         # We found our grid!
 *         for fi in range(3):
 *             dds = ((right_edges[gi,fi] - left_edges[gi,fi])/             # <<<<<<<<<<<<<<
 *                    (<np.float64_t> dimensions[gi,fi]))
 *             ind[fi] = <int> ((point[fi] - left_edges[gi,fi])/dds)
 */
      __pyx_t_30 = __pyx_v_gi;
      __pyx_t_31 = __pyx_v_fi;
      __pyx_t_32 = __pyx_v_gi;
      __pyx_t_33 = __pyx_v_fi;

      /* "yt/utilities/lib/misc_utilities.pyx":300
 *         for fi in range(3):
 *             dds = ((right_edges[gi,fi] - left_edges[gi,fi])/
 *                    (<np.float64_t> dimensions[gi,fi]))             # <<<<<<<<<<<<<<
 *             ind[fi] = <int> ((point[fi] - left_edges[gi,fi])/dds)
 *         grid = grid_objects[gi]
 */
      __pyx_t_34 = __pyx_v_gi;
      __pyx_t_35 = __pyx_v_fi;
      __pyx_v_dds = (((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_31, __pyx_pybuffernd_right_edges.diminfo[1].strides)) - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_33, __pyx_pybuffernd_left_edges.diminfo[1].strides))) / ((__pyx_t_5numpy_float64_t)(*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_dimensions.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_dimensions.diminfo[0].strides, __pyx_t_35, __pyx_pybuffernd_dimensions.diminfo[1].strides))));

      /* "yt/utilities/lib/misc_utilities.pyx":301
 *             dds = ((right_edges[gi,fi] - left_edges[gi,fi])/
 *                    (<np.float64_t> dimensions[gi,fi]))
 *             ind[fi] = <int> ((point[fi] - left_edges[gi,fi])/dds)             # <<<<<<<<<<<<<<
 *         grid = grid_objects[gi]
 *         for fi in range(nf):
 */
      __pyx_t_36 = __pyx_v_fi;
      __pyx_t_37 = __pyx_v_gi;
      __pyx_t_38 = __pyx_v_fi;
      (__pyx_v_ind[__pyx_v_fi]) = ((int)(((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_point.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_point.diminfo[0].strides)) - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_38, __pyx_pybuffernd_left_edges.diminfo[1].strides))) / __pyx_v_dds));
    }

    /* "yt/utilities/lib/misc_utilities.pyx":302
 *                    (<np.float64_t> dimensions[gi,fi]))
 *             ind[fi] = <int> ((point[fi] - left_edges[gi,fi])/dds)
 *         grid = grid_objects[gi]             # <<<<<<<<<<<<<<
 *         for fi in range(nf):
 *             field = grid[field_names[fi]]
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_grid_objects, __pyx_v_gi, sizeof(int), PyInt_FromLong); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 302; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_grid = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "yt/utilities/lib/misc_utilities.pyx":303
 *             ind[fi] = <int> ((point[fi] - left_edges[gi,fi])/dds)
 *         grid = grid_objects[gi]
 *         for fi in range(nf):             # <<<<<<<<<<<<<<
 *             field = grid[field_names[fi]]
 *             rv[fi] = field[ind[0], ind[1], ind[2]]
 */
    __pyx_t_29 = __pyx_v_nf;
    for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_29; __pyx_t_39+=1) {
      __pyx_v_fi = __pyx_t_39;

      /* "yt/utilities/lib/misc_utilities.pyx":304
 *         grid = grid_objects[gi]
 *         for fi in range(nf):
 *             field = grid[field_names[fi]]             # <<<<<<<<<<<<<<
 *             rv[fi] = field[ind[0], ind[1], ind[2]]
 *         return rv
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_field_names, __pyx_v_fi, sizeof(int), PyInt_FromLong); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 304; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyObject_GetItem(__pyx_v_grid, __pyx_t_5); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 304; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 304; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_40 = ((PyArrayObject *)__pyx_t_2);
      {
        __Pyx_BufFmt_StackElem __pyx_stack[1];
        __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_field.rcbuffer->pybuffer);
        __pyx_t_41 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_field.rcbuffer->pybuffer, (PyObject*)__pyx_t_40, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
        if (unlikely(__pyx_t_41 < 0)) {
          PyErr_Fetch(&__pyx_t_42, &__pyx_t_43, &__pyx_t_44);
          if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_field.rcbuffer->pybuffer, (PyObject*)__pyx_v_field, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
            Py_XDECREF(__pyx_t_42); Py_XDECREF(__pyx_t_43); Py_XDECREF(__pyx_t_44);
            __Pyx_RaiseBufferFallbackError();
          } else {
            PyErr_Restore(__pyx_t_42, __pyx_t_43, __pyx_t_44);
          }
        }
        __pyx_pybuffernd_field.diminfo[0].strides = __pyx_pybuffernd_field.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_field.diminfo[0].shape = __pyx_pybuffernd_field.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_field.diminfo[1].strides = __pyx_pybuffernd_field.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_field.diminfo[1].shape = __pyx_pybuffernd_field.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_field.diminfo[2].strides = __pyx_pybuffernd_field.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_field.diminfo[2].shape = __pyx_pybuffernd_field.rcbuffer->pybuffer.shape[2];
        if (unlikely(__pyx_t_41 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 304; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_t_40 = 0;
      __Pyx_XDECREF(((PyObject *)__pyx_v_field));
      __pyx_v_field = ((PyArrayObject *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "yt/utilities/lib/misc_utilities.pyx":305
 *         for fi in range(nf):
 *             field = grid[field_names[fi]]
 *             rv[fi] = field[ind[0], ind[1], ind[2]]             # <<<<<<<<<<<<<<
 *         return rv
 *     raise KeyError
 */
      __pyx_t_41 = (__pyx_v_ind[0]);
      __pyx_t_45 = (__pyx_v_ind[1]);
      __pyx_t_46 = (__pyx_v_ind[2]);
      __pyx_t_47 = __pyx_v_fi;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rv.rcbuffer->pybuffer.buf, __pyx_t_47, __pyx_pybuffernd_rv.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_field.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_field.diminfo[0].strides, __pyx_t_45, __pyx_pybuffernd_field.diminfo[1].strides, __pyx_t_46, __pyx_pybuffernd_field.diminfo[2].strides));
    }

    /* "yt/utilities/lib/misc_utilities.pyx":306
 *             field = grid[field_names[fi]]
 *             rv[fi] = field[ind[0], ind[1], ind[2]]
 *         return rv             # <<<<<<<<<<<<<<
 *     raise KeyError
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_rv));
    __pyx_r = ((PyObject *)__pyx_v_rv);
    goto __pyx_L0;
    __pyx_L3_continue:;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":307
 *             rv[fi] = field[ind[0], ind[1], ind[2]]
 *         return rv
 *     raise KeyError             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_Raise(__pyx_builtin_KeyError, 0, 0, 0);
  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dimensions.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_field.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_point.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rv.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.find_values_at_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dimensions.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_field.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_point.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rv.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF((PyObject *)__pyx_v_rv);
  __Pyx_XDECREF(__pyx_v_grid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_19kdtree_get_choices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_19kdtree_get_choices = {__Pyx_NAMESTR("kdtree_get_choices"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_19kdtree_get_choices, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_19kdtree_get_choices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_data = 0;
  PyArrayObject *__pyx_v_l_corner = 0;
  PyArrayObject *__pyx_v_r_corner = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("kdtree_get_choices (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__data,&__pyx_n_s__l_corner,&__pyx_n_s__r_corner,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__l_corner)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("kdtree_get_choices", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__r_corner)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("kdtree_get_choices", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "kdtree_get_choices") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_data = ((PyArrayObject *)values[0]);
    __pyx_v_l_corner = ((PyArrayObject *)values[1]);
    __pyx_v_r_corner = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("kdtree_get_choices", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.kdtree_get_choices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), __pyx_ptype_5numpy_ndarray, 1, "data", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_l_corner), __pyx_ptype_5numpy_ndarray, 1, "l_corner", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_r_corner), __pyx_ptype_5numpy_ndarray, 1, "r_corner", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 314; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_18kdtree_get_choices(__pyx_self, __pyx_v_data, __pyx_v_l_corner, __pyx_v_r_corner);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":312
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def kdtree_get_choices(np.ndarray[np.float64_t, ndim=3] data,             # <<<<<<<<<<<<<<
 *                        np.ndarray[np.float64_t, ndim=1] l_corner,
 *                        np.ndarray[np.float64_t, ndim=1] r_corner):
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_18kdtree_get_choices(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_data, PyArrayObject *__pyx_v_l_corner, PyArrayObject *__pyx_v_r_corner) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_dim;
  int __pyx_v_n_unique;
  int __pyx_v_best_dim;
  int __pyx_v_n_grids;
  int __pyx_v_my_split;
  __pyx_t_5numpy_float64_t **__pyx_v_uniquedims;
  __pyx_t_5numpy_float64_t *__pyx_v_uniques;
  __pyx_t_5numpy_float64_t __pyx_v_split;
  long __pyx_v_my_max;
  long __pyx_v_skipit;
  PyArrayObject *__pyx_v_tarr = 0;
  PyArrayObject *__pyx_v_less_ids = 0;
  PyArrayObject *__pyx_v_greater_ids = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_data;
  __Pyx_Buffer __pyx_pybuffer_data;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_greater_ids;
  __Pyx_Buffer __pyx_pybuffer_greater_ids;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_l_corner;
  __Pyx_Buffer __pyx_pybuffer_l_corner;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_less_ids;
  __Pyx_Buffer __pyx_pybuffer_less_ids;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_r_corner;
  __Pyx_Buffer __pyx_pybuffer_r_corner;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tarr;
  __Pyx_Buffer __pyx_pybuffer_tarr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyArrayObject *__pyx_t_26 = NULL;
  long __pyx_t_27;
  PyArrayObject *__pyx_t_28 = NULL;
  PyArrayObject *__pyx_t_29 = NULL;
  int __pyx_t_30;
  int __pyx_t_31;
  int __pyx_t_32;
  int __pyx_t_33;
  int __pyx_t_34;
  long __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_t_38;
  PyObject *__pyx_t_39 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("kdtree_get_choices", 0);
  __pyx_pybuffer_tarr.pybuffer.buf = NULL;
  __pyx_pybuffer_tarr.refcount = 0;
  __pyx_pybuffernd_tarr.data = NULL;
  __pyx_pybuffernd_tarr.rcbuffer = &__pyx_pybuffer_tarr;
  __pyx_pybuffer_less_ids.pybuffer.buf = NULL;
  __pyx_pybuffer_less_ids.refcount = 0;
  __pyx_pybuffernd_less_ids.data = NULL;
  __pyx_pybuffernd_less_ids.rcbuffer = &__pyx_pybuffer_less_ids;
  __pyx_pybuffer_greater_ids.pybuffer.buf = NULL;
  __pyx_pybuffer_greater_ids.refcount = 0;
  __pyx_pybuffernd_greater_ids.data = NULL;
  __pyx_pybuffernd_greater_ids.rcbuffer = &__pyx_pybuffer_greater_ids;
  __pyx_pybuffer_data.pybuffer.buf = NULL;
  __pyx_pybuffer_data.refcount = 0;
  __pyx_pybuffernd_data.data = NULL;
  __pyx_pybuffernd_data.rcbuffer = &__pyx_pybuffer_data;
  __pyx_pybuffer_l_corner.pybuffer.buf = NULL;
  __pyx_pybuffer_l_corner.refcount = 0;
  __pyx_pybuffernd_l_corner.data = NULL;
  __pyx_pybuffernd_l_corner.rcbuffer = &__pyx_pybuffer_l_corner;
  __pyx_pybuffer_r_corner.pybuffer.buf = NULL;
  __pyx_pybuffer_r_corner.refcount = 0;
  __pyx_pybuffernd_r_corner.data = NULL;
  __pyx_pybuffernd_r_corner.rcbuffer = &__pyx_pybuffer_r_corner;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_data.rcbuffer->pybuffer, (PyObject*)__pyx_v_data, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_data.diminfo[0].strides = __pyx_pybuffernd_data.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_data.diminfo[0].shape = __pyx_pybuffernd_data.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_data.diminfo[1].strides = __pyx_pybuffernd_data.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_data.diminfo[1].shape = __pyx_pybuffernd_data.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_data.diminfo[2].strides = __pyx_pybuffernd_data.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_data.diminfo[2].shape = __pyx_pybuffernd_data.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_l_corner.rcbuffer->pybuffer, (PyObject*)__pyx_v_l_corner, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_l_corner.diminfo[0].strides = __pyx_pybuffernd_l_corner.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_l_corner.diminfo[0].shape = __pyx_pybuffernd_l_corner.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_r_corner.rcbuffer->pybuffer, (PyObject*)__pyx_v_r_corner, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_r_corner.diminfo[0].strides = __pyx_pybuffernd_r_corner.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_r_corner.diminfo[0].shape = __pyx_pybuffernd_r_corner.rcbuffer->pybuffer.shape[0];

  /* "yt/utilities/lib/misc_utilities.pyx":316
 *                        np.ndarray[np.float64_t, ndim=1] r_corner):
 *     cdef int i, j, k, dim, n_unique, best_dim, n_best, n_grids, addit, my_split
 *     n_grids = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef np.float64_t **uniquedims, *uniques, split
 *     uniquedims = <np.float64_t **> alloca(3 * sizeof(np.float64_t*))
 */
  __pyx_v_n_grids = (__pyx_v_data->dimensions[0]);

  /* "yt/utilities/lib/misc_utilities.pyx":318
 *     n_grids = data.shape[0]
 *     cdef np.float64_t **uniquedims, *uniques, split
 *     uniquedims = <np.float64_t **> alloca(3 * sizeof(np.float64_t*))             # <<<<<<<<<<<<<<
 *     for i in range(3):
 *         uniquedims[i] = <np.float64_t *> \
 */
  __pyx_v_uniquedims = ((__pyx_t_5numpy_float64_t **)alloca((3 * (sizeof(__pyx_t_5numpy_float64_t *)))));

  /* "yt/utilities/lib/misc_utilities.pyx":319
 *     cdef np.float64_t **uniquedims, *uniques, split
 *     uniquedims = <np.float64_t **> alloca(3 * sizeof(np.float64_t*))
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         uniquedims[i] = <np.float64_t *> \
 *                 alloca(2*n_grids * sizeof(np.float64_t))
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/misc_utilities.pyx":320
 *     uniquedims = <np.float64_t **> alloca(3 * sizeof(np.float64_t*))
 *     for i in range(3):
 *         uniquedims[i] = <np.float64_t *> \             # <<<<<<<<<<<<<<
 *                 alloca(2*n_grids * sizeof(np.float64_t))
 *     my_max = 0
 */
    (__pyx_v_uniquedims[__pyx_v_i]) = ((__pyx_t_5numpy_float64_t *)alloca(((2 * __pyx_v_n_grids) * (sizeof(__pyx_t_5numpy_float64_t)))));
  }

  /* "yt/utilities/lib/misc_utilities.pyx":322
 *         uniquedims[i] = <np.float64_t *> \
 *                 alloca(2*n_grids * sizeof(np.float64_t))
 *     my_max = 0             # <<<<<<<<<<<<<<
 *     best_dim = -1
 *     for dim in range(3):
 */
  __pyx_v_my_max = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":323
 *                 alloca(2*n_grids * sizeof(np.float64_t))
 *     my_max = 0
 *     best_dim = -1             # <<<<<<<<<<<<<<
 *     for dim in range(3):
 *         n_unique = 0
 */
  __pyx_v_best_dim = -1;

  /* "yt/utilities/lib/misc_utilities.pyx":324
 *     my_max = 0
 *     best_dim = -1
 *     for dim in range(3):             # <<<<<<<<<<<<<<
 *         n_unique = 0
 *         uniques = uniquedims[dim]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_dim = __pyx_t_1;

    /* "yt/utilities/lib/misc_utilities.pyx":325
 *     best_dim = -1
 *     for dim in range(3):
 *         n_unique = 0             # <<<<<<<<<<<<<<
 *         uniques = uniquedims[dim]
 *         for i in range(n_grids):
 */
    __pyx_v_n_unique = 0;

    /* "yt/utilities/lib/misc_utilities.pyx":326
 *     for dim in range(3):
 *         n_unique = 0
 *         uniques = uniquedims[dim]             # <<<<<<<<<<<<<<
 *         for i in range(n_grids):
 *             # Check for disqualification
 */
    __pyx_v_uniques = (__pyx_v_uniquedims[__pyx_v_dim]);

    /* "yt/utilities/lib/misc_utilities.pyx":327
 *         n_unique = 0
 *         uniques = uniquedims[dim]
 *         for i in range(n_grids):             # <<<<<<<<<<<<<<
 *             # Check for disqualification
 *             for j in range(2):
 */
    __pyx_t_2 = __pyx_v_n_grids;
    for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "yt/utilities/lib/misc_utilities.pyx":329
 *         for i in range(n_grids):
 *             # Check for disqualification
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 #print "Checking against", i,j,dim,data[i,j,dim]
 *                 if not (l_corner[dim] < data[i, j, dim] and
 */
      for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
        __pyx_v_j = __pyx_t_4;

        /* "yt/utilities/lib/misc_utilities.pyx":331
 *             for j in range(2):
 *                 #print "Checking against", i,j,dim,data[i,j,dim]
 *                 if not (l_corner[dim] < data[i, j, dim] and             # <<<<<<<<<<<<<<
 *                         data[i, j, dim] < r_corner[dim]):
 *                     #print "Skipping ", data[i,j,dim]
 */
        __pyx_t_5 = __pyx_v_dim;
        __pyx_t_6 = __pyx_v_i;
        __pyx_t_7 = __pyx_v_j;
        __pyx_t_8 = __pyx_v_dim;
        __pyx_t_9 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_l_corner.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_l_corner.diminfo[0].strides)) < (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_data.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_data.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_data.diminfo[1].strides, __pyx_t_8, __pyx_pybuffernd_data.diminfo[2].strides)));
        if (__pyx_t_9) {

          /* "yt/utilities/lib/misc_utilities.pyx":332
 *                 #print "Checking against", i,j,dim,data[i,j,dim]
 *                 if not (l_corner[dim] < data[i, j, dim] and
 *                         data[i, j, dim] < r_corner[dim]):             # <<<<<<<<<<<<<<
 *                     #print "Skipping ", data[i,j,dim]
 *                     continue
 */
          __pyx_t_10 = __pyx_v_i;
          __pyx_t_11 = __pyx_v_j;
          __pyx_t_12 = __pyx_v_dim;
          __pyx_t_13 = __pyx_v_dim;
          __pyx_t_14 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_data.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_data.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_data.diminfo[1].strides, __pyx_t_12, __pyx_pybuffernd_data.diminfo[2].strides)) < (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_r_corner.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_r_corner.diminfo[0].strides)));
          __pyx_t_15 = __pyx_t_14;
        } else {
          __pyx_t_15 = __pyx_t_9;
        }
        __pyx_t_9 = (!__pyx_t_15);
        if (__pyx_t_9) {

          /* "yt/utilities/lib/misc_utilities.pyx":334
 *                         data[i, j, dim] < r_corner[dim]):
 *                     #print "Skipping ", data[i,j,dim]
 *                     continue             # <<<<<<<<<<<<<<
 *                 skipit = 0
 *                 # Add our left ...
 */
          goto __pyx_L9_continue;
          goto __pyx_L11;
        }
        __pyx_L11:;

        /* "yt/utilities/lib/misc_utilities.pyx":335
 *                     #print "Skipping ", data[i,j,dim]
 *                     continue
 *                 skipit = 0             # <<<<<<<<<<<<<<
 *                 # Add our left ...
 *                 for k in range(n_unique):
 */
        __pyx_v_skipit = 0;

        /* "yt/utilities/lib/misc_utilities.pyx":337
 *                 skipit = 0
 *                 # Add our left ...
 *                 for k in range(n_unique):             # <<<<<<<<<<<<<<
 *                     if uniques[k] == data[i, j, dim]:
 *                         skipit = 1
 */
        __pyx_t_16 = __pyx_v_n_unique;
        for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
          __pyx_v_k = __pyx_t_17;

          /* "yt/utilities/lib/misc_utilities.pyx":338
 *                 # Add our left ...
 *                 for k in range(n_unique):
 *                     if uniques[k] == data[i, j, dim]:             # <<<<<<<<<<<<<<
 *                         skipit = 1
 *                         #print "Identified", uniques[k], data[i,j,dim], n_unique
 */
          __pyx_t_18 = __pyx_v_i;
          __pyx_t_19 = __pyx_v_j;
          __pyx_t_20 = __pyx_v_dim;
          __pyx_t_9 = ((__pyx_v_uniques[__pyx_v_k]) == (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_data.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_data.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_data.diminfo[1].strides, __pyx_t_20, __pyx_pybuffernd_data.diminfo[2].strides)));
          if (__pyx_t_9) {

            /* "yt/utilities/lib/misc_utilities.pyx":339
 *                 for k in range(n_unique):
 *                     if uniques[k] == data[i, j, dim]:
 *                         skipit = 1             # <<<<<<<<<<<<<<
 *                         #print "Identified", uniques[k], data[i,j,dim], n_unique
 *                         break
 */
            __pyx_v_skipit = 1;

            /* "yt/utilities/lib/misc_utilities.pyx":341
 *                         skipit = 1
 *                         #print "Identified", uniques[k], data[i,j,dim], n_unique
 *                         break             # <<<<<<<<<<<<<<
 *                 if skipit == 0:
 *                     uniques[n_unique] = data[i, j, dim]
 */
            goto __pyx_L13_break;
            goto __pyx_L14;
          }
          __pyx_L14:;
        }
        __pyx_L13_break:;

        /* "yt/utilities/lib/misc_utilities.pyx":342
 *                         #print "Identified", uniques[k], data[i,j,dim], n_unique
 *                         break
 *                 if skipit == 0:             # <<<<<<<<<<<<<<
 *                     uniques[n_unique] = data[i, j, dim]
 *                     n_unique += 1
 */
        __pyx_t_9 = (__pyx_v_skipit == 0);
        if (__pyx_t_9) {

          /* "yt/utilities/lib/misc_utilities.pyx":343
 *                         break
 *                 if skipit == 0:
 *                     uniques[n_unique] = data[i, j, dim]             # <<<<<<<<<<<<<<
 *                     n_unique += 1
 *         if n_unique > my_max:
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_17 = __pyx_v_j;
          __pyx_t_21 = __pyx_v_dim;
          (__pyx_v_uniques[__pyx_v_n_unique]) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_data.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_data.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_data.diminfo[1].strides, __pyx_t_21, __pyx_pybuffernd_data.diminfo[2].strides));

          /* "yt/utilities/lib/misc_utilities.pyx":344
 *                 if skipit == 0:
 *                     uniques[n_unique] = data[i, j, dim]
 *                     n_unique += 1             # <<<<<<<<<<<<<<
 *         if n_unique > my_max:
 *             best_dim = dim
 */
          __pyx_v_n_unique = (__pyx_v_n_unique + 1);
          goto __pyx_L15;
        }
        __pyx_L15:;
        __pyx_L9_continue:;
      }
    }

    /* "yt/utilities/lib/misc_utilities.pyx":345
 *                     uniques[n_unique] = data[i, j, dim]
 *                     n_unique += 1
 *         if n_unique > my_max:             # <<<<<<<<<<<<<<
 *             best_dim = dim
 *             my_max = n_unique
 */
    __pyx_t_9 = (__pyx_v_n_unique > __pyx_v_my_max);
    if (__pyx_t_9) {

      /* "yt/utilities/lib/misc_utilities.pyx":346
 *                     n_unique += 1
 *         if n_unique > my_max:
 *             best_dim = dim             # <<<<<<<<<<<<<<
 *             my_max = n_unique
 *             my_split = (n_unique-1)/2
 */
      __pyx_v_best_dim = __pyx_v_dim;

      /* "yt/utilities/lib/misc_utilities.pyx":347
 *         if n_unique > my_max:
 *             best_dim = dim
 *             my_max = n_unique             # <<<<<<<<<<<<<<
 *             my_split = (n_unique-1)/2
 *     # I recognize how lame this is.
 */
      __pyx_v_my_max = __pyx_v_n_unique;

      /* "yt/utilities/lib/misc_utilities.pyx":348
 *             best_dim = dim
 *             my_max = n_unique
 *             my_split = (n_unique-1)/2             # <<<<<<<<<<<<<<
 *     # I recognize how lame this is.
 *     cdef np.ndarray[np.float64_t, ndim=1] tarr = np.empty(my_max, dtype='float64')
 */
      __pyx_v_my_split = ((__pyx_v_n_unique - 1) / 2);
      goto __pyx_L16;
    }
    __pyx_L16:;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":350
 *             my_split = (n_unique-1)/2
 *     # I recognize how lame this is.
 *     cdef np.ndarray[np.float64_t, ndim=1] tarr = np.empty(my_max, dtype='float64')             # <<<<<<<<<<<<<<
 *     for i in range(my_max):
 *         #print "Setting tarr: ", i, uniquedims[best_dim][i]
 */
  __pyx_t_22 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = PyObject_GetAttr(__pyx_t_22, __pyx_n_s__empty); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = PyInt_FromLong(__pyx_v_my_max); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  __pyx_t_22 = 0;
  __pyx_t_22 = PyDict_New(); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_22));
  if (PyDict_SetItem(__pyx_t_22, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__float64)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_25 = PyObject_Call(__pyx_t_23, ((PyObject *)__pyx_t_24), ((PyObject *)__pyx_t_22)); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_24)); __pyx_t_24 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  if (!(likely(((__pyx_t_25) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_25, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_26 = ((PyArrayObject *)__pyx_t_25);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tarr.rcbuffer->pybuffer, (PyObject*)__pyx_t_26, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_tarr = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tarr.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_tarr.diminfo[0].strides = __pyx_pybuffernd_tarr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tarr.diminfo[0].shape = __pyx_pybuffernd_tarr.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_26 = 0;
  __pyx_v_tarr = ((PyArrayObject *)__pyx_t_25);
  __pyx_t_25 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":351
 *     # I recognize how lame this is.
 *     cdef np.ndarray[np.float64_t, ndim=1] tarr = np.empty(my_max, dtype='float64')
 *     for i in range(my_max):             # <<<<<<<<<<<<<<
 *         #print "Setting tarr: ", i, uniquedims[best_dim][i]
 *         tarr[i] = uniquedims[best_dim][i]
 */
  __pyx_t_27 = __pyx_v_my_max;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_27; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/misc_utilities.pyx":353
 *     for i in range(my_max):
 *         #print "Setting tarr: ", i, uniquedims[best_dim][i]
 *         tarr[i] = uniquedims[best_dim][i]             # <<<<<<<<<<<<<<
 *     tarr.sort()
 *     split = tarr[my_split]
 */
    __pyx_t_2 = __pyx_v_i;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tarr.rcbuffer->pybuffer.buf, __pyx_t_2, __pyx_pybuffernd_tarr.diminfo[0].strides) = ((__pyx_v_uniquedims[__pyx_v_best_dim])[__pyx_v_i]);
  }

  /* "yt/utilities/lib/misc_utilities.pyx":354
 *         #print "Setting tarr: ", i, uniquedims[best_dim][i]
 *         tarr[i] = uniquedims[best_dim][i]
 *     tarr.sort()             # <<<<<<<<<<<<<<
 *     split = tarr[my_split]
 *     cdef np.ndarray[np.uint8_t, ndim=1] less_ids = np.empty(n_grids, dtype='uint8')
 */
  __pyx_t_25 = PyObject_GetAttr(((PyObject *)__pyx_v_tarr), __pyx_n_s__sort); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 354; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_22 = PyObject_Call(__pyx_t_25, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 354; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":355
 *         tarr[i] = uniquedims[best_dim][i]
 *     tarr.sort()
 *     split = tarr[my_split]             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=1] less_ids = np.empty(n_grids, dtype='uint8')
 *     cdef np.ndarray[np.uint8_t, ndim=1] greater_ids = np.empty(n_grids, dtype='uint8')
 */
  __pyx_t_1 = __pyx_v_my_split;
  __pyx_v_split = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tarr.rcbuffer->pybuffer.buf, __pyx_t_1, __pyx_pybuffernd_tarr.diminfo[0].strides));

  /* "yt/utilities/lib/misc_utilities.pyx":356
 *     tarr.sort()
 *     split = tarr[my_split]
 *     cdef np.ndarray[np.uint8_t, ndim=1] less_ids = np.empty(n_grids, dtype='uint8')             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=1] greater_ids = np.empty(n_grids, dtype='uint8')
 *     for i in range(n_grids):
 */
  __pyx_t_22 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_25 = PyObject_GetAttr(__pyx_t_22, __pyx_n_s__empty); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = PyInt_FromLong(__pyx_v_n_grids); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  __pyx_t_22 = 0;
  __pyx_t_22 = PyDict_New(); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_22));
  if (PyDict_SetItem(__pyx_t_22, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__uint8)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_23 = PyObject_Call(__pyx_t_25, ((PyObject *)__pyx_t_24), ((PyObject *)__pyx_t_22)); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_24)); __pyx_t_24 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  if (!(likely(((__pyx_t_23) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_23, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_28 = ((PyArrayObject *)__pyx_t_23);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_less_ids.rcbuffer->pybuffer, (PyObject*)__pyx_t_28, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_less_ids = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_less_ids.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_less_ids.diminfo[0].strides = __pyx_pybuffernd_less_ids.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_less_ids.diminfo[0].shape = __pyx_pybuffernd_less_ids.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_28 = 0;
  __pyx_v_less_ids = ((PyArrayObject *)__pyx_t_23);
  __pyx_t_23 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":357
 *     split = tarr[my_split]
 *     cdef np.ndarray[np.uint8_t, ndim=1] less_ids = np.empty(n_grids, dtype='uint8')
 *     cdef np.ndarray[np.uint8_t, ndim=1] greater_ids = np.empty(n_grids, dtype='uint8')             # <<<<<<<<<<<<<<
 *     for i in range(n_grids):
 *         if data[i, 0, best_dim] < split:
 */
  __pyx_t_23 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_22 = PyObject_GetAttr(__pyx_t_23, __pyx_n_s__empty); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = PyInt_FromLong(__pyx_v_n_grids); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  __pyx_t_23 = 0;
  __pyx_t_23 = PyDict_New(); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_23));
  if (PyDict_SetItem(__pyx_t_23, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__uint8)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_25 = PyObject_Call(__pyx_t_22, ((PyObject *)__pyx_t_24), ((PyObject *)__pyx_t_23)); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_24)); __pyx_t_24 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_23)); __pyx_t_23 = 0;
  if (!(likely(((__pyx_t_25) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_25, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_29 = ((PyArrayObject *)__pyx_t_25);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_greater_ids.rcbuffer->pybuffer, (PyObject*)__pyx_t_29, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_greater_ids = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_greater_ids.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_greater_ids.diminfo[0].strides = __pyx_pybuffernd_greater_ids.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_greater_ids.diminfo[0].shape = __pyx_pybuffernd_greater_ids.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_29 = 0;
  __pyx_v_greater_ids = ((PyArrayObject *)__pyx_t_25);
  __pyx_t_25 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":358
 *     cdef np.ndarray[np.uint8_t, ndim=1] less_ids = np.empty(n_grids, dtype='uint8')
 *     cdef np.ndarray[np.uint8_t, ndim=1] greater_ids = np.empty(n_grids, dtype='uint8')
 *     for i in range(n_grids):             # <<<<<<<<<<<<<<
 *         if data[i, 0, best_dim] < split:
 *             less_ids[i] = 1
 */
  __pyx_t_3 = __pyx_v_n_grids;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/utilities/lib/misc_utilities.pyx":359
 *     cdef np.ndarray[np.uint8_t, ndim=1] greater_ids = np.empty(n_grids, dtype='uint8')
 *     for i in range(n_grids):
 *         if data[i, 0, best_dim] < split:             # <<<<<<<<<<<<<<
 *             less_ids[i] = 1
 *         else:
 */
    __pyx_t_30 = __pyx_v_i;
    __pyx_t_27 = 0;
    __pyx_t_31 = __pyx_v_best_dim;
    __pyx_t_9 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_data.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_data.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_data.diminfo[1].strides, __pyx_t_31, __pyx_pybuffernd_data.diminfo[2].strides)) < __pyx_v_split);
    if (__pyx_t_9) {

      /* "yt/utilities/lib/misc_utilities.pyx":360
 *     for i in range(n_grids):
 *         if data[i, 0, best_dim] < split:
 *             less_ids[i] = 1             # <<<<<<<<<<<<<<
 *         else:
 *             less_ids[i] = 0
 */
      __pyx_t_32 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_less_ids.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_less_ids.diminfo[0].strides) = 1;
      goto __pyx_L21;
    }
    /*else*/ {

      /* "yt/utilities/lib/misc_utilities.pyx":362
 *             less_ids[i] = 1
 *         else:
 *             less_ids[i] = 0             # <<<<<<<<<<<<<<
 *         if data[i, 1, best_dim] > split:
 *             greater_ids[i] = 1
 */
      __pyx_t_33 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_less_ids.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_less_ids.diminfo[0].strides) = 0;
    }
    __pyx_L21:;

    /* "yt/utilities/lib/misc_utilities.pyx":363
 *         else:
 *             less_ids[i] = 0
 *         if data[i, 1, best_dim] > split:             # <<<<<<<<<<<<<<
 *             greater_ids[i] = 1
 *         else:
 */
    __pyx_t_34 = __pyx_v_i;
    __pyx_t_35 = 1;
    __pyx_t_36 = __pyx_v_best_dim;
    __pyx_t_9 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_data.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_data.diminfo[0].strides, __pyx_t_35, __pyx_pybuffernd_data.diminfo[1].strides, __pyx_t_36, __pyx_pybuffernd_data.diminfo[2].strides)) > __pyx_v_split);
    if (__pyx_t_9) {

      /* "yt/utilities/lib/misc_utilities.pyx":364
 *             less_ids[i] = 0
 *         if data[i, 1, best_dim] > split:
 *             greater_ids[i] = 1             # <<<<<<<<<<<<<<
 *         else:
 *             greater_ids[i] = 0
 */
      __pyx_t_37 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_greater_ids.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_greater_ids.diminfo[0].strides) = 1;
      goto __pyx_L22;
    }
    /*else*/ {

      /* "yt/utilities/lib/misc_utilities.pyx":366
 *             greater_ids[i] = 1
 *         else:
 *             greater_ids[i] = 0             # <<<<<<<<<<<<<<
 *     # Return out unique values
 *     return best_dim, split, less_ids.view("bool"), greater_ids.view("bool")
 */
      __pyx_t_38 = __pyx_v_i;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_greater_ids.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_greater_ids.diminfo[0].strides) = 0;
    }
    __pyx_L22:;
  }

  /* "yt/utilities/lib/misc_utilities.pyx":368
 *             greater_ids[i] = 0
 *     # Return out unique values
 *     return best_dim, split, less_ids.view("bool"), greater_ids.view("bool")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_25 = PyInt_FromLong(__pyx_v_best_dim); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_23 = PyFloat_FromDouble(__pyx_v_split); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = PyObject_GetAttr(((PyObject *)__pyx_v_less_ids), __pyx_n_s__view); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_22 = PyObject_Call(__pyx_t_24, ((PyObject *)__pyx_k_tuple_1), NULL); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = PyObject_GetAttr(((PyObject *)__pyx_v_greater_ids), __pyx_n_s__view); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_39 = PyObject_Call(__pyx_t_24, ((PyObject *)__pyx_k_tuple_2), NULL); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = PyTuple_New(4); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  PyTuple_SET_ITEM(__pyx_t_24, 2, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_24, 3, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __pyx_t_25 = 0;
  __pyx_t_23 = 0;
  __pyx_t_22 = 0;
  __pyx_t_39 = 0;
  __pyx_r = ((PyObject *)__pyx_t_24);
  __pyx_t_24 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_39);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_data.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_greater_ids.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_l_corner.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_less_ids.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_r_corner.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tarr.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.kdtree_get_choices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_data.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_greater_ids.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_l_corner.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_less_ids.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_r_corner.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tarr.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tarr);
  __Pyx_XDECREF((PyObject *)__pyx_v_less_ids);
  __Pyx_XDECREF((PyObject *)__pyx_v_greater_ids);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_21grow_flagging_field(PyObject *__pyx_self, PyObject *__pyx_v_oofield); /*proto*/
static PyMethodDef __pyx_mdef_2yt_9utilities_3lib_14misc_utilities_21grow_flagging_field = {__Pyx_NAMESTR("grow_flagging_field"), (PyCFunction)__pyx_pw_2yt_9utilities_3lib_14misc_utilities_21grow_flagging_field, METH_O, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_2yt_9utilities_3lib_14misc_utilities_21grow_flagging_field(PyObject *__pyx_self, PyObject *__pyx_v_oofield) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("grow_flagging_field (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_9utilities_3lib_14misc_utilities_20grow_flagging_field(__pyx_self, ((PyObject *)__pyx_v_oofield));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/misc_utilities.pyx":371
 * 
 * 
 * def grow_flagging_field(oofield):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")
 *     cdef np.ndarray[np.uint8_t, ndim=3] nfield
 */

static PyObject *__pyx_pf_2yt_9utilities_3lib_14misc_utilities_20grow_flagging_field(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_oofield) {
  PyArrayObject *__pyx_v_ofield = 0;
  PyArrayObject *__pyx_v_nfield = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ni;
  int __pyx_v_nj;
  int __pyx_v_nk;
  int __pyx_v_oi;
  int __pyx_v_oj;
  int __pyx_v_ok;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_nfield;
  __Pyx_Buffer __pyx_pybuffer_nfield;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_ofield;
  __Pyx_Buffer __pyx_pybuffer_ofield;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyArrayObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  npy_intp __pyx_t_10;
  npy_intp __pyx_t_11;
  int __pyx_t_12;
  npy_intp __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  int __pyx_t_23;
  int __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("grow_flagging_field", 0);
  __pyx_pybuffer_ofield.pybuffer.buf = NULL;
  __pyx_pybuffer_ofield.refcount = 0;
  __pyx_pybuffernd_ofield.data = NULL;
  __pyx_pybuffernd_ofield.rcbuffer = &__pyx_pybuffer_ofield;
  __pyx_pybuffer_nfield.pybuffer.buf = NULL;
  __pyx_pybuffer_nfield.refcount = 0;
  __pyx_pybuffernd_nfield.data = NULL;
  __pyx_pybuffernd_nfield.rcbuffer = &__pyx_pybuffer_nfield;

  /* "yt/utilities/lib/misc_utilities.pyx":372
 * 
 * def grow_flagging_field(oofield):
 *     cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=3] nfield
 *     nfield = np.zeros_like(ofield)
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_oofield, __pyx_n_s__astype); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_k_tuple_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_ofield.rcbuffer->pybuffer, (PyObject*)__pyx_t_3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_ofield = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_ofield.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_ofield.diminfo[0].strides = __pyx_pybuffernd_ofield.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_ofield.diminfo[0].shape = __pyx_pybuffernd_ofield.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_ofield.diminfo[1].strides = __pyx_pybuffernd_ofield.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_ofield.diminfo[1].shape = __pyx_pybuffernd_ofield.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_ofield.diminfo[2].strides = __pyx_pybuffernd_ofield.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_ofield.diminfo[2].shape = __pyx_pybuffernd_ofield.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_3 = 0;
  __pyx_v_ofield = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":374
 *     cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")
 *     cdef np.ndarray[np.uint8_t, ndim=3] nfield
 *     nfield = np.zeros_like(ofield)             # <<<<<<<<<<<<<<
 *     cdef int i, j, k, ni, nj, nk
 *     cdef int oi, oj, ok
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 374; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__zeros_like); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 374; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 374; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_ofield));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_ofield));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_ofield));
  __pyx_t_4 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 374; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 374; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nfield.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nfield.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nfield.rcbuffer->pybuffer, (PyObject*)__pyx_v_nfield, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_nfield.diminfo[0].strides = __pyx_pybuffernd_nfield.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nfield.diminfo[0].shape = __pyx_pybuffernd_nfield.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_nfield.diminfo[1].strides = __pyx_pybuffernd_nfield.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_nfield.diminfo[1].shape = __pyx_pybuffernd_nfield.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_nfield.diminfo[2].strides = __pyx_pybuffernd_nfield.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_nfield.diminfo[2].shape = __pyx_pybuffernd_nfield.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 374; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_5 = 0;
  __pyx_v_nfield = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":377
 *     cdef int i, j, k, ni, nj, nk
 *     cdef int oi, oj, ok
 *     for ni in range(ofield.shape[0]):             # <<<<<<<<<<<<<<
 *         for nj in range(ofield.shape[1]):
 *             for nk in range(ofield.shape[2]):
 */
  __pyx_t_10 = (__pyx_v_ofield->dimensions[0]);
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_10; __pyx_t_6+=1) {
    __pyx_v_ni = __pyx_t_6;

    /* "yt/utilities/lib/misc_utilities.pyx":378
 *     cdef int oi, oj, ok
 *     for ni in range(ofield.shape[0]):
 *         for nj in range(ofield.shape[1]):             # <<<<<<<<<<<<<<
 *             for nk in range(ofield.shape[2]):
 *                 for oi in range(3):
 */
    __pyx_t_11 = (__pyx_v_ofield->dimensions[1]);
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_nj = __pyx_t_12;

      /* "yt/utilities/lib/misc_utilities.pyx":379
 *     for ni in range(ofield.shape[0]):
 *         for nj in range(ofield.shape[1]):
 *             for nk in range(ofield.shape[2]):             # <<<<<<<<<<<<<<
 *                 for oi in range(3):
 *                     i = ni + (oi - 1)
 */
      __pyx_t_13 = (__pyx_v_ofield->dimensions[2]);
      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
        __pyx_v_nk = __pyx_t_14;

        /* "yt/utilities/lib/misc_utilities.pyx":380
 *         for nj in range(ofield.shape[1]):
 *             for nk in range(ofield.shape[2]):
 *                 for oi in range(3):             # <<<<<<<<<<<<<<
 *                     i = ni + (oi - 1)
 *                     if i < 0 or i >= ofield.shape[0]: continue
 */
        for (__pyx_t_15 = 0; __pyx_t_15 < 3; __pyx_t_15+=1) {
          __pyx_v_oi = __pyx_t_15;

          /* "yt/utilities/lib/misc_utilities.pyx":381
 *             for nk in range(ofield.shape[2]):
 *                 for oi in range(3):
 *                     i = ni + (oi - 1)             # <<<<<<<<<<<<<<
 *                     if i < 0 or i >= ofield.shape[0]: continue
 *                     for oj in range(3):
 */
          __pyx_v_i = (__pyx_v_ni + (__pyx_v_oi - 1));

          /* "yt/utilities/lib/misc_utilities.pyx":382
 *                 for oi in range(3):
 *                     i = ni + (oi - 1)
 *                     if i < 0 or i >= ofield.shape[0]: continue             # <<<<<<<<<<<<<<
 *                     for oj in range(3):
 *                         j = nj + (oj - 1)
 */
          __pyx_t_16 = (__pyx_v_i < 0);
          if (!__pyx_t_16) {
            __pyx_t_17 = (__pyx_v_i >= (__pyx_v_ofield->dimensions[0]));
            __pyx_t_18 = __pyx_t_17;
          } else {
            __pyx_t_18 = __pyx_t_16;
          }
          if (__pyx_t_18) {
            goto __pyx_L9_continue;
            goto __pyx_L11;
          }
          __pyx_L11:;

          /* "yt/utilities/lib/misc_utilities.pyx":383
 *                     i = ni + (oi - 1)
 *                     if i < 0 or i >= ofield.shape[0]: continue
 *                     for oj in range(3):             # <<<<<<<<<<<<<<
 *                         j = nj + (oj - 1)
 *                         if j < 0 or j >= ofield.shape[1]: continue
 */
          for (__pyx_t_19 = 0; __pyx_t_19 < 3; __pyx_t_19+=1) {
            __pyx_v_oj = __pyx_t_19;

            /* "yt/utilities/lib/misc_utilities.pyx":384
 *                     if i < 0 or i >= ofield.shape[0]: continue
 *                     for oj in range(3):
 *                         j = nj + (oj - 1)             # <<<<<<<<<<<<<<
 *                         if j < 0 or j >= ofield.shape[1]: continue
 *                         for ok in range(3):
 */
            __pyx_v_j = (__pyx_v_nj + (__pyx_v_oj - 1));

            /* "yt/utilities/lib/misc_utilities.pyx":385
 *                     for oj in range(3):
 *                         j = nj + (oj - 1)
 *                         if j < 0 or j >= ofield.shape[1]: continue             # <<<<<<<<<<<<<<
 *                         for ok in range(3):
 *                             k = nk + (ok - 1)
 */
            __pyx_t_18 = (__pyx_v_j < 0);
            if (!__pyx_t_18) {
              __pyx_t_16 = (__pyx_v_j >= (__pyx_v_ofield->dimensions[1]));
              __pyx_t_17 = __pyx_t_16;
            } else {
              __pyx_t_17 = __pyx_t_18;
            }
            if (__pyx_t_17) {
              goto __pyx_L12_continue;
              goto __pyx_L14;
            }
            __pyx_L14:;

            /* "yt/utilities/lib/misc_utilities.pyx":386
 *                         j = nj + (oj - 1)
 *                         if j < 0 or j >= ofield.shape[1]: continue
 *                         for ok in range(3):             # <<<<<<<<<<<<<<
 *                             k = nk + (ok - 1)
 *                             if k < 0 or k >= ofield.shape[2]: continue
 */
            for (__pyx_t_20 = 0; __pyx_t_20 < 3; __pyx_t_20+=1) {
              __pyx_v_ok = __pyx_t_20;

              /* "yt/utilities/lib/misc_utilities.pyx":387
 *                         if j < 0 or j >= ofield.shape[1]: continue
 *                         for ok in range(3):
 *                             k = nk + (ok - 1)             # <<<<<<<<<<<<<<
 *                             if k < 0 or k >= ofield.shape[2]: continue
 *                             if ofield[i, j, k] == 1:
 */
              __pyx_v_k = (__pyx_v_nk + (__pyx_v_ok - 1));

              /* "yt/utilities/lib/misc_utilities.pyx":388
 *                         for ok in range(3):
 *                             k = nk + (ok - 1)
 *                             if k < 0 or k >= ofield.shape[2]: continue             # <<<<<<<<<<<<<<
 *                             if ofield[i, j, k] == 1:
 *                                 nfield[ni, nj, nk] = 1
 */
              __pyx_t_17 = (__pyx_v_k < 0);
              if (!__pyx_t_17) {
                __pyx_t_18 = (__pyx_v_k >= (__pyx_v_ofield->dimensions[2]));
                __pyx_t_16 = __pyx_t_18;
              } else {
                __pyx_t_16 = __pyx_t_17;
              }
              if (__pyx_t_16) {
                goto __pyx_L15_continue;
                goto __pyx_L17;
              }
              __pyx_L17:;

              /* "yt/utilities/lib/misc_utilities.pyx":389
 *                             k = nk + (ok - 1)
 *                             if k < 0 or k >= ofield.shape[2]: continue
 *                             if ofield[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                                 nfield[ni, nj, nk] = 1
 *     return nfield.astype("bool")
 */
              __pyx_t_21 = __pyx_v_i;
              __pyx_t_22 = __pyx_v_j;
              __pyx_t_23 = __pyx_v_k;
              __pyx_t_24 = -1;
              if (__pyx_t_21 < 0) {
                __pyx_t_21 += __pyx_pybuffernd_ofield.diminfo[0].shape;
                if (unlikely(__pyx_t_21 < 0)) __pyx_t_24 = 0;
              } else if (unlikely(__pyx_t_21 >= __pyx_pybuffernd_ofield.diminfo[0].shape)) __pyx_t_24 = 0;
              if (__pyx_t_22 < 0) {
                __pyx_t_22 += __pyx_pybuffernd_ofield.diminfo[1].shape;
                if (unlikely(__pyx_t_22 < 0)) __pyx_t_24 = 1;
              } else if (unlikely(__pyx_t_22 >= __pyx_pybuffernd_ofield.diminfo[1].shape)) __pyx_t_24 = 1;
              if (__pyx_t_23 < 0) {
                __pyx_t_23 += __pyx_pybuffernd_ofield.diminfo[2].shape;
                if (unlikely(__pyx_t_23 < 0)) __pyx_t_24 = 2;
              } else if (unlikely(__pyx_t_23 >= __pyx_pybuffernd_ofield.diminfo[2].shape)) __pyx_t_24 = 2;
              if (unlikely(__pyx_t_24 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_24);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              __pyx_t_16 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_ofield.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_ofield.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_ofield.diminfo[1].strides, __pyx_t_23, __pyx_pybuffernd_ofield.diminfo[2].strides)) == 1);
              if (__pyx_t_16) {

                /* "yt/utilities/lib/misc_utilities.pyx":390
 *                             if k < 0 or k >= ofield.shape[2]: continue
 *                             if ofield[i, j, k] == 1:
 *                                 nfield[ni, nj, nk] = 1             # <<<<<<<<<<<<<<
 *     return nfield.astype("bool")
 */
                __pyx_t_24 = __pyx_v_ni;
                __pyx_t_25 = __pyx_v_nj;
                __pyx_t_26 = __pyx_v_nk;
                __pyx_t_27 = -1;
                if (__pyx_t_24 < 0) {
                  __pyx_t_24 += __pyx_pybuffernd_nfield.diminfo[0].shape;
                  if (unlikely(__pyx_t_24 < 0)) __pyx_t_27 = 0;
                } else if (unlikely(__pyx_t_24 >= __pyx_pybuffernd_nfield.diminfo[0].shape)) __pyx_t_27 = 0;
                if (__pyx_t_25 < 0) {
                  __pyx_t_25 += __pyx_pybuffernd_nfield.diminfo[1].shape;
                  if (unlikely(__pyx_t_25 < 0)) __pyx_t_27 = 1;
                } else if (unlikely(__pyx_t_25 >= __pyx_pybuffernd_nfield.diminfo[1].shape)) __pyx_t_27 = 1;
                if (__pyx_t_26 < 0) {
                  __pyx_t_26 += __pyx_pybuffernd_nfield.diminfo[2].shape;
                  if (unlikely(__pyx_t_26 < 0)) __pyx_t_27 = 2;
                } else if (unlikely(__pyx_t_26 >= __pyx_pybuffernd_nfield.diminfo[2].shape)) __pyx_t_27 = 2;
                if (unlikely(__pyx_t_27 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_27);
                  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 390; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                }
                *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_nfield.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_nfield.diminfo[0].strides, __pyx_t_25, __pyx_pybuffernd_nfield.diminfo[1].strides, __pyx_t_26, __pyx_pybuffernd_nfield.diminfo[2].strides) = 1;
                goto __pyx_L18;
              }
              __pyx_L18:;
              __pyx_L15_continue:;
            }
            __pyx_L12_continue:;
          }
          __pyx_L9_continue:;
        }
      }
    }
  }

  /* "yt/utilities/lib/misc_utilities.pyx":391
 *                             if ofield[i, j, k] == 1:
 *                                 nfield[ni, nj, nk] = 1
 *     return nfield.astype("bool")             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyObject_GetAttr(((PyObject *)__pyx_v_nfield), __pyx_n_s__astype); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_k_tuple_4), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nfield.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ofield.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.utilities.lib.misc_utilities.grow_flagging_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nfield.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ofield.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ofield);
  __Pyx_XDECREF((PyObject *)__pyx_v_nfield);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":194
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "numpy.pxd":200
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = (__pyx_v_info == NULL);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":203
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":204
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":206
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "numpy.pxd":208
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = ((sizeof(npy_intp)) != (sizeof(Py_ssize_t)));
  if (__pyx_t_1) {

    /* "numpy.pxd":209
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "numpy.pxd":211
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "numpy.pxd":213
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS);
  if (__pyx_t_1) {

    /* "numpy.pxd":214
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
    __pyx_t_2 = (!PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS));
    __pyx_t_3 = __pyx_t_2;
  } else {
    __pyx_t_3 = __pyx_t_1;
  }
  if (__pyx_t_3) {

    /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_6), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "numpy.pxd":217
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_3 = ((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS);
  if (__pyx_t_3) {

    /* "numpy.pxd":218
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
    __pyx_t_1 = (!PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS));
    __pyx_t_2 = __pyx_t_1;
  } else {
    __pyx_t_2 = __pyx_t_3;
  }
  if (__pyx_t_2) {

    /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_8), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "numpy.pxd":221
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "numpy.pxd":222
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "numpy.pxd":223
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  if (__pyx_v_copy_shape) {

    /* "numpy.pxd":226
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "numpy.pxd":227
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "numpy.pxd":228
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_5 = __pyx_v_ndim;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "numpy.pxd":229
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "numpy.pxd":230
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }
    goto __pyx_L7;
  }
  /*else*/ {

    /* "numpy.pxd":232
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "numpy.pxd":233
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L7:;

  /* "numpy.pxd":234
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "numpy.pxd":235
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "numpy.pxd":236
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!PyArray_ISWRITEABLE(__pyx_v_self));

  /* "numpy.pxd":239
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef list stack
 */
  __pyx_v_f = NULL;

  /* "numpy.pxd":240
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef list stack
 *             cdef int offset
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self->descr));
  __pyx_v_descr = __pyx_v_self->descr;

  /* "numpy.pxd":244
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "numpy.pxd":246
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = (!__pyx_v_hasfields);
  if (__pyx_t_2) {
    __pyx_t_3 = (!__pyx_v_copy_shape);
    __pyx_t_1 = __pyx_t_3;
  } else {
    __pyx_t_1 = __pyx_t_2;
  }
  if (__pyx_t_1) {

    /* "numpy.pxd":248
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;
    goto __pyx_L10;
  }
  /*else*/ {

    /* "numpy.pxd":251
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L10:;

  /* "numpy.pxd":253
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = (!__pyx_v_hasfields);
  if (__pyx_t_1) {

    /* "numpy.pxd":254
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_v_t = __pyx_v_descr->type_num;

    /* "numpy.pxd":255
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_1 = (__pyx_v_descr->byteorder == '>');
    if (__pyx_t_1) {
      __pyx_t_2 = __pyx_v_little_endian;
    } else {
      __pyx_t_2 = __pyx_t_1;
    }
    if (!__pyx_t_2) {

      /* "numpy.pxd":256
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
      __pyx_t_1 = (__pyx_v_descr->byteorder == '<');
      if (__pyx_t_1) {
        __pyx_t_3 = (!__pyx_v_little_endian);
        __pyx_t_7 = __pyx_t_3;
      } else {
        __pyx_t_7 = __pyx_t_1;
      }
      __pyx_t_1 = __pyx_t_7;
    } else {
      __pyx_t_1 = __pyx_t_2;
    }
    if (__pyx_t_1) {

      /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_10), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L12;
    }
    __pyx_L12:;

    /* "numpy.pxd":258
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_BYTE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__b;
      goto __pyx_L13;
    }

    /* "numpy.pxd":259
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_UBYTE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__B;
      goto __pyx_L13;
    }

    /* "numpy.pxd":260
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_SHORT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__h;
      goto __pyx_L13;
    }

    /* "numpy.pxd":261
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_USHORT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__H;
      goto __pyx_L13;
    }

    /* "numpy.pxd":262
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_INT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__i;
      goto __pyx_L13;
    }

    /* "numpy.pxd":263
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_UINT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__I;
      goto __pyx_L13;
    }

    /* "numpy.pxd":264
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__l;
      goto __pyx_L13;
    }

    /* "numpy.pxd":265
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_ULONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__L;
      goto __pyx_L13;
    }

    /* "numpy.pxd":266
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONGLONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__q;
      goto __pyx_L13;
    }

    /* "numpy.pxd":267
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_ULONGLONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Q;
      goto __pyx_L13;
    }

    /* "numpy.pxd":268
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_FLOAT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__f;
      goto __pyx_L13;
    }

    /* "numpy.pxd":269
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_DOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__d;
      goto __pyx_L13;
    }

    /* "numpy.pxd":270
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONGDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__g;
      goto __pyx_L13;
    }

    /* "numpy.pxd":271
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CFLOAT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zf;
      goto __pyx_L13;
    }

    /* "numpy.pxd":272
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zd;
      goto __pyx_L13;
    }

    /* "numpy.pxd":273
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CLONGDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zg;
      goto __pyx_L13;
    }

    /* "numpy.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
    __pyx_t_1 = (__pyx_v_t == NPY_OBJECT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__O;
      goto __pyx_L13;
    }
    /*else*/ {

      /* "numpy.pxd":276
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_4 = PyInt_FromLong(__pyx_v_t); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_11), __pyx_t_4); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_8));
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_8));
      __Pyx_GIVEREF(((PyObject *)__pyx_t_8));
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_L13:;

    /* "numpy.pxd":277
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "numpy.pxd":278
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L11;
  }
  /*else*/ {

    /* "numpy.pxd":280
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
    __pyx_v_info->format = ((char *)malloc(255));

    /* "numpy.pxd":281
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "numpy.pxd":282
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "numpy.pxd":285
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)             # <<<<<<<<<<<<<<
 *                 f[0] = c'\0' # Terminate format string
 * 
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 255), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_f = __pyx_t_9;

    /* "numpy.pxd":286
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }
  __pyx_L11:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));
  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":288
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "numpy.pxd":289
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = PyArray_HASFIELDS(__pyx_v_self);
  if (__pyx_t_1) {

    /* "numpy.pxd":290
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":291
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = ((sizeof(npy_intp)) != (sizeof(Py_ssize_t)));
  if (__pyx_t_1) {

    /* "numpy.pxd":292
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);
    goto __pyx_L4;
  }
  __pyx_L4:;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":768
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "numpy.pxd":769
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":771
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "numpy.pxd":772
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":774
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "numpy.pxd":775
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":777
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "numpy.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 778; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":780
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "numpy.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 781; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":783
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  long __pyx_t_11;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "numpy.pxd":790
 *     cdef int delta_offset
 *     cdef tuple i
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":791
 *     cdef tuple i
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(((PyObject *)__pyx_v_descr->names) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_1 = ((PyObject *)__pyx_v_descr->names); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __Pyx_XDECREF(__pyx_v_childname);
    __pyx_v_childname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "numpy.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (!__pyx_t_3) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected tuple, got %.200s", Py_TYPE(__pyx_t_3)->tp_name), 0))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_fields));
    __pyx_v_fields = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "numpy.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 */
    if (likely(PyTuple_CheckExact(((PyObject *)__pyx_v_fields)))) {
      PyObject* sequence = ((PyObject *)__pyx_v_fields);
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else if (1) {
      __Pyx_RaiseNoneNotIterableError(); {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else
    {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(((PyObject *)__pyx_v_fields)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_child));
    __pyx_v_child = ((PyArray_Descr *)__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_v_new_offset);
    __pyx_v_new_offset = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "numpy.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = PyInt_FromLong((__pyx_v_end - __pyx_v_f)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_int_15, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_13), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L7;
    }
    __pyx_L7:;

    /* "numpy.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = (__pyx_v_child->byteorder == '>');
    if (__pyx_t_7) {
      __pyx_t_8 = __pyx_v_little_endian;
    } else {
      __pyx_t_8 = __pyx_t_7;
    }
    if (!__pyx_t_8) {

      /* "numpy.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
      __pyx_t_7 = (__pyx_v_child->byteorder == '<');
      if (__pyx_t_7) {
        __pyx_t_9 = (!__pyx_v_little_endian);
        __pyx_t_10 = __pyx_t_9;
      } else {
        __pyx_t_10 = __pyx_t_7;
      }
      __pyx_t_7 = __pyx_t_10;
    } else {
      __pyx_t_7 = __pyx_t_8;
    }
    if (__pyx_t_7) {

      /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_14), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "numpy.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_5 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!__pyx_t_7) break;

      /* "numpy.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 120;

      /* "numpy.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "numpy.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_11 = 0;
      (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + 1);
    }

    /* "numpy.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_11 = 0;
    (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + __pyx_v_child->elsize);

    /* "numpy.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_7 = (!PyDataType_HASFIELDS(__pyx_v_child));
    if (__pyx_t_7) {

      /* "numpy.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_3 = PyInt_FromLong(__pyx_v_child->type_num); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 821; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF(__pyx_v_t);
      __pyx_v_t = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "numpy.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_7 = ((__pyx_v_end - __pyx_v_f) < 5);
      if (__pyx_t_7) {

        /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_3 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_16), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        goto __pyx_L12;
      }
      __pyx_L12:;

      /* "numpy.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_BYTE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L13;
      }

      /* "numpy.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UBYTE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L13;
      }

      /* "numpy.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_SHORT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 104;
        goto __pyx_L13;
      }

      /* "numpy.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_USHORT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L13;
      }

      /* "numpy.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_INT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 105;
        goto __pyx_L13;
      }

      /* "numpy.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UINT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L13;
      }

      /* "numpy.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 108;
        goto __pyx_L13;
      }

      /* "numpy.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L13;
      }

      /* "numpy.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGLONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 113;
        goto __pyx_L13;
      }

      /* "numpy.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONGLONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L13;
      }

      /* "numpy.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_FLOAT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 102;
        goto __pyx_L13;
      }

      /* "numpy.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_5 = PyInt_FromLong(NPY_DOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 100;
        goto __pyx_L13;
      }

      /* "numpy.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 103;
        goto __pyx_L13;
      }

      /* "numpy.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CFLOAT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 102;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_CDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 100;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 103;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_3 = PyInt_FromLong(NPY_OBJECT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L13;
      }
      /*else*/ {

        /* "numpy.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
        __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_11), __pyx_v_t); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(((PyObject *)__pyx_t_5));
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
        __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
        __pyx_t_5 = 0;
        __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_L13:;

      /* "numpy.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);
      goto __pyx_L11;
    }
    /*else*/ {

      /* "numpy.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
      __pyx_t_12 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_12 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_v_f = __pyx_t_12;
    }
    __pyx_L11:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numpy.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":965
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "numpy.pxd":967
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  if (__pyx_t_1) {

    /* "numpy.pxd":968
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":970
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
    Py_INCREF(__pyx_v_base);

    /* "numpy.pxd":971
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "numpy.pxd":972
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "numpy.pxd":973
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "numpy.pxd":976
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = (__pyx_v_arr->base == NULL);
  if (__pyx_t_1) {

    /* "numpy.pxd":977
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":979
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    __Pyx_NAMESTR("misc_utilities"),
    __Pyx_DOCSTR(__pyx_k_17), /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_11, __pyx_k_11, sizeof(__pyx_k_11), 0, 1, 0, 0},
  {&__pyx_kp_u_12, __pyx_k_12, sizeof(__pyx_k_12), 0, 1, 0, 0},
  {&__pyx_kp_u_15, __pyx_k_15, sizeof(__pyx_k_15), 0, 1, 0, 0},
  {&__pyx_kp_s_20, __pyx_k_20, sizeof(__pyx_k_20), 0, 0, 1, 0},
  {&__pyx_n_s_21, __pyx_k_21, sizeof(__pyx_k_21), 0, 0, 1, 1},
  {&__pyx_n_s_36, __pyx_k_36, sizeof(__pyx_k_36), 0, 0, 1, 1},
  {&__pyx_n_s_39, __pyx_k_39, sizeof(__pyx_k_39), 0, 0, 1, 1},
  {&__pyx_kp_u_5, __pyx_k_5, sizeof(__pyx_k_5), 0, 1, 0, 0},
  {&__pyx_kp_u_7, __pyx_k_7, sizeof(__pyx_k_7), 0, 1, 0, 0},
  {&__pyx_kp_u_9, __pyx_k_9, sizeof(__pyx_k_9), 0, 1, 0, 0},
  {&__pyx_n_s__KeyError, __pyx_k__KeyError, sizeof(__pyx_k__KeyError), 0, 0, 1, 1},
  {&__pyx_n_s__R, __pyx_k__R, sizeof(__pyx_k__R), 0, 0, 1, 1},
  {&__pyx_n_s__RuntimeError, __pyx_k__RuntimeError, sizeof(__pyx_k__RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s__ValueError, __pyx_k__ValueError, sizeof(__pyx_k__ValueError), 0, 0, 1, 1},
  {&__pyx_n_s____main__, __pyx_k____main__, sizeof(__pyx_k____main__), 0, 0, 1, 1},
  {&__pyx_n_s____test__, __pyx_k____test__, sizeof(__pyx_k____test__), 0, 0, 1, 1},
  {&__pyx_n_s__addit, __pyx_k__addit, sizeof(__pyx_k__addit), 0, 0, 1, 1},
  {&__pyx_n_s__alpha, __pyx_k__alpha, sizeof(__pyx_k__alpha), 0, 0, 1, 1},
  {&__pyx_n_s__astype, __pyx_k__astype, sizeof(__pyx_k__astype), 0, 0, 1, 1},
  {&__pyx_n_s__best_dim, __pyx_k__best_dim, sizeof(__pyx_k__best_dim), 0, 0, 1, 1},
  {&__pyx_n_s__bin, __pyx_k__bin, sizeof(__pyx_k__bin), 0, 0, 1, 1},
  {&__pyx_n_s__bin_profile1d, __pyx_k__bin_profile1d, sizeof(__pyx_k__bin_profile1d), 0, 0, 1, 1},
  {&__pyx_n_s__bin_profile2d, __pyx_k__bin_profile2d, sizeof(__pyx_k__bin_profile2d), 0, 0, 1, 1},
  {&__pyx_n_s__bin_profile3d, __pyx_k__bin_profile3d, sizeof(__pyx_k__bin_profile3d), 0, 0, 1, 1},
  {&__pyx_n_s__bini, __pyx_k__bini, sizeof(__pyx_k__bini), 0, 0, 1, 1},
  {&__pyx_n_s__binj, __pyx_k__binj, sizeof(__pyx_k__binj), 0, 0, 1, 1},
  {&__pyx_n_s__bink, __pyx_k__bink, sizeof(__pyx_k__bink), 0, 0, 1, 1},
  {&__pyx_n_s__bins_x, __pyx_k__bins_x, sizeof(__pyx_k__bins_x), 0, 0, 1, 1},
  {&__pyx_n_s__bins_y, __pyx_k__bins_y, sizeof(__pyx_k__bins_y), 0, 0, 1, 1},
  {&__pyx_n_s__bins_z, __pyx_k__bins_z, sizeof(__pyx_k__bins_z), 0, 0, 1, 1},
  {&__pyx_n_s__bool, __pyx_k__bool, sizeof(__pyx_k__bool), 0, 0, 1, 1},
  {&__pyx_n_s__bresult, __pyx_k__bresult, sizeof(__pyx_k__bresult), 0, 0, 1, 1},
  {&__pyx_n_s__bsource, __pyx_k__bsource, sizeof(__pyx_k__bsource), 0, 0, 1, 1},
  {&__pyx_n_s__bval, __pyx_k__bval, sizeof(__pyx_k__bval), 0, 0, 1, 1},
  {&__pyx_n_s__colors, __pyx_k__colors, sizeof(__pyx_k__colors), 0, 0, 1, 1},
  {&__pyx_n_s__data, __pyx_k__data, sizeof(__pyx_k__data), 0, 0, 1, 1},
  {&__pyx_n_s__dds, __pyx_k__dds, sizeof(__pyx_k__dds), 0, 0, 1, 1},
  {&__pyx_n_s__dim, __pyx_k__dim, sizeof(__pyx_k__dim), 0, 0, 1, 1},
  {&__pyx_n_s__dimensions, __pyx_k__dimensions, sizeof(__pyx_k__dimensions), 0, 0, 1, 1},
  {&__pyx_n_s__dtype, __pyx_k__dtype, sizeof(__pyx_k__dtype), 0, 0, 1, 1},
  {&__pyx_n_s__dx, __pyx_k__dx, sizeof(__pyx_k__dx), 0, 0, 1, 1},
  {&__pyx_n_s__dy, __pyx_k__dy, sizeof(__pyx_k__dy), 0, 0, 1, 1},
  {&__pyx_n_s__e2, __pyx_k__e2, sizeof(__pyx_k__e2), 0, 0, 1, 1},
  {&__pyx_n_s__empty, __pyx_k__empty, sizeof(__pyx_k__empty), 0, 0, 1, 1},
  {&__pyx_n_s__eps, __pyx_k__eps, sizeof(__pyx_k__eps), 0, 0, 1, 1},
  {&__pyx_n_s__err, __pyx_k__err, sizeof(__pyx_k__err), 0, 0, 1, 1},
  {&__pyx_n_s__fi, __pyx_k__fi, sizeof(__pyx_k__fi), 0, 0, 1, 1},
  {&__pyx_n_s__field, __pyx_k__field, sizeof(__pyx_k__field), 0, 0, 1, 1},
  {&__pyx_n_s__field_names, __pyx_k__field_names, sizeof(__pyx_k__field_names), 0, 0, 1, 1},
  {&__pyx_n_s__finfo, __pyx_k__finfo, sizeof(__pyx_k__finfo), 0, 0, 1, 1},
  {&__pyx_n_s__float64, __pyx_k__float64, sizeof(__pyx_k__float64), 0, 0, 1, 1},
  {&__pyx_n_s__get_box_grids_level, __pyx_k__get_box_grids_level, sizeof(__pyx_k__get_box_grids_level), 0, 0, 1, 1},
  {&__pyx_n_s__get_color_bounds, __pyx_k__get_color_bounds, sizeof(__pyx_k__get_color_bounds), 0, 0, 1, 1},
  {&__pyx_n_s__gi, __pyx_k__gi, sizeof(__pyx_k__gi), 0, 0, 1, 1},
  {&__pyx_n_s__greater_ids, __pyx_k__greater_ids, sizeof(__pyx_k__greater_ids), 0, 0, 1, 1},
  {&__pyx_n_s__grid, __pyx_k__grid, sizeof(__pyx_k__grid), 0, 0, 1, 1},
  {&__pyx_n_s__grid_objects, __pyx_k__grid_objects, sizeof(__pyx_k__grid_objects), 0, 0, 1, 1},
  {&__pyx_n_s__grow_flagging_field, __pyx_k__grow_flagging_field, sizeof(__pyx_k__grow_flagging_field), 0, 0, 1, 1},
  {&__pyx_n_s__has_alpha, __pyx_k__has_alpha, sizeof(__pyx_k__has_alpha), 0, 0, 1, 1},
  {&__pyx_n_s__i, __pyx_k__i, sizeof(__pyx_k__i), 0, 0, 1, 1},
  {&__pyx_n_s__image, __pyx_k__image, sizeof(__pyx_k__image), 0, 0, 1, 1},
  {&__pyx_n_s__ind, __pyx_k__ind, sizeof(__pyx_k__ind), 0, 0, 1, 1},
  {&__pyx_n_s__inside, __pyx_k__inside, sizeof(__pyx_k__inside), 0, 0, 1, 1},
  {&__pyx_n_s__j, __pyx_k__j, sizeof(__pyx_k__j), 0, 0, 1, 1},
  {&__pyx_n_s__k, __pyx_k__k, sizeof(__pyx_k__k), 0, 0, 1, 1},
  {&__pyx_n_s__kdtree_get_choices, __pyx_k__kdtree_get_choices, sizeof(__pyx_k__kdtree_get_choices), 0, 0, 1, 1},
  {&__pyx_n_s__l_corner, __pyx_k__l_corner, sizeof(__pyx_k__l_corner), 0, 0, 1, 1},
  {&__pyx_n_s__left_edge, __pyx_k__left_edge, sizeof(__pyx_k__left_edge), 0, 0, 1, 1},
  {&__pyx_n_s__left_edges, __pyx_k__left_edges, sizeof(__pyx_k__left_edges), 0, 0, 1, 1},
  {&__pyx_n_s__leftx, __pyx_k__leftx, sizeof(__pyx_k__leftx), 0, 0, 1, 1},
  {&__pyx_n_s__lefty, __pyx_k__lefty, sizeof(__pyx_k__lefty), 0, 0, 1, 1},
  {&__pyx_n_s__less_ids, __pyx_k__less_ids, sizeof(__pyx_k__less_ids), 0, 0, 1, 1},
  {&__pyx_n_s__level, __pyx_k__level, sizeof(__pyx_k__level), 0, 0, 1, 1},
  {&__pyx_n_s__levels, __pyx_k__levels, sizeof(__pyx_k__levels), 0, 0, 1, 1},
  {&__pyx_n_s__lines, __pyx_k__lines, sizeof(__pyx_k__lines), 0, 0, 1, 1},
  {&__pyx_n_s__ma, __pyx_k__ma, sizeof(__pyx_k__ma), 0, 0, 1, 1},
  {&__pyx_n_s__mask, __pyx_k__mask, sizeof(__pyx_k__mask), 0, 0, 1, 1},
  {&__pyx_n_s__maxdx, __pyx_k__maxdx, sizeof(__pyx_k__maxdx), 0, 0, 1, 1},
  {&__pyx_n_s__mi, __pyx_k__mi, sizeof(__pyx_k__mi), 0, 0, 1, 1},
  {&__pyx_n_s__min_index, __pyx_k__min_index, sizeof(__pyx_k__min_index), 0, 0, 1, 1},
  {&__pyx_n_s__min_level, __pyx_k__min_level, sizeof(__pyx_k__min_level), 0, 0, 1, 1},
  {&__pyx_n_s__mindx, __pyx_k__mindx, sizeof(__pyx_k__mindx), 0, 0, 1, 1},
  {&__pyx_n_s__mresult, __pyx_k__mresult, sizeof(__pyx_k__mresult), 0, 0, 1, 1},
  {&__pyx_n_s__my_max, __pyx_k__my_max, sizeof(__pyx_k__my_max), 0, 0, 1, 1},
  {&__pyx_n_s__my_split, __pyx_k__my_split, sizeof(__pyx_k__my_split), 0, 0, 1, 1},
  {&__pyx_n_s__n, __pyx_k__n, sizeof(__pyx_k__n), 0, 0, 1, 1},
  {&__pyx_n_s__n_best, __pyx_k__n_best, sizeof(__pyx_k__n_best), 0, 0, 1, 1},
  {&__pyx_n_s__n_grids, __pyx_k__n_grids, sizeof(__pyx_k__n_grids), 0, 0, 1, 1},
  {&__pyx_n_s__n_unique, __pyx_k__n_unique, sizeof(__pyx_k__n_unique), 0, 0, 1, 1},
  {&__pyx_n_s__nf, __pyx_k__nf, sizeof(__pyx_k__nf), 0, 0, 1, 1},
  {&__pyx_n_s__nfield, __pyx_k__nfield, sizeof(__pyx_k__nfield), 0, 0, 1, 1},
  {&__pyx_n_s__ni, __pyx_k__ni, sizeof(__pyx_k__ni), 0, 0, 1, 1},
  {&__pyx_n_s__nj, __pyx_k__nj, sizeof(__pyx_k__nj), 0, 0, 1, 1},
  {&__pyx_n_s__nk, __pyx_k__nk, sizeof(__pyx_k__nk), 0, 0, 1, 1},
  {&__pyx_n_s__nl, __pyx_k__nl, sizeof(__pyx_k__nl), 0, 0, 1, 1},
  {&__pyx_n_s__np, __pyx_k__np, sizeof(__pyx_k__np), 0, 0, 1, 1},
  {&__pyx_n_s__numpy, __pyx_k__numpy, sizeof(__pyx_k__numpy), 0, 0, 1, 1},
  {&__pyx_n_s__nx, __pyx_k__nx, sizeof(__pyx_k__nx), 0, 0, 1, 1},
  {&__pyx_n_s__ny, __pyx_k__ny, sizeof(__pyx_k__ny), 0, 0, 1, 1},
  {&__pyx_n_s__ofield, __pyx_k__ofield, sizeof(__pyx_k__ofield), 0, 0, 1, 1},
  {&__pyx_n_s__oi, __pyx_k__oi, sizeof(__pyx_k__oi), 0, 0, 1, 1},
  {&__pyx_n_s__oj, __pyx_k__oj, sizeof(__pyx_k__oj), 0, 0, 1, 1},
  {&__pyx_n_s__ok, __pyx_k__ok, sizeof(__pyx_k__ok), 0, 0, 1, 1},
  {&__pyx_n_s__oofield, __pyx_k__oofield, sizeof(__pyx_k__oofield), 0, 0, 1, 1},
  {&__pyx_n_s__pdx, __pyx_k__pdx, sizeof(__pyx_k__pdx), 0, 0, 1, 1},
  {&__pyx_n_s__pdy, __pyx_k__pdy, sizeof(__pyx_k__pdy), 0, 0, 1, 1},
  {&__pyx_n_s__point, __pyx_k__point, sizeof(__pyx_k__point), 0, 0, 1, 1},
  {&__pyx_n_s__points_per_color, __pyx_k__points_per_color, sizeof(__pyx_k__points_per_color), 0, 0, 1, 1},
  {&__pyx_n_s__px, __pyx_k__px, sizeof(__pyx_k__px), 0, 0, 1, 1},
  {&__pyx_n_s__py, __pyx_k__py, sizeof(__pyx_k__py), 0, 0, 1, 1},
  {&__pyx_n_s__qresult, __pyx_k__qresult, sizeof(__pyx_k__qresult), 0, 0, 1, 1},
  {&__pyx_n_s__r_corner, __pyx_k__r_corner, sizeof(__pyx_k__r_corner), 0, 0, 1, 1},
  {&__pyx_n_s__range, __pyx_k__range, sizeof(__pyx_k__range), 0, 0, 1, 1},
  {&__pyx_n_s__right_edge, __pyx_k__right_edge, sizeof(__pyx_k__right_edge), 0, 0, 1, 1},
  {&__pyx_n_s__right_edges, __pyx_k__right_edges, sizeof(__pyx_k__right_edges), 0, 0, 1, 1},
  {&__pyx_n_s__rightx, __pyx_k__rightx, sizeof(__pyx_k__rightx), 0, 0, 1, 1},
  {&__pyx_n_s__righty, __pyx_k__righty, sizeof(__pyx_k__righty), 0, 0, 1, 1},
  {&__pyx_n_s__rotate_vectors, __pyx_k__rotate_vectors, sizeof(__pyx_k__rotate_vectors), 0, 0, 1, 1},
  {&__pyx_n_s__rotated, __pyx_k__rotated, sizeof(__pyx_k__rotated), 0, 0, 1, 1},
  {&__pyx_n_s__rv, __pyx_k__rv, sizeof(__pyx_k__rv), 0, 0, 1, 1},
  {&__pyx_n_s__skipit, __pyx_k__skipit, sizeof(__pyx_k__skipit), 0, 0, 1, 1},
  {&__pyx_n_s__sort, __pyx_k__sort, sizeof(__pyx_k__sort), 0, 0, 1, 1},
  {&__pyx_n_s__split, __pyx_k__split, sizeof(__pyx_k__split), 0, 0, 1, 1},
  {&__pyx_n_s__sx, __pyx_k__sx, sizeof(__pyx_k__sx), 0, 0, 1, 1},
  {&__pyx_n_s__sy, __pyx_k__sy, sizeof(__pyx_k__sy), 0, 0, 1, 1},
  {&__pyx_n_s__tarr, __pyx_k__tarr, sizeof(__pyx_k__tarr), 0, 0, 1, 1},
  {&__pyx_n_s__uint8, __pyx_k__uint8, sizeof(__pyx_k__uint8), 0, 0, 1, 1},
  {&__pyx_n_s__uniquedims, __pyx_k__uniquedims, sizeof(__pyx_k__uniquedims), 0, 0, 1, 1},
  {&__pyx_n_s__uniques, __pyx_k__uniques, sizeof(__pyx_k__uniques), 0, 0, 1, 1},
  {&__pyx_n_s__used, __pyx_k__used, sizeof(__pyx_k__used), 0, 0, 1, 1},
  {&__pyx_n_s__v, __pyx_k__v, sizeof(__pyx_k__v), 0, 0, 1, 1},
  {&__pyx_n_s__value, __pyx_k__value, sizeof(__pyx_k__value), 0, 0, 1, 1},
  {&__pyx_n_s__vecs, __pyx_k__vecs, sizeof(__pyx_k__vecs), 0, 0, 1, 1},
  {&__pyx_n_s__view, __pyx_k__view, sizeof(__pyx_k__view), 0, 0, 1, 1},
  {&__pyx_n_s__wresult, __pyx_k__wresult, sizeof(__pyx_k__wresult), 0, 0, 1, 1},
  {&__pyx_n_s__wsource, __pyx_k__wsource, sizeof(__pyx_k__wsource), 0, 0, 1, 1},
  {&__pyx_n_s__wval, __pyx_k__wval, sizeof(__pyx_k__wval), 0, 0, 1, 1},
  {&__pyx_n_s__x0, __pyx_k__x0, sizeof(__pyx_k__x0), 0, 0, 1, 1},
  {&__pyx_n_s__x1, __pyx_k__x1, sizeof(__pyx_k__x1), 0, 0, 1, 1},
  {&__pyx_n_s__xs, __pyx_k__xs, sizeof(__pyx_k__xs), 0, 0, 1, 1},
  {&__pyx_n_s__y0, __pyx_k__y0, sizeof(__pyx_k__y0), 0, 0, 1, 1},
  {&__pyx_n_s__y1, __pyx_k__y1, sizeof(__pyx_k__y1), 0, 0, 1, 1},
  {&__pyx_n_s__ys, __pyx_k__ys, sizeof(__pyx_k__ys), 0, 0, 1, 1},
  {&__pyx_n_s__zeros, __pyx_k__zeros, sizeof(__pyx_k__zeros), 0, 0, 1, 1},
  {&__pyx_n_s__zeros_like, __pyx_k__zeros_like, sizeof(__pyx_k__zeros_like), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetName(__pyx_b, __pyx_n_s__range); if (!__pyx_builtin_range) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 47; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_KeyError = __Pyx_GetName(__pyx_b, __pyx_n_s__KeyError); if (!__pyx_builtin_KeyError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetName(__pyx_b, __pyx_n_s__ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetName(__pyx_b, __pyx_n_s__RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "yt/utilities/lib/misc_utilities.pyx":368
 *             greater_ids[i] = 0
 *     # Return out unique values
 *     return best_dim, split, less_ids.view("bool"), greater_ids.view("bool")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_k_tuple_1 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_1);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bool));
  PyTuple_SET_ITEM(__pyx_k_tuple_1, 0, ((PyObject *)__pyx_n_s__bool));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bool));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_1));
  __pyx_k_tuple_2 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_2);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bool));
  PyTuple_SET_ITEM(__pyx_k_tuple_2, 0, ((PyObject *)__pyx_n_s__bool));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bool));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_2));

  /* "yt/utilities/lib/misc_utilities.pyx":372
 * 
 * def grow_flagging_field(oofield):
 *     cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=3] nfield
 *     nfield = np.zeros_like(ofield)
 */
  __pyx_k_tuple_3 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_3);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__uint8));
  PyTuple_SET_ITEM(__pyx_k_tuple_3, 0, ((PyObject *)__pyx_n_s__uint8));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__uint8));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_3));

  /* "yt/utilities/lib/misc_utilities.pyx":391
 *                             if ofield[i, j, k] == 1:
 *                                 nfield[ni, nj, nk] = 1
 *     return nfield.astype("bool")             # <<<<<<<<<<<<<<
 */
  __pyx_k_tuple_4 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_4);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bool));
  PyTuple_SET_ITEM(__pyx_k_tuple_4, 0, ((PyObject *)__pyx_n_s__bool));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bool));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_4));

  /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_k_tuple_6 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_6)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_6);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_5));
  PyTuple_SET_ITEM(__pyx_k_tuple_6, 0, ((PyObject *)__pyx_kp_u_5));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_5));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_6));

  /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_k_tuple_8 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_8);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_7));
  PyTuple_SET_ITEM(__pyx_k_tuple_8, 0, ((PyObject *)__pyx_kp_u_7));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_7));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_8));

  /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_k_tuple_10 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_10)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_10);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_9));
  PyTuple_SET_ITEM(__pyx_k_tuple_10, 0, ((PyObject *)__pyx_kp_u_9));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_9));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_10));

  /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_k_tuple_13 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_13)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_13);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_12));
  PyTuple_SET_ITEM(__pyx_k_tuple_13, 0, ((PyObject *)__pyx_kp_u_12));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_12));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_13));

  /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_k_tuple_14 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_14)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_14);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_9));
  PyTuple_SET_ITEM(__pyx_k_tuple_14, 0, ((PyObject *)__pyx_kp_u_9));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_9));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_14));

  /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_k_tuple_16 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_16)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_16);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_15));
  PyTuple_SET_ITEM(__pyx_k_tuple_16, 0, ((PyObject *)__pyx_kp_u_15));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_15));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_16));

  /* "yt/utilities/lib/misc_utilities.pyx":37
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile1d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.float64_t, ndim=1] wsource,
 *                   np.ndarray[np.float64_t, ndim=1] bsource,
 */
  __pyx_k_tuple_18 = PyTuple_New(12); if (unlikely(!__pyx_k_tuple_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_18);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bins_x));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 0, ((PyObject *)__pyx_n_s__bins_x));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bins_x));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wsource));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 1, ((PyObject *)__pyx_n_s__wsource));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wsource));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bsource));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 2, ((PyObject *)__pyx_n_s__bsource));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bsource));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 3, ((PyObject *)__pyx_n_s__wresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 4, ((PyObject *)__pyx_n_s__bresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 5, ((PyObject *)__pyx_n_s__mresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__qresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 6, ((PyObject *)__pyx_n_s__qresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__qresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__used));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 7, ((PyObject *)__pyx_n_s__used));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__used));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 8, ((PyObject *)__pyx_n_s__n));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bin));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 9, ((PyObject *)__pyx_n_s__bin));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bin));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wval));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 10, ((PyObject *)__pyx_n_s__wval));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wval));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bval));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 11, ((PyObject *)__pyx_n_s__bval));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bval));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_18));
  __pyx_k_codeobj_19 = (PyObject*)__Pyx_PyCode_New(8, 0, 12, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__bin_profile1d, 37, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":62
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile2d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=1] bins_y,
 *                   np.ndarray[np.float64_t, ndim=1] wsource,
 */
  __pyx_k_tuple_22 = PyTuple_New(15); if (unlikely(!__pyx_k_tuple_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_22);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bins_x));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 0, ((PyObject *)__pyx_n_s__bins_x));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bins_x));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bins_y));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 1, ((PyObject *)__pyx_n_s__bins_y));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bins_y));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wsource));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 2, ((PyObject *)__pyx_n_s__wsource));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wsource));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bsource));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 3, ((PyObject *)__pyx_n_s__bsource));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bsource));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 4, ((PyObject *)__pyx_n_s__wresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 5, ((PyObject *)__pyx_n_s__bresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 6, ((PyObject *)__pyx_n_s__mresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__qresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 7, ((PyObject *)__pyx_n_s__qresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__qresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__used));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 8, ((PyObject *)__pyx_n_s__used));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__used));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 9, ((PyObject *)__pyx_n_s__n));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bini));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 10, ((PyObject *)__pyx_n_s__bini));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bini));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__binj));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 11, ((PyObject *)__pyx_n_s__binj));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__binj));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bin));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 12, ((PyObject *)__pyx_n_s__bin));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bin));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wval));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 13, ((PyObject *)__pyx_n_s__wval));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wval));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bval));
  PyTuple_SET_ITEM(__pyx_k_tuple_22, 14, ((PyObject *)__pyx_n_s__bval));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bval));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_22));
  __pyx_k_codeobj_23 = (PyObject*)__Pyx_PyCode_New(9, 0, 15, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__bin_profile2d, 62, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":90
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile3d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=1] bins_y,
 *                   np.ndarray[np.int64_t, ndim=1] bins_z,
 */
  __pyx_k_tuple_24 = PyTuple_New(17); if (unlikely(!__pyx_k_tuple_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_24);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bins_x));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 0, ((PyObject *)__pyx_n_s__bins_x));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bins_x));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bins_y));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 1, ((PyObject *)__pyx_n_s__bins_y));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bins_y));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bins_z));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 2, ((PyObject *)__pyx_n_s__bins_z));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bins_z));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wsource));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 3, ((PyObject *)__pyx_n_s__wsource));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wsource));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bsource));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 4, ((PyObject *)__pyx_n_s__bsource));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bsource));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 5, ((PyObject *)__pyx_n_s__wresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 6, ((PyObject *)__pyx_n_s__bresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 7, ((PyObject *)__pyx_n_s__mresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__qresult));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 8, ((PyObject *)__pyx_n_s__qresult));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__qresult));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__used));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 9, ((PyObject *)__pyx_n_s__used));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__used));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 10, ((PyObject *)__pyx_n_s__n));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bini));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 11, ((PyObject *)__pyx_n_s__bini));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bini));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__binj));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 12, ((PyObject *)__pyx_n_s__binj));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__binj));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bink));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 13, ((PyObject *)__pyx_n_s__bink));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bink));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bin));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 14, ((PyObject *)__pyx_n_s__bin));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bin));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__wval));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 15, ((PyObject *)__pyx_n_s__wval));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__wval));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__bval));
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 16, ((PyObject *)__pyx_n_s__bval));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__bval));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_24));
  __pyx_k_codeobj_25 = (PyObject*)__Pyx_PyCode_New(10, 0, 17, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__bin_profile3d, 90, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":120
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def lines(np.ndarray[np.float64_t, ndim=3] image,             # <<<<<<<<<<<<<<
 *           np.ndarray[np.int64_t, ndim=1] xs,
 *           np.ndarray[np.int64_t, ndim=1] ys,
 */
  __pyx_k_tuple_26 = PyTuple_New(22); if (unlikely(!__pyx_k_tuple_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_26);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__image));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 0, ((PyObject *)__pyx_n_s__image));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__image));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__xs));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 1, ((PyObject *)__pyx_n_s__xs));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__xs));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ys));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 2, ((PyObject *)__pyx_n_s__ys));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ys));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__colors));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 3, ((PyObject *)__pyx_n_s__colors));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__colors));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__points_per_color));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 4, ((PyObject *)__pyx_n_s__points_per_color));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__points_per_color));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nx));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 5, ((PyObject *)__pyx_n_s__nx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ny));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 6, ((PyObject *)__pyx_n_s__ny));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ny));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nl));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 7, ((PyObject *)__pyx_n_s__nl));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nl));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__alpha));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 8, ((PyObject *)__pyx_n_s__alpha));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__alpha));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 9, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 10, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dx));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 11, ((PyObject *)__pyx_n_s__dx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dy));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 12, ((PyObject *)__pyx_n_s__dy));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dy));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__sx));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 13, ((PyObject *)__pyx_n_s__sx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__sx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__sy));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 14, ((PyObject *)__pyx_n_s__sy));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__sy));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__e2));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 15, ((PyObject *)__pyx_n_s__e2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__e2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__err));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 16, ((PyObject *)__pyx_n_s__err));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__err));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__x0));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 17, ((PyObject *)__pyx_n_s__x0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__x0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__x1));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 18, ((PyObject *)__pyx_n_s__x1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__x1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__y0));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 19, ((PyObject *)__pyx_n_s__y0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__y0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__y1));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 20, ((PyObject *)__pyx_n_s__y1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__y1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__has_alpha));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 21, ((PyObject *)__pyx_n_s__has_alpha));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__has_alpha));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_26));
  __pyx_k_codeobj_27 = (PyObject*)__Pyx_PyCode_New(5, 0, 22, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__lines, 120, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":179
 *     return
 * 
 * def rotate_vectors(np.ndarray[np.float64_t, ndim=3] vecs,             # <<<<<<<<<<<<<<
 *         np.ndarray[np.float64_t, ndim=2] R):
 *     cdef int nx = vecs.shape[0]
 */
  __pyx_k_tuple_28 = PyTuple_New(8); if (unlikely(!__pyx_k_tuple_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_28);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__vecs));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 0, ((PyObject *)__pyx_n_s__vecs));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__vecs));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__R));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 1, ((PyObject *)__pyx_n_s__R));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__R));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nx));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 2, ((PyObject *)__pyx_n_s__nx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ny));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 3, ((PyObject *)__pyx_n_s__ny));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ny));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__rotated));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 4, ((PyObject *)__pyx_n_s__rotated));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__rotated));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 5, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 6, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 7, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_28));
  __pyx_k_codeobj_29 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__rotate_vectors, 179, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":194
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_color_bounds(np.ndarray[np.float64_t, ndim=1] px,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=1] py,
 *                      np.ndarray[np.float64_t, ndim=1] pdx,
 */
  __pyx_k_tuple_30 = PyTuple_New(16); if (unlikely(!__pyx_k_tuple_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_30);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__px));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 0, ((PyObject *)__pyx_n_s__px));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__px));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__py));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 1, ((PyObject *)__pyx_n_s__py));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__py));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pdx));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 2, ((PyObject *)__pyx_n_s__pdx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pdx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pdy));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 3, ((PyObject *)__pyx_n_s__pdy));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pdy));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__value));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 4, ((PyObject *)__pyx_n_s__value));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__value));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__leftx));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 5, ((PyObject *)__pyx_n_s__leftx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__leftx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__rightx));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 6, ((PyObject *)__pyx_n_s__rightx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__rightx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__lefty));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 7, ((PyObject *)__pyx_n_s__lefty));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__lefty));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__righty));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 8, ((PyObject *)__pyx_n_s__righty));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__righty));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mindx));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 9, ((PyObject *)__pyx_n_s__mindx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mindx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__maxdx));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 10, ((PyObject *)__pyx_n_s__maxdx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__maxdx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 11, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mi));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 12, ((PyObject *)__pyx_n_s__mi));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mi));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ma));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 13, ((PyObject *)__pyx_n_s__ma));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ma));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 14, ((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__np));
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 15, ((PyObject *)__pyx_n_s__np));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__np));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_30));
  __pyx_k_codeobj_31 = (PyObject*)__Pyx_PyCode_New(11, 0, 16, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__get_color_bounds, 194, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":222
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_box_grids_level(np.ndarray[np.float64_t, ndim=1] left_edge,             # <<<<<<<<<<<<<<
 *                         np.ndarray[np.float64_t, ndim=1] right_edge,
 *                         int level,
 */
  __pyx_k_tuple_32 = PyTuple_New(13); if (unlikely(!__pyx_k_tuple_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_32);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__left_edge));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 0, ((PyObject *)__pyx_n_s__left_edge));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__left_edge));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__right_edge));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 1, ((PyObject *)__pyx_n_s__right_edge));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__right_edge));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__level));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 2, ((PyObject *)__pyx_n_s__level));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__level));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__left_edges));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 3, ((PyObject *)__pyx_n_s__left_edges));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__left_edges));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__right_edges));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 4, ((PyObject *)__pyx_n_s__right_edges));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__right_edges));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__levels));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 5, ((PyObject *)__pyx_n_s__levels));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__levels));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 6, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__min_index));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 7, ((PyObject *)__pyx_n_s__min_index));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__min_index));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 8, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 9, ((PyObject *)__pyx_n_s__n));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nx));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 10, ((PyObject *)__pyx_n_s__nx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__inside));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 11, ((PyObject *)__pyx_n_s__inside));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__inside));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__eps));
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 12, ((PyObject *)__pyx_n_s__eps));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__eps));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_32));
  __pyx_k_codeobj_33 = (PyObject*)__Pyx_PyCode_New(8, 0, 13, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__get_box_grids_level, 222, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":250
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_box_grids_below_level(             # <<<<<<<<<<<<<<
 *                         np.ndarray[np.float64_t, ndim=1] left_edge,
 *                         np.ndarray[np.float64_t, ndim=1] right_edge,
 */
  __pyx_k_tuple_34 = PyTuple_New(13); if (unlikely(!__pyx_k_tuple_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_34);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__left_edge));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 0, ((PyObject *)__pyx_n_s__left_edge));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__left_edge));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__right_edge));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 1, ((PyObject *)__pyx_n_s__right_edge));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__right_edge));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__level));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 2, ((PyObject *)__pyx_n_s__level));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__level));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__left_edges));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 3, ((PyObject *)__pyx_n_s__left_edges));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__left_edges));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__right_edges));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 4, ((PyObject *)__pyx_n_s__right_edges));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__right_edges));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__levels));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 5, ((PyObject *)__pyx_n_s__levels));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__levels));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 6, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__min_level));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 7, ((PyObject *)__pyx_n_s__min_level));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__min_level));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 8, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 9, ((PyObject *)__pyx_n_s__n));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nx));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 10, ((PyObject *)__pyx_n_s__nx));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nx));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__inside));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 11, ((PyObject *)__pyx_n_s__inside));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__inside));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__eps));
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 12, ((PyObject *)__pyx_n_s__eps));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__eps));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_34));
  __pyx_k_codeobj_35 = (PyObject*)__Pyx_PyCode_New(8, 0, 13, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s_36, 250, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":277
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def find_values_at_point(np.ndarray[np.float64_t, ndim=1] point,             # <<<<<<<<<<<<<<
 *                          np.ndarray[np.float64_t, ndim=2] left_edges,
 *                          np.ndarray[np.float64_t, ndim=2] right_edges,
 */
  __pyx_k_tuple_37 = PyTuple_New(14); if (unlikely(!__pyx_k_tuple_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_37);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__point));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 0, ((PyObject *)__pyx_n_s__point));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__point));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__left_edges));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 1, ((PyObject *)__pyx_n_s__left_edges));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__left_edges));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__right_edges));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 2, ((PyObject *)__pyx_n_s__right_edges));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__right_edges));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dimensions));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 3, ((PyObject *)__pyx_n_s__dimensions));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dimensions));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__field_names));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 4, ((PyObject *)__pyx_n_s__field_names));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__field_names));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__grid_objects));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 5, ((PyObject *)__pyx_n_s__grid_objects));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__grid_objects));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ind));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 6, ((PyObject *)__pyx_n_s__ind));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ind));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__gi));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 7, ((PyObject *)__pyx_n_s__gi));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__gi));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fi));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 8, ((PyObject *)__pyx_n_s__fi));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fi));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nf));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 9, ((PyObject *)__pyx_n_s__nf));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nf));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dds));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 10, ((PyObject *)__pyx_n_s__dds));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dds));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__field));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 11, ((PyObject *)__pyx_n_s__field));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__field));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__rv));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 12, ((PyObject *)__pyx_n_s__rv));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__rv));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__grid));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 13, ((PyObject *)__pyx_n_s__grid));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__grid));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_37));
  __pyx_k_codeobj_38 = (PyObject*)__Pyx_PyCode_New(6, 0, 14, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s_39, 277, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":312
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def kdtree_get_choices(np.ndarray[np.float64_t, ndim=3] data,             # <<<<<<<<<<<<<<
 *                        np.ndarray[np.float64_t, ndim=1] l_corner,
 *                        np.ndarray[np.float64_t, ndim=1] r_corner):
 */
  __pyx_k_tuple_40 = PyTuple_New(21); if (unlikely(!__pyx_k_tuple_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_40);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__data));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 0, ((PyObject *)__pyx_n_s__data));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__data));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l_corner));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 1, ((PyObject *)__pyx_n_s__l_corner));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l_corner));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__r_corner));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 2, ((PyObject *)__pyx_n_s__r_corner));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__r_corner));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 3, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 4, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 5, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dim));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 6, ((PyObject *)__pyx_n_s__dim));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dim));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n_unique));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 7, ((PyObject *)__pyx_n_s__n_unique));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n_unique));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__best_dim));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 8, ((PyObject *)__pyx_n_s__best_dim));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__best_dim));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n_best));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 9, ((PyObject *)__pyx_n_s__n_best));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n_best));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__n_grids));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 10, ((PyObject *)__pyx_n_s__n_grids));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__n_grids));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__addit));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 11, ((PyObject *)__pyx_n_s__addit));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__addit));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__my_split));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 12, ((PyObject *)__pyx_n_s__my_split));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__my_split));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__uniquedims));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 13, ((PyObject *)__pyx_n_s__uniquedims));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__uniquedims));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__uniques));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 14, ((PyObject *)__pyx_n_s__uniques));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__uniques));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__split));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 15, ((PyObject *)__pyx_n_s__split));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__split));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__my_max));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 16, ((PyObject *)__pyx_n_s__my_max));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__my_max));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__skipit));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 17, ((PyObject *)__pyx_n_s__skipit));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__skipit));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__tarr));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 18, ((PyObject *)__pyx_n_s__tarr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__tarr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__less_ids));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 19, ((PyObject *)__pyx_n_s__less_ids));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__less_ids));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__greater_ids));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 20, ((PyObject *)__pyx_n_s__greater_ids));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__greater_ids));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_40));
  __pyx_k_codeobj_41 = (PyObject*)__Pyx_PyCode_New(3, 0, 21, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__kdtree_get_choices, 312, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "yt/utilities/lib/misc_utilities.pyx":371
 * 
 * 
 * def grow_flagging_field(oofield):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")
 *     cdef np.ndarray[np.uint8_t, ndim=3] nfield
 */
  __pyx_k_tuple_42 = PyTuple_New(12); if (unlikely(!__pyx_k_tuple_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 371; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_42);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__oofield));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 0, ((PyObject *)__pyx_n_s__oofield));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__oofield));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ofield));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 1, ((PyObject *)__pyx_n_s__ofield));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ofield));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nfield));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 2, ((PyObject *)__pyx_n_s__nfield));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nfield));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 3, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 4, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 5, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ni));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 6, ((PyObject *)__pyx_n_s__ni));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ni));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nj));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 7, ((PyObject *)__pyx_n_s__nj));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nj));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nk));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 8, ((PyObject *)__pyx_n_s__nk));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nk));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__oi));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 9, ((PyObject *)__pyx_n_s__oi));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__oi));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__oj));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 10, ((PyObject *)__pyx_n_s__oj));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__oj));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ok));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 11, ((PyObject *)__pyx_n_s__ok));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ok));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_42));
  __pyx_k_codeobj_43 = (PyObject*)__Pyx_PyCode_New(1, 0, 12, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_20, __pyx_n_s__grow_flagging_field, 371, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 371; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initmisc_utilities(void); /*proto*/
PyMODINIT_FUNC initmisc_utilities(void)
#else
PyMODINIT_FUNC PyInit_misc_utilities(void); /*proto*/
PyMODINIT_FUNC PyInit_misc_utilities(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_misc_utilities(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4(__Pyx_NAMESTR("misc_utilities"), __pyx_methods, __Pyx_DOCSTR(__pyx_k_17), 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_b = PyImport_AddModule(__Pyx_NAMESTR(__Pyx_BUILTIN_MODULE_NAME)); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (__Pyx_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_module_is_main_yt__utilities__lib__misc_utilities) {
    if (__Pyx_SetAttrString(__pyx_m, "__name__", __pyx_n_s____main__) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "yt/utilities/lib/misc_utilities.pyx":26
 * """
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * cimport cython
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__numpy), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 26; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__np, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 26; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":37
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile1d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.float64_t, ndim=1] wsource,
 *                   np.ndarray[np.float64_t, ndim=1] bsource,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_1bin_profile1d, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__bin_profile1d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":62
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile2d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=1] bins_y,
 *                   np.ndarray[np.float64_t, ndim=1] wsource,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_3bin_profile2d, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__bin_profile2d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":90
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def bin_profile3d(np.ndarray[np.int64_t, ndim=1] bins_x,             # <<<<<<<<<<<<<<
 *                   np.ndarray[np.int64_t, ndim=1] bins_y,
 *                   np.ndarray[np.int64_t, ndim=1] bins_z,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_5bin_profile3d, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__bin_profile3d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":120
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def lines(np.ndarray[np.float64_t, ndim=3] image,             # <<<<<<<<<<<<<<
 *           np.ndarray[np.int64_t, ndim=1] xs,
 *           np.ndarray[np.int64_t, ndim=1] ys,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_7lines, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__lines, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":179
 *     return
 * 
 * def rotate_vectors(np.ndarray[np.float64_t, ndim=3] vecs,             # <<<<<<<<<<<<<<
 *         np.ndarray[np.float64_t, ndim=2] R):
 *     cdef int nx = vecs.shape[0]
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_9rotate_vectors, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__rotate_vectors, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":194
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_color_bounds(np.ndarray[np.float64_t, ndim=1] px,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=1] py,
 *                      np.ndarray[np.float64_t, ndim=1] pdx,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_11get_color_bounds, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__get_color_bounds, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":222
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_box_grids_level(np.ndarray[np.float64_t, ndim=1] left_edge,             # <<<<<<<<<<<<<<
 *                         np.ndarray[np.float64_t, ndim=1] right_edge,
 *                         int level,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_13get_box_grids_level, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__get_box_grids_level, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":250
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def get_box_grids_below_level(             # <<<<<<<<<<<<<<
 *                         np.ndarray[np.float64_t, ndim=1] left_edge,
 *                         np.ndarray[np.float64_t, ndim=1] right_edge,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_15get_box_grids_below_level, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_36, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":277
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def find_values_at_point(np.ndarray[np.float64_t, ndim=1] point,             # <<<<<<<<<<<<<<
 *                          np.ndarray[np.float64_t, ndim=2] left_edges,
 *                          np.ndarray[np.float64_t, ndim=2] right_edges,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_17find_values_at_point, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_39, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":312
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def kdtree_get_choices(np.ndarray[np.float64_t, ndim=3] data,             # <<<<<<<<<<<<<<
 *                        np.ndarray[np.float64_t, ndim=1] l_corner,
 *                        np.ndarray[np.float64_t, ndim=1] r_corner):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_19kdtree_get_choices, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__kdtree_get_choices, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 312; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":371
 * 
 * 
 * def grow_flagging_field(oofield):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")
 *     cdef np.ndarray[np.uint8_t, ndim=3] nfield
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_2yt_9utilities_3lib_14misc_utilities_21grow_flagging_field, NULL, __pyx_n_s_21); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 371; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__grow_flagging_field, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 371; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/utilities/lib/misc_utilities.pyx":1
 * """             # <<<<<<<<<<<<<<
 * Simple utilities that don't fit anywhere else
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s____test__, ((PyObject *)__pyx_t_1)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;

  /* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    __Pyx_AddTraceback("init yt.utilities.lib.misc_utilities", __pyx_clineno, __pyx_lineno, __pyx_filename);
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init yt.utilities.lib.misc_utilities");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif /* CYTHON_REFNANNY */

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name) {
    PyObject *result;
    result = PyObject_GetAttr(dict, name);
    if (!result) {
        if (dict != __pyx_b) {
            PyErr_Clear();
            result = PyObject_GetAttr(__pyx_b, name);
        }
        if (!result) {
            PyErr_SetObject(PyExc_NameError, name);
        }
    }
    return result;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%s() takes %s %" CYTHON_FORMAT_SSIZE_T "d positional argument%s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (!type) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (Py_TYPE(obj) == type) return 1;
    }
    else {
        if (PyObject_TypeCheck(obj, type)) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%s' has incorrect type (expected %s, got %s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

static CYTHON_INLINE int __Pyx_IsLittleEndian(void) {
  unsigned int n = 1;
  return *(unsigned char*)(&n) != 0;
}
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t < '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1) /* First char was not a digit */
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count; /* Consume from buffer string */
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break; /* breaks both loops as ctx->enc_count == 0 */
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue; /* empty struct */
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static CYTHON_INLINE PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        if (isspace(*ts))
            continue;
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
                return ts;
      case ' ':
      case 10:
      case 13:
        ++ts;
        break;
      case '<':
        if (!__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T': /* substruct */
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0; /* Erase processed last struct element */
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}': /* end of substruct; either repeat or move on */
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0; /* Erase processed last struct element */
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }        /* fall through */
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 's': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
        } else {
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_count = ctx->new_count;
          ctx->enc_packmode = ctx->new_packmode;
          ctx->enc_type = *ts;
          ctx->is_complex = got_Z;
        }
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}
static CYTHON_INLINE void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static CYTHON_INLINE int __Pyx_GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  if (obj == Py_None || obj == NULL) {
    __Pyx_ZeroBuffer(buf);
    return 0;
  }
  buf->buf = NULL;
  if (__Pyx_GetBuffer(obj, buf, flags) == -1) goto fail;
  if (buf->ndim != nd) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if ((unsigned)buf->itemsize != dtype->size) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_ZeroBuffer(buf);
  return -1;
}
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (info->buf == NULL) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_Format(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    #if PY_VERSION_HEX < 0x02050000
    if (PyClass_Check(type)) {
    #else
    if (PyType_Check(type)) {
    #endif
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        #if PY_VERSION_HEX < 0x02050000
            if (PyInstance_Check(type)) {
                type = (PyObject*) ((PyInstanceObject*)type)->in_class;
                Py_INCREF(type);
            }
            else {
                type = 0;
                PyErr_SetString(PyExc_TypeError,
                    "raise: exception must be an old-style class or instance");
                goto raise_error;
            }
        #else
            type = (PyObject*) Py_TYPE(type);
            Py_INCREF(type);
            if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
                PyErr_SetString(PyExc_TypeError,
                    "raise: exception class must be a subclass of BaseException");
                goto raise_error;
            }
        #endif
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else /* Python 3+ */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *args;
        if (!value)
            args = PyTuple_New(0);
        else if (PyTuple_Check(value)) {
            Py_INCREF(value);
            args = value;
        }
        else
            args = PyTuple_Pack(1, value);
        if (!args)
            goto bad;
        owned_instance = PyEval_CallObject(type, args);
        Py_DECREF(args);
        if (!owned_instance)
            goto bad;
        value = owned_instance;
        if (!PyExceptionInstance_Check(value)) {
            PyErr_Format(PyExc_TypeError,
                         "calling %R should have returned an instance of "
                         "BaseException, not %R",
                         type, Py_TYPE(value));
            goto bad;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause && cause != Py_None) {
        PyObject *fixed_cause;
        if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        }
        else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        }
        else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%s to unpack",
                 index, (index == 1) ? "" : "s");
}

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    CYTHON_UNUSED PyObject *getbuffer_cobj;
  #if PY_VERSION_HEX >= 0x02060000
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
  #endif
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
  #if PY_VERSION_HEX < 0x02060000
    if (obj->ob_type->tp_dict &&
        (getbuffer_cobj = PyMapping_GetItemString(obj->ob_type->tp_dict,
                                             "__pyx_getbuffer"))) {
        getbufferproc func;
      #if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION == 0)
        func = (getbufferproc) PyCapsule_GetPointer(getbuffer_cobj, "getbuffer(obj, view, flags)");
      #else
        func = (getbufferproc) PyCObject_AsVoidPtr(getbuffer_cobj);
      #endif
        Py_DECREF(getbuffer_cobj);
        if (!func)
            goto fail;
        return func(obj, view, flags);
    } else {
        PyErr_Clear();
    }
  #endif
    PyErr_Format(PyExc_TypeError, "'%100s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
#if PY_VERSION_HEX < 0x02060000
fail:
#endif
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    CYTHON_UNUSED PyObject *releasebuffer_cobj;
    if (!obj) return;
  #if PY_VERSION_HEX >= 0x02060000
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
  #endif
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) { __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view); return; }
  #if PY_VERSION_HEX < 0x02060000
    if (obj->ob_type->tp_dict &&
        (releasebuffer_cobj = PyMapping_GetItemString(obj->ob_type->tp_dict,
                                                      "__pyx_releasebuffer"))) {
        releasebufferproc func;
      #if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION == 0)
        func = (releasebufferproc) PyCapsule_GetPointer(releasebuffer_cobj, "releasebuffer(obj, view)");
      #else
        func = (releasebufferproc) PyCObject_AsVoidPtr(releasebuffer_cobj);
      #endif
        Py_DECREF(releasebuffer_cobj);
        if (!func)
            goto fail;
        func(obj, view);
        return;
    } else {
        PyErr_Clear();
    }
  #endif
    goto nofail;
#if PY_VERSION_HEX < 0x02060000
fail:
#endif
    PyErr_WriteUnraisable(obj);
nofail:
    Py_DECREF(obj);
    view->obj = NULL;
}
#endif /*  PY_MAJOR_VERSION < 3 */


    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, long level) {
    PyObject *py_import = 0;
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    py_import = __Pyx_GetAttrString(__pyx_b, "__import__");
    if (!py_import)
        goto bad;
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    #if PY_VERSION_HEX >= 0x02050000
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                /* try package relative import first */
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0; /* try absolute import on failure */
        }
        #endif
        if (!module) {
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
        }
    }
    #else
    if (level>0) {
        PyErr_SetString(PyExc_RuntimeError, "Relative import is not supported for Python <=2.4.");
        goto bad;
    }
    module = PyObject_CallFunctionObjArgs(py_import,
        name, global_dict, empty_dict, list, NULL);
    #endif
bad:
    Py_XDECREF(empty_list);
    Py_XDECREF(py_import);
    Py_XDECREF(empty_dict);
    return module;
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t val) {
    const Py_intptr_t neg_one = (Py_intptr_t)-1, const_zero = (Py_intptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(Py_intptr_t) == sizeof(char))  ||
        (sizeof(Py_intptr_t) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(Py_intptr_t) == sizeof(int)) ||
               (sizeof(Py_intptr_t) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(Py_intptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_npy_int64(npy_int64 val) {
    const npy_int64 neg_one = (npy_int64)-1, const_zero = (npy_int64)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(npy_int64) == sizeof(char))  ||
        (sizeof(npy_int64) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(npy_int64) == sizeof(int)) ||
               (sizeof(npy_int64) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(npy_int64) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int64),
                                     little, !is_unsigned);
    }
}

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        float denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(a, a);
                    case 3:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, a);
                    case 4:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_absf(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        double denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(a, a);
                    case 3:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, a);
                    case 4:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_abs(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject* x) {
    const unsigned char neg_one = (unsigned char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned char" :
                    "value too large to convert to unsigned char");
            }
            return (unsigned char)-1;
        }
        return (unsigned char)val;
    }
    return (unsigned char)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject* x) {
    const unsigned short neg_one = (unsigned short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned short" :
                    "value too large to convert to unsigned short");
            }
            return (unsigned short)-1;
        }
        return (unsigned short)val;
    }
    return (unsigned short)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject* x) {
    const unsigned int neg_one = (unsigned int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned int" :
                    "value too large to convert to unsigned int");
            }
            return (unsigned int)-1;
        }
        return (unsigned int)val;
    }
    return (unsigned int)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject* x) {
    const char neg_one = (char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to char" :
                    "value too large to convert to char");
            }
            return (char)-1;
        }
        return (char)val;
    }
    return (char)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject* x) {
    const short neg_one = (short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to short" :
                    "value too large to convert to short");
            }
            return (short)-1;
        }
        return (short)val;
    }
    return (short)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject* x) {
    const signed char neg_one = (signed char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed char" :
                    "value too large to convert to signed char");
            }
            return (signed char)-1;
        }
        return (signed char)val;
    }
    return (signed char)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject* x) {
    const signed short neg_one = (signed short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed short" :
                    "value too large to convert to signed short");
            }
            return (signed short)-1;
        }
        return (signed short)val;
    }
    return (signed short)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject* x) {
    const signed int neg_one = (signed int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed int" :
                    "value too large to convert to signed int");
            }
            return (signed int)-1;
        }
        return (signed int)val;
    }
    return (signed int)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject* x) {
    const unsigned long neg_one = (unsigned long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned long");
            return (unsigned long)-1;
        }
        return (unsigned long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned long");
                return (unsigned long)-1;
            }
            return (unsigned long)PyLong_AsUnsignedLong(x);
        } else {
            return (unsigned long)PyLong_AsLong(x);
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned long)-1;
        val = __Pyx_PyInt_AsUnsignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned PY_LONG_LONG");
            return (unsigned PY_LONG_LONG)-1;
        }
        return (unsigned PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned PY_LONG_LONG");
                return (unsigned PY_LONG_LONG)-1;
            }
            return (unsigned PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (unsigned PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsUnsignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject* x) {
    const long neg_one = (long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to long");
            return (long)-1;
        }
        return (long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to long");
                return (long)-1;
            }
            return (long)PyLong_AsUnsignedLong(x);
        } else {
            return (long)PyLong_AsLong(x);
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long)-1;
        val = __Pyx_PyInt_AsLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject* x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to PY_LONG_LONG");
            return (PY_LONG_LONG)-1;
        }
        return (PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to PY_LONG_LONG");
                return (PY_LONG_LONG)-1;
            }
            return (PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject* x) {
    const signed long neg_one = (signed long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed long");
            return (signed long)-1;
        }
        return (signed long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed long");
                return (signed long)-1;
            }
            return (signed long)PyLong_AsUnsignedLong(x);
        } else {
            return (signed long)PyLong_AsLong(x);
        }
    } else {
        signed long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed long)-1;
        val = __Pyx_PyInt_AsSignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject* x) {
    const signed PY_LONG_LONG neg_one = (signed PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed PY_LONG_LONG");
            return (signed PY_LONG_LONG)-1;
        }
        return (signed PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed PY_LONG_LONG");
                return (signed PY_LONG_LONG)-1;
            }
            return (signed PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (signed PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        signed PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsSignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        #if PY_VERSION_HEX < 0x02050000
        return PyErr_Warn(NULL, message);
        #else
        return PyErr_WarnEx(NULL, message, 1);
        #endif
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%s.%s is not a type object",
            module_name, class_name);
        goto bad;
    }
    if (!strict && (size_t)((PyTypeObject *)result)->tp_basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        #if PY_VERSION_HEX < 0x02050000
        if (PyErr_Warn(NULL, warning) < 0) goto bad;
        #else
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
        #endif
    }
    else if ((size_t)((PyTypeObject *)result)->tp_basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%s.%s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,            /*int argcount,*/
        0,            /*int kwonlyargcount,*/
        0,            /*int nlocals,*/
        0,            /*int stacksize,*/
        0,            /*int flags,*/
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,      /*int firstlineno,*/
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_globals = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_globals = PyModule_GetDict(__pyx_m);
    if (!py_globals) goto bad;
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        py_globals,          /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else  /* Python 3+ has unicode identifiers */
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}


/* Type Conversion Functions */

static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}

static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_VERSION_HEX < 0x03000000
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_VERSION_HEX < 0x03000000
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%s__ returned non-%s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}

static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject* x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}

static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
#if PY_VERSION_HEX < 0x02050000
   if (ival <= LONG_MAX)
       return PyInt_FromLong((long)ival);
   else {
       unsigned char *bytes = (unsigned char *) &ival;
       int one = 1; int little = (int)*(unsigned char*)&one;
       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
   }
#else
   return PyInt_FromSize_t(ival);
#endif
}

static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject* x) {
   unsigned PY_LONG_LONG val = __Pyx_PyInt_AsUnsignedLongLong(x);
   if (unlikely(val == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred())) {
       return (size_t)-1;
   } else if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
       PyErr_SetString(PyExc_OverflowError,
                       "value too large to convert to size_t");
       return (size_t)-1;
   }
   return (size_t)val;
}


#endif /* Py_PYTHON_H */
